(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

signature AST =
  sig

      include Lexeme where type identifier = Lexeme.identifier
                     where type operator = Lexeme.operator
		     where type constant = Lexeme.constant

      type pos
      type region = pos * pos
      type label
      structure LabKey : ORD_KEY where type ord_key = label
      val mkFreshLab  : unit -> label
      val labToString : label -> string

      structure ExtraInfo : ExtraInfo
      type scopeInfo = ExtraInfo.scopeInfo
      type symInfo = ExtraInfo.symInfo

      (* For lookup function *)
      datatype namespace =
	  VarNames
	| StructNames

      datatype expression =
          Variable of identifier * astInfo
        | Constant of constant * astInfo
        | Conditional of expression * expression * expression * astInfo
        | Application of expression * expression list * astInfo * label
        | Operation of operator * expression list * astInfo
        | Assignment of expression * operator option * expression * astInfo
        | Comma of expression * expression * astInfo
        | Cast of syntacticType * expression * astInfo
	| CastList of syntacticType * initializer list * astInfo
        | SizeOfExpr of expression * astInfo
        | SizeOfTypeExpr of syntacticType * astInfo
	| AlignOfExpr of expression * astInfo
	| AlignOfTypeExpr of syntacticType * astInfo
        | FieldDirect of expression * identifier * astInfo
        | FieldIndirect of expression * identifier * astInfo
        | Postfix of operator * expression * astInfo
        | Index of expression * expression * astInfo
	| Extension of statement * astInfo (* gcc allows blocks as expressions *)
        
      and syntacticType = 
	  SyntacticType of storageClass option * typeSpecifier *
	                   typeQualifier list *  declarator list

      (* Aggregate types *)
      and aggregate =
            Struct
          | Union

      (* Storage classes *)
      and storageClass =
            Register
          | Auto 
          | Static 
          | Extern
          | Typedef 

      (* Type qualifiers *)
      and typeQualifier = 
            Const
          | Volatile
	  | Inline
	  | NonConst
	  | Qualifier of identifier

      and declarator = 
            Pointer of typeQualifier list
          | Array of expression option (* must be integer constant! *)
          | BitField of expression (* must be integer constant! *)
          | Function of declaration list
          | OldFunction of (region * identifier) list * region option (* option marks ellipsis *)

      (* Basic types *)
      and typeSpecifier =
	  Void
          | Char
          | UnsignedChar
          | ShortInt
          | UnsignedShortInt    
          | Int                 
          | UnsignedInt         
          | LongInt             
	  | LongLong			(* new 64 bit type *)
          | UnsignedLongInt     
          | UnsignedLongLong		(* new 64 bit type *)
          | Float
	| Label
          | Double              
          | LongDouble              
          | NamedType of identifier
          | Enum of identifier option * (identifier * expression option) list
          | Aggregate of aggregate * identifier option * declaration list
          | NamedAggregate of aggregate * identifier
	| TypeOfExpr of expression * astInfo		(* johno added this *)
	| TypeOfTypeExpr of syntacticType * astInfo
	  
      (* Declarations *)
      and declaration =
            EllipsisDecl of {region : region}
          | VarDecl of {name  : identifier, 
			region : region,
			typ   : syntacticType,
			init  : initializer option}
          | TypeDef of {name  : identifier,
			region : region,
			typ   : syntacticType}
          | FunDef of {name   : identifier,
		       typ    : syntacticType,
		       region : region,
		       return : syntacticType,
		       params : declaration list,
		       decls  : declaration list,
		       body   : statement list,
		       lookup : lookupFun}
          | FunDecl of {name   : identifier,
			typ    : syntacticType,
			region : region,
			return : syntacticType,
			params : declaration list option}

      and initializer =
            Initial of expression
          | InitialList of initializer list


      and statement =
            Return of expression option * identifier option * astInfo
	  (* notates enclosing fun *)
          | Expression of expression option * astInfo
          | IfThen of expression * statement * astInfo
          | IfThenElse of expression * statement * statement * astInfo
          | WhileDo of expression * statement * astInfo
          | DoWhile of statement * expression * astInfo
          | For of expression option * expression option * expression option *
    	  statement * astInfo
          | GoTo of identifier * astInfo
          | Switch of expression * statement * astInfo
          | Case of expression option * statement option * astInfo
          | Labeled of identifier * statement option * astInfo
          | Break of astInfo
          | Continue of astInfo
          | Compound of {decls : declaration list,
			 body : statement list,
			 region : region,
			 extraInfo : ExtraInfo.astInfo option ref,
			 lookup : lookupFun}

      and program =
	  Program of {decls : declaration list,
		      lookup : lookupFun}

      and symbolInfo = SymInfo of {typ : syntacticType,
				   typLookup : lookupFun,
				   isParam : bool,
				   isGlobal : bool,
				   scopeInfo : scopeInfo option,
				   symInfo : symInfo option ref}

      withtype lookupFun = namespace -> identifier -> symbolInfo option

      and astInfo = {region : region,
		      extraInfo : ExtraInfo.astInfo option ref,
		      typeInfo : syntacticType option}
	  
      val getExprRegion : expression -> region
      val getDeclRegion : declaration -> region
      val getStmtRegion : statement -> region
      val typeQualEq : typeQualifier * typeQualifier -> bool

  end

