(* Chris Harrelson - August 6, 1999
 * Code to typecheck a C AST *)

structure typeCheck : typeCheck =
   struct

      structure AST = typeCheckCParser.AST
      open AST

      exception Unsupported of region * string
      exception ParseError of region * string
      exception UnifyError of region * string

      (* datatypes *)

      datatype general_type =
	 NoGeneralType | ArithmeticType of {is_int : bool, is_const : bool} | UnionType | StructureType | PointerType of {is_void : bool}
       | FunctionType of {is_void : bool} | ArrayType | BitfieldType | StringType | CharacterType

      type program_function_type = (SourceMap.sourcemap * AST.program -> (AST.program option)) option
      type statement_function_type = (SourceMap.sourcemap * AST.statement -> (AST.statement option)) option
      type expression_function_type = (SourceMap.sourcemap * AST.expression -> (AST.expression option)) option
      type declaration_function_type = (SourceMap.sourcemap * AST.declaration -> (AST.declaration option)) option

      type error_report = AST.region * string

      type ast_type = {sourceMap : SourceMap.sourcemap, tree : typeCheckCParser.GenParser.result}

      datatype callback_type =
	 Callbacks of
	 {(* the whole program *)
	  program : program_function_type,
	  (* statement types *)
	  return : statement_function_type,
	  expression : statement_function_type,
	  ifThen : statement_function_type,
	  ifThenElse : statement_function_type,
	  whileDo : statement_function_type,
	  doWhile : statement_function_type,
	  for : statement_function_type,
	  goTo : statement_function_type,
	  switch : statement_function_type,
	  caseStat : statement_function_type, (* case is a reserved word, so we can't use it *)
	  labeled : statement_function_type,
	  break : statement_function_type,
	  continue : statement_function_type,
	  compound : statement_function_type,
	  (* expression types *)
	  variable : expression_function_type,
	  constant : expression_function_type,
	  conditional : expression_function_type,
	  application : expression_function_type,
	  operation : expression_function_type,
	  assignment : expression_function_type,
	  comma : expression_function_type,
	  cast : expression_function_type,
	  sizeOfExpr : expression_function_type,
	  sizeOfTypeExpr : expression_function_type,
	  fieldDirect : expression_function_type,
	  fieldIndirect : expression_function_type,
	  postfix : expression_function_type,
	  index : expression_function_type,
	  extension : expression_function_type,
	  (* declaration types *)
	  ellipsisDecl : declaration_function_type,
	  varDecl : declaration_function_type,
	  typeDef : declaration_function_type,
	  funDef : declaration_function_type,
	  funDecl : declaration_function_type}

      (* end datatypes *)

      (* functions *)

      fun folder func (el, (els, errors)) =
	 let val (new_el, new_errors) = func el
	 in
	    (new_el :: els, new_errors @ errors)
	 end

      fun check_types {sourceMap, tree} (cb as (Callbacks callbacks)) =
	 let val (tree, errors) = check_types_prog sourceMap cb tree
	 in
	    ({sourceMap = sourceMap, tree = cb_program (#program callbacks, sourceMap, tree)}, errors)
	 end

      and check_types_prog sourceMap cb (Program {decls, lookup}) =
	 let
	    val (decls, errors) = foldr (folder (check_types_declaration lookup sourceMap cb)) ([], []) decls
	 in
	    (Program {decls = decls, lookup = lookup}, errors)
	 end
	 
      and check_types_declaration lookup_param sourceMap (cb as (Callbacks callbacks)) tree =
	 case tree of
	    FunDef {name, typ=typ as SyntacticType (_, return_type, qual_list, _), region, return, params, decls, body, lookup} =>
	       let
		  val (ret_params, ret_errors1) =
		     foldr (folder (check_types_declaration lookup_param sourceMap cb))
		           ([], []) params
		  val (ret_decls, ret_errors2) =
		     foldr (folder (check_types_declaration lookup sourceMap cb))
		           ([], []) decls
		  val (ret_body, ret_errors3) =
		     foldr (folder (check_types_statement (lookup,
							   (* return type: *)
							   SyntacticType (SOME Static,
									  return_type, qual_list, []),
							   (* case promotion type (see switch statement
							    * handling code): *)
							   SyntacticType (NONE, Int, [], []))
				    sourceMap cb))
		           ([], []) body
		  val ret_val = FunDef {name = name, typ = typ, region = region, return = return,
					params = ret_params,
					decls = ret_decls,
					body = ret_body,
					lookup = lookup}
	       in
		  (cb_declaration (#funDef callbacks, sourceMap, ret_val), ret_errors1 @ ret_errors2 @ ret_errors3)
	       end
	  | VarDecl {name, region = var_region, typ, init = SOME init} =>
	       let
		  val (ret_init, ret_errors_1) = check_types_initializer (lookup_param,
									 SyntacticType (NONE, Int, [], []),
									 SyntacticType (NONE, Int, [], []))
		                                                        sourceMap (Variable (name, {region = var_region,
												    extraInfo = ref NONE,
												    (* this ref isn't my fault!! :> *)
												    typeInfo = SOME typ}))
									typ true cb init
		  val ret_errors_2 =
		     (case typ of
			 SyntacticType (_, ((NamedType id) | Aggregate (_, SOME id, _) | NamedAggregate (_, id)), _, _) =>
			    (case lookup_param StructNames id of
				SOME _ => []
			      | NONE => [(getDeclRegion tree, "Undeclared aggregate variable type")])
		       | _ => [])
	       in
		  (cb_declaration (#varDecl callbacks, sourceMap,
				   VarDecl {name = name, region = var_region, typ = typ, init = SOME init}),
		   ret_errors_2 @ ret_errors_1) (* it's quicker to append them in this order since ret_errors2 only has at most one element
						* and the order has no real significance *)
	       end
	  | VarDecl _ => (cb_declaration (#varDecl callbacks, sourceMap, tree), [])
	  | TypeDef _ => (cb_declaration (#typeDef callbacks, sourceMap, tree), [])
	  | FunDecl _ => (cb_declaration (#funDecl callbacks, sourceMap, tree), [])
	  | EllipsisDecl _ => (cb_declaration (#ellipsisDecl callbacks, sourceMap, tree), [])

      and check_types_initializer (lookup_param as (lookup, _, _)) sourceMap expr_declared_var typ_declared_var is_top cb init =
	 case init of
	    Initial expr =>
	       let
		  val (ret_expr, typ_expr, errors_expr) = check_types_expression lookup_param sourceMap NONE cb expr
		  val (_, errors) =
		     (regular_assign_check lookup_param sourceMap cb {is_initialization = true}
		                           (expr_declared_var, typ_declared_var) (ret_expr, typ_expr), [])
		     handle ParseError value => (SyntacticType (NONE, Int, [], []), [value])
		  val out_errors =
		     case errors of
			[] => []
		      | a :: rest => (getExprRegion expr, "initializer_check: incorrect type in initializer (1)") :: rest
	       (* need to also check for initializers of static values having to be constants!!! *)
	       in
		  (Initial ret_expr, out_errors @ errors_expr)
	       end
	| InitialList inits => 
	       case typ_declared_var of
		  (* if it's an array, we must be at the top level, so remove all array dimensions and just traverse over the rest
		   * to make sure it's all uniform *)
		  SyntacticType (a, b, c, (Array _) :: rest) =>
		     let
			fun remover ((Array _) :: rest) =
			   remover rest
			  | remover decl_list = decl_list
			   
			val (ret_inits, ret_errors) =
			   foldr (folder (check_types_initializer lookup_param sourceMap expr_declared_var
					                          (SyntacticType (a, b, c, remover rest)) false cb))
			         ([], []) inits
		     in
			(InitialList ret_inits, ret_errors)
		     end
		| SyntacticType (a, specifier, c, []) =>
		     let
			fun raise_error () = (inits, 
					      [(getExprRegion expr_declared_var, "initializer_check: incorrect type in initializer (2)")])

			fun check_decls_against_inits decls inits =
			   let
			      fun double_folder (decl, init, (inits, errors)) =
				 let
				    val typ_decl =
				       case decl of
					  VarDecl {typ, ...} => typ
					| TypeDef {typ, ...} => typ
					| FunDef {typ, ...} => typ
					| FunDecl {typ, ...} => typ
					| _ => raise Unsupported (getExprRegion expr_declared_var, "this should never happen")

				    val (new_init, errors_new_init) =
				       check_types_initializer lookup_param sourceMap expr_declared_var typ_decl true cb init
				 in
				    (init::inits, errors_new_init @ errors)
				 end
			      val len_decls = length decls
			      val len_inits = length inits

			      val (fixed_decls, fixed_inits, errors) =
				 if len_decls > len_inits then
				    (List.take (decls, len_inits), inits,
				     [(getExprRegion expr_declared_var,
			   "initializer_check: not as many initializer expressions as fields in the structure/union being initialized")])
				 else if len_inits > len_decls then
				    (decls, List.take (inits, len_decls),
				     [(getExprRegion expr_declared_var,
			   "initializer_check: more initializer expressions than fields in the structure/union being initialized")])
				 else
				    (decls, inits, [])
			   in
			      ListPair.foldr double_folder ([], errors) (decls, inits)
			   end

			fun check_aggregates specifier =
			   case specifier of
			      NamedType id =>
				 (case lookup StructNames id of
				     SOME (SymInfo {typ, ...}) =>
					(case typ of
					    SyntacticType (a, specifier', c, []) =>
					       check_aggregates specifier'
					  | _ => raise_error ())
				   | NONE => raise Unsupported (getExprRegion expr_declared_var,
								"initializer_check: this shouldn't happen (1)"))
			    | Aggregate (_, id, decls) => check_decls_against_inits decls inits
			    | NamedAggregate (_, id) =>
				 (case lookup StructNames id of
				     SOME (SymInfo {typ, ...}) =>
					(case typ of
					    SyntacticType (a, specifier', c, []) =>
					       check_aggregates specifier'
					  | _ => raise_error())
				   | NONE => raise Unsupported (getExprRegion expr_declared_var,
								"initializer_check: this shouldn't happen (2)"))
			    | _ =>
				 if is_top then
				    (inits, [(getExprRegion expr_declared_var,
					      "initializer_check: braces are allowed only for arrays and structures/unions")])
				 else
				    foldr (folder (check_types_initializer lookup_param sourceMap expr_declared_var typ_declared_var
						   false cb)) ([], []) inits
		       val (ret_inits, ret_errors) = check_aggregates specifier
			   handle ((ParseError value) | (Unsupported value)) => (inits, [value])
		     in
			(InitialList ret_inits, ret_errors)
		     end
		| _ =>
		     if is_top then
			raise Unsupported (getExprRegion expr_declared_var, "initializer_check: this shouldn't happen (3)")
		     else
			let
			   val (ret_inits, ret_errors) =
			      foldr (folder (check_types_initializer lookup_param sourceMap expr_declared_var typ_declared_var false
					     cb)) ([], []) inits
			in
			   (InitialList ret_inits, ret_errors)
			end
      
      and check_types_statement (lookup_param as (lookup, func_return_type, case_promotion_type)) sourceMap
	                        (cb as (Callbacks callbacks)) tree =
	 let
	    val curry_check_types_expr_long = check_types_expression lookup_param sourceMap NONE cb
	    val curry_check_types_stat = check_types_statement lookup_param sourceMap cb
	 in
	    (* statements have no real type, since they don't return values, so they only need to be recursively parsed for the
	     * expressions inside *)
	    case tree of
	       Return (SOME expr, NONE, astinfo as {region, ...}) => (* return val; *)
		  let
		     val (ret_expr, typ_expr, errors) = curry_check_types_expr_long expr
		     val cb_out = cb_statement (#return callbacks, sourceMap, Return (SOME ret_expr, NONE, astinfo))
		  in
		     (* the type in the return statement must agree with the return type of the function: *)
		     if equal_types {check_pointer_type_qualifiers=true} typ_expr func_return_type then
			(cb_out, errors)
		     else
			(cb_out,
			 (region, "stat_check: incorrect type in return statement; it must agree with the function return type") :: errors)
		  end
	     | Return (NONE, NONE, astinfo) => (* return; *)
		  (cb_statement (#return callbacks, sourceMap, tree), [])
	     | Return (_, SOME _, {region, ...}) => raise Unsupported (region, "Can't do a return statement with a non-NONE second parameter")
	     | Expression (SOME expr, astinfo) =>
		  let val (ret_expr, typ_expr, errors) = curry_check_types_expr_long expr
		  in
		     (cb_statement (#expression callbacks, sourceMap,
				    Expression (SOME ret_expr, astinfo)),
		      errors)
		  end
	     | Expression (NONE, astinfo) =>
		  (cb_statement (#expression callbacks, sourceMap, tree), [])
	     | IfThen (pred, stat_true, astinfo as {region, ...}) => (* if (pred) stat_true; *)
		  let
		     val (ret_pred, typ_pred, errors_pred) = curry_check_types_expr_long pred
		     val (ret_stat_true, errors_stat_true) = curry_check_types_stat stat_true
		     val cb_out = cb_statement (#ifThen callbacks, sourceMap, IfThen (ret_pred, ret_stat_true, astinfo))
		  in
		     (* the predicate must have an arithmetic or pointer type: *)
		     case convert_to_general_types typ_pred of
			(ArithmeticType {is_int=_, is_const=_} | PointerType {is_void=_}) =>
			   (cb_out, errors_pred @ errors_stat_true)
		      | _ =>
			   (cb_out,
			    (region,
			     "stat_check: incorrect type in predicate of if statement; it must have either an arithmetic or a pointer type")
			    :: (errors_pred @ errors_stat_true))
		  end
	     | IfThenElse (pred, stat_true, stat_false, astinfo as {region, ...}) => (* if (pred) stat_true else stat_false; *)
		  let
		     val (ret_pred, typ_pred, errors_pred) = curry_check_types_expr_long pred
		     val (ret_stat_true, errors_stat_true) = curry_check_types_stat stat_true
		     val (ret_stat_false, errors_stat_false) = curry_check_types_stat stat_false
		     val cb_out = cb_statement (#ifThenElse callbacks, sourceMap, IfThenElse (ret_pred, ret_stat_true, ret_stat_false, astinfo))
		  in
		     (* the predicate must have an arithmetic or pointer type: *)
		     case convert_to_general_types typ_pred of
			(ArithmeticType {is_int=_, is_const=_} | PointerType {is_void=_}) =>
			   (cb_out, errors_pred @ errors_stat_true @ errors_stat_false)
		      | _ =>
			   (cb_out,
			    (region, "stat_check: incorrect type in predicate of if statement; it must have an arithmetic or pointer type")
			    :: (errors_pred @ errors_stat_true @ errors_stat_false))
		  end
	     | WhileDo (pred, stat_inside, astinfo as {region, ...}) => (* while (pred) stat_inside; *)
		  let
		     val (ret_pred, typ_pred, errors_pred) = curry_check_types_expr_long pred
		     val (ret_stat_inside, errors_stat_inside) = curry_check_types_stat stat_inside
		     val cb_out = cb_statement (#whileDo callbacks, sourceMap,
						WhileDo (ret_pred, stat_inside, astinfo))
		  in
		     (* the predicate must have an arithmetic or pointer type: *)
		     case convert_to_general_types typ_pred of
			(ArithmeticType {is_int=_, is_const=_} | PointerType {is_void=_}) =>
			   (cb_out, errors_pred @ errors_stat_inside)
		      | _ =>
			   (cb_out,
			    (region, "stat_check: incorrect type in predicate of while statement; it must have an arithmetic or pointer type")
			    :: (errors_pred @ errors_stat_inside))
		  end
	     | DoWhile (stat_inside, pred, astinfo as {region, ...}) => (* do stat while (pred); *)
		  let
		     val (ret_pred, typ_pred, errors_pred) = curry_check_types_expr_long pred
		     val (ret_stat_inside, errors_stat_inside) = curry_check_types_stat stat_inside
		     val cb_out = cb_statement (#doWhile callbacks, sourceMap,
						DoWhile (ret_stat_inside, ret_pred, astinfo))
		  in
		     (* the predicate must have an arithmetic or pointer type: *)
		     case convert_to_general_types typ_pred of
			(ArithmeticType {is_int=_, is_const=_} | PointerType {is_void=_}) =>
			   (cb_out, errors_pred @ errors_stat_inside)
		      | _ => (cb_out,
			      (region,
			       "stat_check: incorrect type in predicate of do..while statement; it must have an arithmetic or pointer type")
			      :: (errors_pred @ errors_stat_inside))
		  end
	     | For (init_expr, test_expr, incr_expr, stat_inside, astinfo) => (* for (init_expr; test_expr; incr_expr) stat *)
		  let
		     fun mini_fun expr = 
			case expr of
			   SOME expr =>
			      let val (ret_expr, typ_expr, errors_expr) = curry_check_types_expr_long expr
			      in (SOME ret_expr, SOME typ_expr, errors_expr)
			      end
			 | NONE => (NONE, NONE, [])

		     val (ret_init_expr, typ_init_expr, errors_init_expr) = mini_fun init_expr
		     val (ret_test_expr, typ_test_expr, errors_test_expr) = mini_fun test_expr
		     val (ret_incr_expr, typ_incr_expr, errors_incr_expr) = mini_fun incr_expr
		     val (ret_stat_inside, errors_stat_inside) = curry_check_types_stat stat_inside
		     val cb_out = cb_statement (#for callbacks, sourceMap,
						For (ret_init_expr, ret_test_expr, ret_incr_expr, ret_stat_inside, astinfo))
		     val total_errors = errors_init_expr @ errors_test_expr @ errors_incr_expr @ errors_stat_inside
		  in
		     (* the test expression must have an arithmetic or pointer type: *)
		     case typ_test_expr of
			SOME typ =>
			   (case convert_to_general_types typ of
			       (ArithmeticType {is_int=_, is_const=_} | PointerType {is_void=_}) =>
				  (cb_out, total_errors)
			     | _ =>
				  (cb_out,
				   (getStmtRegion tree,
				  "stat_check: incorrect type in test expression of for statement; it must have an arithmetic or pointer type")
				   :: total_errors))
		      | _ => (cb_out, total_errors)
		  end
	     | GoTo _ => (* goto _ ; note that the identifier argument must be "a label located in the current function" (K&R 2, p. 224),
			  * which is not checked *)
		  (cb_statement (#goTo callbacks, sourceMap, tree), [])

	     | Switch (expr, stat, astinfo) => (* switch (expr) stat *)
		  let
		     (* The control expression (called expr here) must have an integral type, and is subject to the integral promotions.
		      * All case statements within are also converted to the promoted type. *)
		     val (expr', typ_expr', errors_expr) = curry_check_types_expr_long expr
		     val typ_expr = promote_integer_type typ_expr'
		     val ret_expr = set_expr_type expr' typ_expr
		     val (ret_stat, errors_stat) = check_types_statement (lookup, func_return_type, typ_expr) sourceMap cb stat
		     val cb_out = cb_statement (#switch callbacks, sourceMap,
						Switch (ret_expr, ret_stat, astinfo))
		     val total_errors = errors_expr @ errors_stat
		  in
		     case convert_to_general_types typ_expr of
			ArithmeticType {is_int=true, is_const=_} =>
			  (cb_out, total_errors)
		      | _ => (cb_out,
			      (getStmtRegion tree,
			       "stat_check: incorrect type in control expression of switch statement; it must have an integral type")
			      :: total_errors)
		  end
	     | Case (SOME expr, SOME stat, astinfo) => (* case expr : stat *)
		  let
		     (* the type must be promoted to that of the promoted control expression in the enclosing switch statement: *)
		     val (expr', typ_expr', errors_expr) = curry_check_types_expr_long expr
		     val ret_expr = set_expr_type expr case_promotion_type
		     val (ret_stat, errors_stat) = curry_check_types_stat stat
		     val cb_out = cb_statement (#caseStat callbacks, sourceMap,
					 Case (SOME ret_expr, SOME ret_stat, astinfo))
		     val total_errors = errors_expr @ errors_stat
		  in
		     (* the expression must have a constant integral type: *)
		     case convert_to_general_types typ_expr' of
			ArithmeticType {is_int=true, is_const=true} =>
			   (cb_out, total_errors)
		      | _ =>
			   (cb_out,
			    (getStmtRegion tree,
			     "stat_check: incorrect type in case statement expression; it must have a constant integral type")
			    :: total_errors)
		  end
	     | Case (NONE, SOME stat, astinfo) => (* default: stat *)
		  let val (ret_stat, errors_stat) = curry_check_types_stat stat
		  in 
		     (cb_statement (#caseStat callbacks, sourceMap,
				    Case (NONE, SOME ret_stat, astinfo)),
		      errors_stat)
		  end
	     (* | Labeled (id, SOME stat, astinfo) =>
		  let val (ret_stat, errors_stat) = curry_check_types_stat stat
		  in
		     (cb_statement (#labeled callbacks, sourceMap,
				    Labeled (id, ret_stat, astinfo)),
		      errors_stat)
		  end *)
	     | Break _ =>
		  (cb_statement (#break callbacks, sourceMap, tree), [])
	     | Continue _ =>
		  (cb_statement (#continue callbacks, sourceMap, tree), [])
	     | Compound {decls, body, region, extraInfo, lookup} =>
		  let
		     val (ret_decls, errors_decls) =
			foldr (folder (check_types_declaration lookup sourceMap cb)) ([], []) decls
		     val (ret_body, errors_body) = foldr (folder curry_check_types_stat) ([], []) body
		  in
		     (cb_statement (#compound callbacks, sourceMap,
				    Compound {decls = ret_decls, body = ret_body, region = region, extraInfo = extraInfo,
					      lookup = lookup}),
		      errors_decls @ errors_body)
		  end
	 end

      and check_types_expression (lookup_param as (lookup, func_return_type, case_promotion_type)) sourceMap up_node
	                         (cb as (Callbacks callbacks)) tree =
	 let
	    val curry_check_types_expr = check_types_expression lookup_param sourceMap up_node cb
	    val curry_check_types_stat = check_types_statement lookup_param sourceMap cb
	 in
	    case tree of
	       Variable (id, astinfo) => (* id *)
		  let
		     val (ret_typ, ret_errors) =
			(* check if it's a function type; if so, when it is not used as the operand of the & operator, it must
			 * be converted to a pointer *)
			case lookup VarNames id of
			   SOME (SymInfo {typ, ...}) =>
			      (case typ of
				  SyntacticType (a, b, c, (Function decls) :: rest) =>
				     (case up_node of
					 SOME (Operation (Address, _, _)) =>
					    (typ, [])
				       | _ => (SyntacticType (a, b, c, (Pointer []) :: (Function decls) :: rest), []))
				| _ => (typ, []))
			 | NONE =>
			      (case lookup StructNames id of
				  SOME (SymInfo {typ, ...}) => (typ, []) (* none of the special cases above happen here *)
				| NONE => (SyntacticType (NONE, Int, [], []),
					   [(getExprRegion tree, "expr_check: unbound variable: " ^ (Lexeme.idToString id))]))
		  in
		     (cb_expression (#variable callbacks, sourceMap,
				     Variable (id, set_type astinfo ret_typ)),
		      ret_typ, ret_errors)
		  end
	     | Constant (Integer i, astinfo) => (* i *)
		  let val typ = SyntacticType (NONE, Int, [Const], [])
		  in
		     (cb_expression (#constant callbacks, sourceMap,
				     Constant (Integer i, set_type astinfo typ)),
		      typ, [])
		  end		      
	     | Constant (BigInteger i, astinfo) =>
		  let val typ = SyntacticType (NONE, LongInt, [Const], [])
		  in
		     (cb_expression (#constant callbacks, sourceMap,
				     Constant (BigInteger i, set_type astinfo typ)),
		      typ, [])
		  end
	     | Constant (String s, astinfo) =>
		  let
		     (* if the string is the argument of &, ++, -- or sizeof(), or is on the left side of = or ., its type
		      * remains Array; otherwise, it's converted to Pointer *)
		     val typ =
			case up_node of
			   SOME e =>
			      (case e of
				  (Operation ((Address | Incr | Decr), _, _) |
				   Postfix ((Incr | Decr), _, _) |
				   SizeOfExpr _ |
				   FieldDirect _) =>
				     SyntacticType (NONE, Char, [Const], [Array (SOME (Constant (Integer 
												 (String.size (Lexeme.idToString s)),
												 astinfo)))])
				 | Assignment (tree, _, _, _) =>
				     SyntacticType (NONE, Char, [Const], [Array (SOME (Constant (Integer 
												 (String.size (Lexeme.idToString s)),
												 astinfo)))])
				 | _ => SyntacticType (NONE, Char, [Const], [Pointer []]))
			 | NONE => SyntacticType (NONE, Char, [Const], [Pointer []])

		  in
		     (cb_expression (#constant callbacks, sourceMap,
				     Constant (String s, set_type astinfo typ)),
		      typ, [])
		  end
	     | Constant (Character c, astinfo) =>
		  let val typ = SyntacticType (NONE, Char, [Const], [])
		  in
		     (cb_expression (#constant callbacks, sourceMap,
				     Constant (Character c, set_type astinfo typ)),
		      typ, [])
		  end
	     | Constant (Real r, astinfo) =>
		  let val typ = SyntacticType (NONE, Double, [Const], [])
		  in
		     (cb_expression (#constant callbacks, sourceMap,
				     Constant (Real r, set_type astinfo typ)),
		      typ, [])
		  end
	     | Conditional (pred, expr_true, expr_false, astinfo as {region, ...}) =>
		  let
		     val (ret_pred, _, errors_pred) = curry_check_types_expr pred
		     val (ret_expr_true, typ_expr_true, errors_expr_true) = curry_check_types_expr expr_true
		     val (ret_expr_false, typ_expr_false, errors_expr_false) = curry_check_types_expr expr_false
		     val (typ, errors_typ) =
			(unify_types {is_ptr_minus=false} (expr_true, typ_expr_true) (expr_false,typ_expr_false), [])
			handle UnifyError (_, str) => (typ_expr_true,
						       [(region, "expr_check: incompatible types in true and false clauses of conditional expression; the value of the conditional needs to be the unified, common type of the two")])
		  in
		     (cb_expression (#conditional callbacks, sourceMap,
				     Conditional (ret_pred, ret_expr_true, ret_expr_false, set_type astinfo typ)),
		      typ, errors_pred @ errors_expr_true @ errors_expr_false @ errors_typ)
		  end
	     | Application (call_expr, param_expr_list, astinfo, lab) =>
		  let
		     val (ret_call_expr, typ_call_expr, errors_call_expr) = curry_check_types_expr call_expr
		     fun check_then_accum (expr, (a, b, c)) =
			let val (expr, typ, errors) = curry_check_types_expr expr
			in (expr :: a, typ :: b, errors @ c)
			end
		     val (ret_param_expr_list, typ_param_expr_list, errors_param_expr_list) = foldr check_then_accum ([], [], []) param_expr_list
		     val promoted_argument_types = map promote_integer_type typ_param_expr_list
		     val total_errors = errors_call_expr @ errors_param_expr_list
		  in
		     case typ_call_expr of
			SyntacticType (class, return_type, qualifiers, (OldFunction params)::_) =>
			   let
			      val ret_type = SyntacticType (class, return_type, qualifiers, [])
			   in
			      (cb_expression (#application callbacks, sourceMap,
					      Application (ret_call_expr, ret_param_expr_list, set_type astinfo ret_type, lab)),
			       ret_type,
			       (getExprRegion tree, "Old function definitions aren't implemented yet") :: total_errors)
			   end
		      | SyntacticType (class, return_type, qualifiers, (Function decls)::_) =>
			   let
			      val ret_typ = SyntacticType (class, return_type, qualifiers, [])
			      val _ = check_argument_types_for_new_functions promoted_argument_types decls
			   in
			      (cb_expression (#application callbacks, sourceMap,
					      Application (ret_call_expr, ret_param_expr_list, set_type astinfo ret_typ, lab)),
			       ret_typ, total_errors)
			   end
		      | _ =>
			   let
			      val ret_typ = SyntacticType (NONE, Int, [], [])
			   in
			      (cb_expression (#application callbacks, sourceMap,
					      Application (ret_call_expr, ret_param_expr_list, set_type astinfo ret_typ, lab)),
			       ret_typ,
			       (getExprRegion tree, "expr_check: incorrect type in function application") :: total_errors)
			   end
		  end
	     | Operation _ =>
		  let
		     val (ret_expr, typ_expr, errors_expr) = check_types_operation lookup_param sourceMap cb tree
		  in
		     (cb_expression (#operation callbacks, sourceMap, ret_expr), typ_expr, errors_expr)
		  end
	     | Assignment _ =>
		  let
		     val (ret_expr, typ_expr, errors_expr) = check_types_assignment lookup_param sourceMap cb tree
		  in
		     (cb_expression (#assignment callbacks, sourceMap, ret_expr), typ_expr, errors_expr)
		  end
	     | Comma (expr1, expr2, astinfo) =>
		  let
		     val (ret_expr1, _, errors_expr1) = curry_check_types_expr expr1
		     val (ret_expr2, comma_type, errors_expr2) =  curry_check_types_expr expr2
		  in
		     (cb_expression (#comma callbacks, sourceMap,
				     Comma (ret_expr1, ret_expr2, set_type astinfo comma_type)),
		      comma_type, errors_expr1 @ errors_expr2)
		  end
	     | Cast (typ, expr, astinfo) =>
		  let
		     val (ret_expr, typ_expr, errors_expr) = curry_check_types_expr expr
		     val expr_default = Cast (typ, ret_expr, set_type astinfo typ)
		     val cb_out = cb_expression (#cast callbacks, sourceMap, expr_default)
		     val default_out = (cb_out, typ, errors_expr)
		  in
		     case (convert_to_general_types typ, convert_to_general_types typ_expr) of
			(ArithmeticType _, ArithmeticType _) =>
			   default_out
		      | (PointerType _, ArithmeticType {is_int=true, is_const=_}) =>
			   (* casts are allowed to and from pointers and suff. large integral types, between pointer types, between
			    * function types *)
			   (case (typ, typ_expr) of
			       (SyntacticType (_, spec1, _, _), SyntacticType (_, spec2, _, _)) =>
				  if bigger_or_equal_data_size spec2 spec1 then
				     default_out
				  else
				     (cb_out, typ,
				      (getExprRegion tree, 
				       "expr_check: integer type is not large enough to contain a converted pointer value") :: errors_expr))
		      | (ArithmeticType {is_int=true, is_const=_}, PointerType _) =>
			   default_out
		      | (PointerType _, PointerType _) =>
			   default_out
		      | (FunctionType _, FunctionType _) =>
			   default_out
		      | (_, _) =>
			   if equal_types {check_pointer_type_qualifiers=true} typ typ_expr then
			      default_out
			   else
			      (cb_out, typ, [(getExprRegion tree, "expr_check: illegal type cast")])
		  end
	     | SizeOfExpr (expr, astinfo) =>
		  let
		     val (ret_expr, typ_expr, errors_expr) = curry_check_types_expr expr
		     val typ = SyntacticType (NONE, Int, [Const], [])
		     val cb_out = cb_expression (#sizeOfExpr callbacks, sourceMap, SizeOfExpr (expr, set_type astinfo typ))
		  in
		     case convert_to_general_types typ_expr of
			FunctionType _ =>
			   (cb_out, typ,
			    (getExprRegion tree,
			     "expr_check: incorrect argument to the 'sizeof()' operator; the argument may not have a function type")
			    :: errors_expr)
		      | BitfieldType =>
			   (cb_out, typ,
			    (getExprRegion tree, 
			     "expr_check: incorrect argument to the 'sizeof()' operator; the argument may not have a bitfield type")
			    :: errors_expr)
		      | _ =>
			   (cb_out, typ, errors_expr)
		  end
	     | SizeOfTypeExpr (syntacticType, astinfo) =>
		  let
		     val typ = SyntacticType (SOME Static, Int, [Const], [])
		     val cb_out = cb_expression (#sizeOfTypeExpr callbacks, sourceMap,
						 SizeOfTypeExpr (syntacticType, set_type astinfo typ))
		  in
		     case convert_to_general_types typ of
			FunctionType _ =>
			   (cb_out, typ,
			    [(getExprRegion tree,
			      "expr_check: incorrect argument to the 'sizeof()' operator; the argument may not have a function type")])
		      | BitfieldType =>
			   (cb_out, typ,
			    [(getExprRegion tree,
			      "expr_check: incorrect argument to the 'sizeof()' operator; the argument may not have a bitfield type")])
		      | _ =>
			   (cb_out, typ, [])
		  end
	     | ((FieldIndirect (expr, id, astinfo)) | (FieldDirect (expr, id, astinfo))) =>
		  let
		     val (ret_expr, typ_expr, errors_expr) = curry_check_types_expr expr
		    
		     val error_value = (SyntacticType (NONE, Int, [], []),
					[(getExprRegion tree,
					  "expr_check: the argument to the direct ('.') or indirect ('->') field selector operator must be a struct or a union")])

		     fun aggregate_iter (SyntacticType (_, Aggregate (_, _, decl_list), _, [])) =
			let
			   val (typ, errors) =
			      (find_type_of_element_by_name decl_list id, [])
			      handle ParseError (_, str) => (SyntacticType (NONE, Int, [], []),
							     [(getExprRegion tree, "expr_check: " ^ str)])
			in
			   (typ, errors @ errors_expr)
			end
		       | aggregate_iter (SyntacticType (_, NamedType id, _, [])) =
			(case lookup StructNames id of
			    SOME (SymInfo {typ, ...}) => aggregate_iter typ
			  | NONE => error_value)
		       | aggregate_iter (SyntacticType (_, NamedAggregate (_, id), _, [])) =
			    (case lookup StructNames id of
				SOME (SymInfo {typ, ...}) => aggregate_iter typ
			      | NONE => error_value)
		       | aggregate_iter _ = error_value

		     val (ret_typ, errors_typ) =
			case tree of
			   FieldDirect _ => aggregate_iter typ_expr
			  | _ => (* field indirect *)
			      (case typ_expr of
				  SyntacticType (a, b, c, [Pointer _]) => aggregate_iter (SyntacticType (a, b, c, []))
				| _ => error_value)

		     val cb_out =
			case tree of
			   FieldDirect _ =>
			      cb_expression (#fieldDirect callbacks, sourceMap,
					     FieldDirect (ret_expr, id, set_type astinfo ret_typ))
			 | _ => (* field indirect *)
			      cb_expression (#fieldIndirect callbacks, sourceMap,
					     FieldIndirect (ret_expr, id, set_type astinfo ret_typ))
		  in
		     (cb_out, ret_typ, errors_typ @ errors_expr)
		  end
	     | Postfix _ =>
		  let val (ret_expr, typ_expr, errors_expr) = check_types_operation lookup_param sourceMap cb tree
		  in
		     (cb_expression (#postfix callbacks, sourceMap, ret_expr), typ_expr, errors_expr)
		  end
	     | Index (expr1, expr2, astinfo) =>
		  let
		     val (ret_expr1, typ_expr1, errors_expr1) = curry_check_types_expr expr1
		     val (ret_expr2, typ_expr2, errors_expr2) = curry_check_types_expr expr2
		     val total_errors = errors_expr1 @ errors_expr2

		  (* aere we need to generate the type of the result expression.  Since a[x] and x[a] are equivalent in C, we
		   * need to acount for that in the case statement inside the let (note how ret_expr1 and ret_expr2 are switched
		   * in the two calls) *)
		     fun switcher typ_a typ_b =
			case typ_a of
			   SyntacticType (class, element_type, qualifiers, ((Array _) | (Pointer _)) :: rest) =>
			      let val typ = SyntacticType (class, element_type, qualifiers, rest)
			      in
				 (cb_expression (#index callbacks, sourceMap,
						 Index (ret_expr1, ret_expr2, set_type astinfo typ)),
				  typ, total_errors)
			      end
			 | _ => raise Unsupported (getExprRegion tree,
						   "unexpected error; this is a bug")
		  in
		     case (convert_to_general_types typ_expr1, convert_to_general_types typ_expr2) of
			(((PointerType _) | ArrayType), ArithmeticType {is_int=true, ...}) =>
			   switcher typ_expr1 typ_expr2
		      | (ArithmeticType {is_int=true, ...}, (ArrayType | (PointerType _))) =>
			   switcher typ_expr2 typ_expr1
		      | _ => (cb_expression (#index callbacks, sourceMap, Index (ret_expr1, ret_expr2, set_type astinfo typ_expr1)),
			      typ_expr1,
			      (getExprRegion tree, "expr_check: incorrect type for '[]' operator") :: total_errors)
		  end
	     | Extension (stat, astinfo) =>
		  let
		     val typ = SyntacticType (NONE, Int, [], []) (* make up a type until I can figure out what gcc does with this extension *)
		     val (ret_stat, errors_stat) = curry_check_types_stat stat
		  in
		     (cb_expression (#extension callbacks, sourceMap,
				     Extension (ret_stat, set_type astinfo typ)),
		      typ, errors_stat)
		  end
	 end
	       
      and check_types_operation lookup_param sourceMap cb (tree as (Operation (oper, expr_list, astinfo))) =
	 let val _ = 0
	 in
	    case expr_list of
	       [left, right] =>
		  let
		     val (ret_expr_left, typ_expr_left, errors_expr_left) = check_types_expression lookup_param sourceMap NONE cb left
		     val (ret_expr_right, typ_expr_right, errors_expr_right) = check_types_expression lookup_param sourceMap NONE cb right
		     val (typ, typ_errors) =
			(unify_op_types {is_assign=false} oper (ret_expr_left, typ_expr_left) (ret_expr_left, typ_expr_right), [])
			handle UnifyError (_, str) => (typ_expr_left, [(getExprRegion tree, "expr_check: " ^ str)])
		     val out_typ = Operation (oper, [ret_expr_left, ret_expr_right], set_type astinfo typ)
		     val total_errors = errors_expr_left @ errors_expr_right @ typ_errors
		  in
		     case oper of
			(Plus | Minus | Mul | Div | Mod | And | Or | BitAnd | BitOr | BitXor | Lshift | Rshift |
			 Lt | Le | Gt | Ge | Eq | Ne) => (out_typ, typ, total_errors)
		       | _ => (out_typ, typ,
			       (getExprRegion tree, "this should never happen; it's a bug in the type checker") :: total_errors)
		  end
	     | [expr] =>
		  let
		     val (ret_expr, typ_expr as SyntacticType (_, spec, _, _), errors_expr) =
			check_types_expression lookup_param sourceMap (SOME tree) cb expr
			
		     val out_expr_regular = Operation (oper, [ret_expr], set_type astinfo typ_expr)
		  in
		     case oper of
			(Incr | Decr) =>
			   (case (convert_to_general_types typ_expr) of
			       ((ArithmeticType {is_int=_, is_const=_}) | (PointerType {is_void=false})) => (out_expr_regular, typ_expr,
													     errors_expr)
			     | _ => (out_expr_regular, typ_expr,
				     [(getExprRegion tree,
				      "oper_check: incorrect type of the operand to a preincrement or predecrement operation; it must be either an arithmetic value or a pointer")]))
		      | Not => (* ! operator *)
			   let
			      val typ = SyntacticType (NONE, Int, [], [])
			      val expr = Operation (oper, [ret_expr], set_type astinfo typ)
			   in
			      case (convert_to_general_types typ_expr) of
				 ((ArithmeticType {is_int=_, is_const=_}) | (PointerType {is_void=false})) =>
				    (expr, typ, errors_expr)
			       | _ => (expr, typ, errors_expr @
				       (getExprRegion tree,
	     "oper_check: incorrect type of the operand to a logical negation; it must be either an arithmetic value or a pointer") ::
				       errors_expr)
			   end
		      | BitNot => (* ~ operator *)
			   let
			      val typ = promote_integer_type typ_expr
			      val expr = Operation (oper, [ret_expr], set_type astinfo typ)
			   in
			      case (convert_to_general_types typ_expr) of
				 (ArithmeticType {is_int=true, is_const=_}) =>
				    (expr, typ, errors_expr)
			       | _ => (expr, typ,
				       (getExprRegion tree,
		"oper_check: incorrect type of the operand to a bitwise negation; it must be either an arithmetic value or a pointer") ::
				       errors_expr)
			   end
		      | Address => (* & operator *)
			   let
			      val error_typ = SyntacticType (NONE, spec, [], [(Pointer [])]) (* might as well make it a pointer of some sort *)
			      val error_expr_out = Operation (oper, [ret_expr], set_type astinfo error_typ)
			   in
			      case typ_expr of
				 SyntacticType (val1, val2, val3, (BitField expr) :: rest) =>
				    (error_expr_out, error_typ,
				     (getExprRegion tree,
				      "oper_check: incorrect type of the operand to an address-of operation; it cannot be a bitfield")
				     :: errors_expr)
			       | SyntacticType (SOME Register, val2, val3, val4) =>
				    (error_expr_out, error_typ,
				     (getExprRegion tree,
				      "oper_check: incorrect type of the operand to an address-of operation; it cannot be a register variable")
				     :: errors_expr)
			       | SyntacticType (val1, val2, val3, val4) =>
				    let val typ = SyntacticType (val1, val2, val3, (Pointer []) :: val4)
				                  (* make a pointer to the original type *)
				    in
				       (Operation (oper, [ret_expr], set_type astinfo typ), typ, errors_expr)
				    end
			   end
		      | Deref => (* * operator *)
			   (case typ_expr of
			       SyntacticType (val1, val2, val3, (Pointer p) :: rest) =>
				  let val typ = SyntacticType (val1, val2, val3, rest)
				  in
				     (Operation (oper, [ret_expr], set_type astinfo typ), typ, errors_expr)
				  end
			     | _ =>
				  let val typ = SyntacticType (NONE, spec, [], [])
				  in
				     (Operation (oper, [ret_expr], set_type astinfo typ), typ,
				      (getExprRegion tree,
				       "oper_check: incorrect type of the operand to the dereference operator; it must be a pointer value")
				      :: errors_expr)
				  end)
		      | ( Plus | UnaryMinus ) => (* negation or unary plus *)
			   (case (convert_to_general_types typ_expr) of
			       (ArithmeticType {is_int=_, is_const=_}) =>
				  let val typ = promote_integer_type typ_expr
				  in
				     (Operation (oper, [ret_expr], set_type astinfo typ), typ, errors_expr)
				  end
			     | _ =>
				  let val typ = SyntacticType (NONE, Int, [], [])
				  in
				     (Operation (oper, [ret_expr], set_type astinfo typ), typ,
				      (getExprRegion tree,
		       "oper_check: incorrect type of the operand to the unary plus or unary minus operator; it must be an arithmetic value")
				      :: errors_expr)
				  end)
		      | _ => raise Unsupported (getExprRegion tree,
						"One argument should not be passed to this function...this is a bug in the compiler.")
		  end
	     | _ =>
		  raise Unsupported
		     (getExprRegion tree,
		      "Zero or more than two arguments should not be passed to this function...this is a bug in the compiler.")
	 end
     | check_types_operation lookup_param sourceMap cb (tree as (Postfix (oper, expr, astinfo))) =
	 let val _ = 0
	 in
	    case oper of
	       (Incr | Decr) =>
		  let
		     val (ret_expr, typ_expr, errors_expr) = check_types_expression lookup_param sourceMap (SOME tree) cb expr
		     val error_typ = SyntacticType (NONE, Int, [], [])
		  in
		     case (convert_to_general_types typ_expr) of
			((ArithmeticType _) | (PointerType {is_void=false})) =>
			   (Postfix (oper, ret_expr, set_type astinfo typ_expr), typ_expr, errors_expr)
		      | _ =>
			   (Postfix (oper, ret_expr, set_type astinfo error_typ), error_typ,
			    (getExprRegion tree,
	 "oper_check: incorrect type of the operand to a postfix increment or decrement; it must be either an arithmetic value or a pointer")
			    :: errors_expr)
		  end
	     | _ => raise Unsupported (getExprRegion tree,
				       "There aren't any other postfix expressions other than '++' and '--'")
	 end
	| check_types_operation _ _ _ tree = raise Unsupported (getExprRegion tree,
							     "Only call this function with operation or postfix expressions")

      and check_types_assignment lookup_param sourceMap cb (tree as (Assignment (expr_left, possible_oper, expr_right, astinfo))) =
         let
	    val (ret_expr_left, typ_expr_left, errors_expr_left) = check_types_expression lookup_param sourceMap (SOME tree) cb expr_left
	    val (ret_expr_right, typ_expr_right, errors_expr_right) = check_types_expression lookup_param sourceMap (SOME tree) cb expr_right

	    val total_errors = errors_expr_left @ errors_expr_right
	    val out_default = Assignment (ret_expr_left, possible_oper, ret_expr_right, set_type astinfo typ_expr_left)
	 in
	    (* if non exception has been thrown, then the left side expression should be ok *)
	    case (convert_to_general_types typ_expr_left) of
	       (ArrayType | FunctionType _) =>
		  (out_default, typ_expr_left,
		   (getExprRegion tree,
	    "assign_check: incorrect type of the expression on the left side of an assignment statement; it cannot be an array or function")
		   :: total_errors)
	     | ((ArithmeticType {is_int=_, is_const=true}) | CharacterType | StringType) =>
		  (out_default, typ_expr_left,
		   (getExprRegion tree,
	    "assign_check: incorrect type of the expression on the left side of an assignment statement; it cannot be a constant value")
		   :: total_errors)
	     | _ => 
		  let val errors =
		     (case (convert_to_general_types typ_expr_left) of
			 (StructureType | UnionType) => check_consts_in_structure_or_union_pred lookup_param typ_expr_left
		       | _ => [])
		      val total_errors = errors @ total_errors
		  in
		     case possible_oper of
			SOME oper =>
			   let
		     (* since a op= b is the same as a = a op b (except that a is evaluated only once, but this doesn't affect the types)
		      * we can call unify_op_types op a b to get the unified type and check it against the left-hand side  *)
			      val (typ, errors_typ) =
				 (unify_op_types {is_assign=true} oper (ret_expr_left, typ_expr_left) (ret_expr_left, typ_expr_right), [])
				 handle UnifyError (_, str) => (typ_expr_left, [(getExprRegion tree, "assign_check " ^ str)])
			   in
			      case oper of
				 (Plus | Minus | Mul | Div | Mod | Lshift | Rshift | BitAnd | BitOr | BitXor) =>
				    (out_default, typ_expr_left, errors_typ @ total_errors)
			       | _ => (out_default, typ_expr_left,
				       (getExprRegion tree, "This exception should not happen; it is a compiler error")
				       :: (errors_typ @ total_errors))
			   end
		      | NONE =>
			   let val (typ, typ_errors) =
			      (regular_assign_check lookup_param sourceMap cb {is_initialization = true}
			                            (ret_expr_left, typ_expr_left) (ret_expr_right, typ_expr_right),
			       [])
			      handle ParseError (_, str) => (typ_expr_left, [(getExprRegion tree, "assign_check: " ^ str)])
			   in
			      (Assignment (ret_expr_left, possible_oper, ret_expr_right, set_type astinfo typ), typ,
			       typ_errors @ total_errors)
			   end
		  end
	 end
     | check_types_assignment _ _ _ tree = raise Unsupported (getExprRegion tree, "Only call this function with an assignment expression")

      and regular_assign_check loookup_param sourceMap cb {is_initialization}
	                       (ret_expr_left, typ_expr_left) (ret_expr_right, typ_expr_right) =
	 case (convert_to_general_types typ_expr_left, convert_to_general_types typ_expr_right) of
	    ((StructureType, StructureType) | (UnionType, UnionType)) =>
	       if equal_types {check_pointer_type_qualifiers=true} typ_expr_left typ_expr_right then
		  typ_expr_left
	       else
		  raise ParseError
		     (getExprRegion ret_expr_left,
		      "incorrect types of the arguments to '='; if they are both structures or unions, they must have the same type")
	  | ((ArithmeticType {is_int=_, is_const=_}), (ArithmeticType {is_int=_, is_const=_})) =>
	       let
		  val typ = unify_types {is_ptr_minus=false} (ret_expr_left, typ_expr_left) (ret_expr_right, typ_expr_right)
		     handle UnifyError (_, str) => raise ParseError (getExprRegion ret_expr_left,
								     "incompatible arithmetic types in assignment")
	       in
		  typ_expr_left
	       end
	  | ((PointerType {is_void=_}), (PointerType {is_void=true})) => typ_expr_left
	  | ((PointerType {is_void=true}), (PointerType {is_void=_})) => typ_expr_left
	  | ((PointerType {is_void=false}), (ArithmeticType {is_int=true, is_const=true})) => typ_expr_left
	  | (PointerType _, PointerType _) =>
	       if equal_types {check_pointer_type_qualifiers=false} typ_expr_left typ_expr_right then
		  typ_expr_left
	       else
		  raise ParseError (getExprRegion ret_expr_left, "incompatible types in assignment: if they are both non-void pointers, they must have the same type")
	  (* assume the integer is 0 for now *)
	  | (ArrayType, PointerType _) =>
	       let val error = (getExprRegion ret_expr_left, "incompatible types in assignment: an array type can only be initialized with a { ... } style initialization, or if it is a character array, by an initialization string")
	       in
		  if is_initialization then
		     case (typ_expr_left, typ_expr_right) of
			(SyntacticType (_, Char, _, (Array _)::rest),
			 SyntacticType (NONE, Char, [Const], [(Pointer _)])) => (* sort of correct: it still may not be a string constant *)
		        typ_expr_left
		       | _ => raise ParseError error
		  else
		     raise ParseError error
	       end
	  | _ =>
	       (case (typ_expr_left, typ_expr_right) of
		   (SyntacticType (val1_left, val2_left, val3_left, (Function decls1) :: rest_left),
		    SyntacticType (val1_right, val2_right, val3_right, (Function decls2) :: rest_right)) =>
		   let
		      fun remove_const_and_volatile els =
			 foldr (fn (el, accum) =>
				case el of
				   Const => accum
				 | Volatile => accum
				 | _ => el :: accum) [] els
		   in
		      if val1_left = val1_right andalso
			 equal_type_specifiers val2_left val2_right andalso
			 (equal_type_qualifiers_list val3_left val3_right
			  orelse equal_type_qualifiers_list (remove_const_and_volatile val3_left)
			  (remove_const_and_volatile val3_right)) andalso
			 equal_declarations_list decls1 decls2 then
			 typ_expr_left
		      else
			 raise ParseError
			    (getExprRegion ret_expr_left,
			     "incorrect types of the arguments to '='; if they are both functions, they must have equivalent types except for the possible absence of 'const' or 'volatile' in the right argument")
		   end
		  | _ => raise ParseError (getExprRegion ret_expr_left,
					   "incorrect types of the arguments to '='; they must both be arithmetic, both pointers, the left be a pointer and the right an integral value, both the same structure type, or both equivalent function types"))


      and set_type {region, extraInfo, typeInfo} typ = {region = region, extraInfo = extraInfo, typeInfo = SOME typ}

      and set_expr_type expr typ =
	 case expr of
	    Variable (id, astinfo) => Variable(id, set_type astinfo typ)
	  | Constant (c, astinfo) => Constant (c, set_type astinfo typ)
	  | Conditional (e1, e2, e3, astinfo) => Conditional (e1, e2, e3, set_type astinfo typ)
	  | Application (e1, e2_l, astinfo, lab) => Application (e1, e2_l, set_type astinfo typ, lab)
	  | Operation (oper, e, astinfo) => Operation (oper, e, set_type astinfo typ)
	  | Assignment (e1, oper, e2, astinfo) => Assignment (e1, oper, e2, set_type astinfo typ)
	  | Comma (e1, e2, astinfo) => Comma (e1, e2, set_type astinfo typ)
	  | Cast (syntype, e, astinfo) => Cast (syntype, e, set_type astinfo typ)
	  | SizeOfExpr (e, astinfo) => SizeOfExpr (e, set_type astinfo typ)
	  | SizeOfTypeExpr (syntype, astinfo) => SizeOfTypeExpr (syntype, set_type astinfo typ)
	  | FieldDirect (e, id, astinfo) => FieldDirect (e, id, set_type astinfo typ)
	  | FieldIndirect (e, id, astinfo) => FieldIndirect (e, id, set_type astinfo typ)
	  | Postfix (oper, e, astinfo) => Postfix (oper, e, set_type astinfo typ)
	  | Index (e1, e2, astinfo) => Index (e1, e2, set_type astinfo typ)
	  | Extension (s, astinfo) => Extension (s, set_type astinfo typ)

      and convert_to_general_types typ =
	 case typ of
	    SyntacticType (_, _, _, (BitField _) :: _) => BitfieldType
	  | SyntacticType (_, (Char | UnsignedChar | ShortInt | UnsignedShortInt | Int | UnsignedInt | LongInt |
			       LongLong | UnsignedLongInt | UnsignedLongLong),
			   qual_list, []) =>
	    (if List.exists (fn a => case a of Const => true | _ => false) qual_list then
		ArithmeticType {is_int=true, is_const=true}
	     else
		ArithmeticType {is_int=true, is_const=false})
	  | SyntacticType (_, (Float | Double | LongDouble), qual_list, []) =>
	      (if List.exists (fn a => case a of Const => true | _ => false) qual_list then
		  ArithmeticType {is_int=false, is_const=true}
	       else
		  ArithmeticType {is_int=false, is_const=false})
	  | SyntacticType (_, (Aggregate (Union, _, _) | NamedAggregate (Union, _)), _, []) => UnionType
	  | SyntacticType (_, (Aggregate (Struct, _, _) | NamedAggregate (Struct, _)), _, []) => StructureType
	  | SyntacticType (_, Void, _, (Pointer _) :: rest) => PointerType {is_void=true}
	  | SyntacticType (_, _, _, (Pointer _) :: rest) => PointerType {is_void=false}
	  | SyntacticType (_, Void, _, ((Function _) | OldFunction _) :: rest) => FunctionType {is_void = true}
	  | SyntacticType (_, _, _, ((Function _) | OldFunction _) :: rest) => FunctionType {is_void = false}
	  | SyntacticType (_, _, _, (Array _) :: rest) => ArrayType
	  | _ => NoGeneralType

      and equal_types {check_pointer_type_qualifiers} (SyntacticType (a_1, a_2, a_3, a_4)) (SyntacticType (b_1, b_2, b_3, b_4)) =
	 equal_type_specifiers a_2 b_2 andalso
	 equal_type_qualifiers_list a_3 b_3 andalso
	 equal_declarators_list {check_pointer_type_qualifiers=check_pointer_type_qualifiers} a_4 b_4
	 
      and equal_identifiers id1 id2 = (Lexeme.idToString id1) = (Lexeme.idToString id2)

      and equal_declarations decl1 decl2 =
	 case (decl1, decl2) of
	    (EllipsisDecl _, EllipsisDecl _) => true
	  | (VarDecl {name=name1, typ=typ1, ...}, VarDecl {name=name2,typ= typ2, ...}) =>
	       (equal_identifiers name1 name2) andalso (equal_types {check_pointer_type_qualifiers=true} typ1 typ2)
	  | (TypeDef {name=name1, typ=typ1, ...}, TypeDef {name=name2, typ=typ2, ...}) =>
	       (equal_identifiers name1 name2) andalso (equal_types {check_pointer_type_qualifiers=true} typ1 typ2)
	  | (FunDef {name=name1, typ=typ1, ...}, FunDef {name=name2, typ=typ2, ...}) =>
	       (equal_identifiers name1 name2) andalso (equal_types {check_pointer_type_qualifiers=true} typ1 typ2)
	  | (FunDecl {name=name1, typ=typ1, ...}, FunDecl {name=name2, typ=typ2, ...}) =>
	       (equal_identifiers name1 name2) andalso (equal_types {check_pointer_type_qualifiers=true} typ1 typ2)
	  | _ => false

      and equal_declarators {check_pointer_type_qualifiers} decl1 decl2 =
	 case (decl1, decl2) of
	    (Pointer tq1, Pointer tq2) =>
	       if check_pointer_type_qualifiers then
		  equal_type_qualifiers_list tq1 tq2
	       else
		  true
	  | (Array (SOME expr1), Array (SOME expr2)) => equal_constants expr1 expr2
	  | (Array NONE, Array NONE) => true
	  | (Array _, Array _) => false
	  | (BitField expr1, BitField expr2) => equal_constants expr1 expr2
	  | (Function declaration1_list, Function declaration2_list) => equal_declarations_list declaration1_list declaration2_list
	  | (OldFunction (list_a, _), OldFunction (list_b, _)) =>
	       ListPair.foldl (fn ((_, id1), (_, id2), accum) => accum andalso equal_identifiers id1 id2)
	                      true (list_a, list_b)
	  | _ => false

      and equal_declarators_list options decl1_list decl2_list =
	 (* the lists must be in the same order, but that's required for the type declarations to be equal anyway *)
	 ListPair.foldl (fn (a,b,accum) => accum andalso equal_declarators options a b) true (decl1_list, decl2_list)
	       
      and equal_declarations_list decl1_list decl2_list =
	 let
	    fun get_name (EllipsisDecl _) = ""
	      | get_name (VarDecl {name, ...}) = Lexeme.idToString name
	      | get_name (TypeDef {name, ...}) = Lexeme.idToString name
	      | get_name (FunDef {name, ...}) = Lexeme.idToString name
	      | get_name (FunDecl {name, ...}) = Lexeme.idToString name
		   
	    fun sort_fun (a,b) =
	       let
		  val name1 = get_name a
		  val name2 = get_name b
	       in
		  String.> (name1, name2)
	       end
	 in
	     ListPair.foldl (fn (a, b, accum) => accum andalso (equal_declarations a b))
	                    true (ListMergeSort.sort sort_fun decl1_list, ListMergeSort.sort sort_fun decl2_list)
	 end

      and equal_constants e1 e2 =
	 case (e1, e2) of
	    (Constant (c1, _), Constant (c2, _)) =>
	       (case (c1, c2) of
		   (Integer i1, Integer i2) => i1 = i2
		 | (BigInteger id1, BigInteger id2) => equal_identifiers id1 id2
		 | (String id1, String id2) => equal_identifiers id1 id2
		 | (Character id1, Character id2) => equal_identifiers id1 id2
		 | (Real id1, Real id2) => equal_identifiers id1 id2
		 | _ => false)
	  | _ => false

      and equal_enum_els els1 els2 =
	 let
	    fun gt ((id1, _), (id2, _)) = String.> (Lexeme.idToString id1, Lexeme.idToString id2)
	    val sorted_els1 = ListMergeSort.sort gt els1
	    val sorted_els2 = ListMergeSort.sort gt els2

	    fun check_func ((id1, expr1), (id2, expr2), accum) =
	       accum andalso
	       (Lexeme.idToString id1) = (Lexeme.idToString id2) andalso
	       (case (expr1, expr2) of
		   (SOME e1, SOME e2) => equal_constants e1 e2
		 | (NONE, NONE) => true
		 | _ => false)
	 in
	    ListPair.foldl check_func true (sorted_els1, sorted_els2)
	 end
	     
      and equal_type_specifiers spec1 spec2 =
	 case (spec1, spec2) of
	    (Void, Void) => true
	  | (Char, Char) => true
	  | (UnsignedChar, UnsignedChar) => true
	  | (ShortInt, ShortInt) => true
	  | (UnsignedShortInt, UnsignedShortInt) => true
	  | (Int, Int) => true
	  | (UnsignedInt, UnsignedInt) => true
	  | (LongInt, LongInt) => true
	  | (UnsignedLongIng, UnsignedLongInt) => true
	  | (LongLong, LongLong) => true
	  | (UnsignedLongLong, UnsignedLongLong) => true
	  | (Float, Float) => true
	  | (Double, Double) => true
	  | (LongDouble, LongDouble) => true
	  | (NamedType v1, NamedType v2) => equal_identifiers v1 v2
	  | (Enum ((SOME v1), _), Enum ((SOME v2), _)) => equal_identifiers v1 v2 (* if the names are the same, they must be the same type...
										   * or at least we are going to assume so *)
	  | (Enum (NONE, els1), Enum (NONE, els2)) => equal_enum_els els1 els2
	  | (Aggregate (Struct, SOME id1, _), Aggregate (Struct, SOME id2, _)) =>
	       equal_identifiers id1 id2 (* if the names are the same...(see comment above) *)
	  | (Aggregate (Union, SOME id1, _), Aggregate (Union, SOME id2, _)) =>
	       equal_identifiers id1 id2   (* if the names are the same...(see comment above) *)
	  | (Aggregate (Struct, NONE, decls1), Aggregate (Struct, NONE, decls2)) => equal_declarations_list decls1 decls2
	  | (Aggregate (Union, NONE, decls1), Aggregate (Union, NONE, decls2)) => equal_declarations_list decls1 decls2
	  | (NamedAggregate (Struct, id1), NamedAggregate (Struct, id2)) => equal_identifiers id1 id2
	  | (NamedAggregate (Union, id1), NamedAggregate (Union, id2)) => equal_identifiers id1 id2
	  | _ => false

      and equal_type_qualifiers q1 q2 =
	 case (q1, q2) of
	    (Const, Const) => true
	  | (Volatile, Volatile) => true
	  | (Inline, Inline) => true
	  | (NonConst, NonConst) => true
	  | (Qualifier id1, Qualifier id2) => equal_identifiers id1 id2
	  | _ => false

      and equal_type_qualifiers_list spec_list1 spec_list2 =
	 let
	    fun hash a =
	       case a of
		  Const => 0 | Volatile => 1 | Inline => 2 | NonConst => 3
		| Qualifier _ => raise Unsupported ((~1, ~1), "Don't call this with a qualifier!")

	    (* impose an arbitrary ordering of : Const < Volatile < Inline < NonConst < Qualifier _, with Qualifiers being differentiated
	     * based on the lexicographical value in the identifer *)
	    fun gt (a,b) =
	       case (a,b) of
		  (Qualifier id1, Qualifier id2) => String.> ((Lexeme.idToString id1), (Lexeme.idToString id2))
		| (Qualifier _, _) => true
		| (_, Qualifier _) => false
		| _ => (hash a) > (hash b)
	 in
	    ListPair.foldl (fn (a, b, accum) => accum andalso (equal_type_qualifiers a b))
	                   true (ListMergeSort.sort gt spec_list1, ListMergeSort.sort gt spec_list2)
	 end

      and find_type_of_element_by_name decl_list id =
	 let
	    val id_str = Lexeme.idToString id
	    fun pred decl =
	       case decl of
		  EllipsisDecl _ => false
		| VarDecl {name, ...} => (Lexeme.idToString name) = id_str
		| TypeDef {name, ...} => (Lexeme.idToString name) = id_str
		| FunDef {name, ...} => (Lexeme.idToString name) = id_str
		| FunDecl {name, ...} => (Lexeme.idToString name) = id_str

	    val decl = List.find pred decl_list
	 in
	    case decl of
	       SOME (EllipsisDecl _) => raise ParseError ((~1, ~1), "No type found")
	     | SOME (VarDecl {typ, ...}) => typ
	     | SOME (TypeDef {typ, ...}) => typ
	     | SOME (FunDef {typ, ...}) => typ
	     | SOME (FunDecl {typ, ...}) => typ
	     | NONE => raise ParseError ((~1, ~1), "No type found")
	 end

      and check_consts_in_structure_or_union_pred lookup_param typ = raise Unsupported ((~1, ~1), "implement this")

      and merge_type_qualifiers quals1 quals2 =
	 let
	    fun hash a =
	       case a of
		  Const => 0 | Volatile => 1 | Inline => 2 | NonConst => 3
		| Qualifier _ => raise Unsupported ((~1, ~1), "Don't call this with a qualifier!")

	    (* impose an arbitrary ordering of : Const < Volatile < Inline < NonConst < Qualifier _, with Qualifiers being differentiated
	     * based on the lexicographical value in the identifer *)
	    fun gt (a,b) =
	       case (a,b) of
		  (Qualifier id1, Qualifier id2) => String.> ((Lexeme.idToString id1), (Lexeme.idToString id2))
		| (Qualifier _, _) => true
		| (_, Qualifier _) => false
		| _ => (hash a) > (hash b)

	    fun merge_quals_sorted (quals1 as (el1 :: rest1)) (quals2 as (el2 :: rest2)) =
	       if gt (el1, el2) then
		  el2 :: (merge_quals_sorted quals1 rest2)
	       else if gt (el2, el1) then
		  el1 :: (merge_quals_sorted rest1 quals2)
	       else
		  el1 :: (merge_quals_sorted rest1 rest2)
	      | merge_quals_sorted [] quals2 =
		  quals2
	      | merge_quals_sorted quals1 [] =
		  quals1
	 in
	    merge_quals_sorted (ListMergeSort.sort gt quals1) (ListMergeSort.sort gt quals2)
	 end

      and evaluates_to_numeric expr v = raise Unsupported ((~1, ~1), "implement this...")

      (* for now for both of these; the real implementation should come later once system-dependent information is added *)
      and bigger_data_size typ1 typ2 = false
      and bigger_or_equal_data_size typ1 typ2 = true

      and promote_integer_type (typ as (SyntacticType (val1, spec as (Char | UnsignedChar | ShortInt | UnsignedShortInt | (Enum _)),
						       val3, []))) =
	 if bigger_data_size Int spec then
	    SyntacticType (val1, Int, val3, [])
	 else
	    SyntacticType (val1, UnsignedInt, val3, [])
	| promote_integer_type (typ as (SyntacticType (val1, spec as (Int | UnsignedInt), val3, (BitField expr) :: rest))) =
	    if bigger_data_size Int spec then
	       SyntacticType (val1, Int, val3, [])
	    else
	       SyntacticType (val1, UnsignedInt, val3, [])
	| promote_integer_type typ = typ

      and unify_arithmetic_types (typ1 as (SyntacticType (_, spec1, _, _))) (typ2 as (SyntacticType (_, spec2, _, _))) =
	 case (convert_to_general_types typ1, convert_to_general_types typ2) of
	    (ArithmeticType _, ArithmeticType _) =>
	       let
		  val new_spec =
		     case (spec1, spec2) of
			(LongDouble, _) => LongDouble
		      | (_, LongDouble) => LongDouble
		      | (Double, _) => Double
		      | (_, Double) => Double
		      | (Float, _) => Float
		      | (_, Float) => Float
		      | (a, b) => (* some sort of integral type *)
			   let
			      val promoted_a = promote_integer_type typ1
			      val promoted_b = promote_integer_type typ2
			   in
			      case (promoted_a, promoted_b) of
				 (SyntacticType (_, spec1, _, _), SyntacticType (_, spec2, _, _)) =>
				    (case (spec1, spec2) of
					(UnsignedLongInt, _) => UnsignedLongInt
				      | (_, UnsignedLongInt) => UnsignedLongInt
				      | ((LongInt, UnsignedInt) | (UnsignedInt, LongInt)) =>
					   if bigger_data_size LongInt UnsignedInt then
					      LongInt
					   else
					      UnsignedLongInt
				      | (LongInt, _) => LongInt
				      | (_, LongInt) => LongInt
				      | (UnsignedInt, _) => UnsignedInt
				      | (_, UnsignedInt) => UnsignedInt
				      | _ => Int)
			   end
	       in
		  SyntacticType (NONE, new_spec, [], [])
	       end
	  | _ => raise Unsupported ((~1, ~1), "Don't call this function with non-arithmetic types")
	 
      and unify_types {is_ptr_minus} (expr1, typ1) (expr2, typ2) =
	 let
	    fun get_pointer_qualifiers (SyntacticType (_, _, _, (Pointer quals) :: rest)) = quals
	      | get_pointer_qualifiers _ = raise Unsupported ((~1, ~1), "Don't call this with a non-pointer")
	       
	    fun set_pointer_qualifiers (SyntacticType (a, b, c, (Pointer quals_pre) :: rest)) quals_post =
	       SyntacticType (a, b, c, (Pointer quals_post) :: rest)
	      | set_pointer_qualifiers _ _ = raise Unsupported ((~1, ~1), "Don't call this with a non-pointer")
	 in
	    case (convert_to_general_types typ1, convert_to_general_types typ2) of
	       (PointerType {is_void=is_void1}, PointerType {is_void=is_void2}) =>
		  if is_ptr_minus then  (* if the expressions are the operands to the binary '-' operator, then they must be the
					 * same type even if they are void pointers *)
		     if (equal_types {check_pointer_type_qualifiers=false} typ1 typ2) then
			set_pointer_qualifiers typ1 (merge_type_qualifiers (get_pointer_qualifiers typ1) (get_pointer_qualifiers typ2))
		     else
			raise UnifyError ((~1, ~1), "can't unify pointers of different types here")
		  else
		  if is_void1 then
		     set_pointer_qualifiers typ1 (merge_type_qualifiers (get_pointer_qualifiers typ1) (get_pointer_qualifiers typ2))
		  else if is_void2 then
		     set_pointer_qualifiers typ2 (merge_type_qualifiers (get_pointer_qualifiers typ1) (get_pointer_qualifiers typ2))
		  else if equal_types {check_pointer_type_qualifiers=false} typ1 typ2 then
		     set_pointer_qualifiers typ1 (merge_type_qualifiers (get_pointer_qualifiers typ1) (get_pointer_qualifiers typ2))
		  else
		     raise UnifyError ((~1, ~1), "can't unify pointers of different non-void types here")
	     | (StructureType, StructureType) =>
		  if equal_types {check_pointer_type_qualifiers=true} typ1 typ2 then typ1
		  else raise UnifyError ((~1, ~1), "can't unify structures of different types")
	     | (UnionType, UnionType) =>
		  if equal_types {check_pointer_type_qualifiers=true} typ1 typ2 then typ1
		  else raise UnifyError ((~1, ~1), "can't unify unions of different types")
	     | (PointerType _, ArithmeticType {is_int=true, is_const=true}) =>
	          if evaluates_to_numeric expr2 0 then typ1
		  else raise UnifyError ((~1, ~1), "can't unify pointer types with constants with value other than zero")
	     | (ArithmeticType {is_int=true, is_const=true}, PointerType _) =>
	          if evaluates_to_numeric expr1 0 then typ2
		  else raise UnifyError ((~1, ~1), "can't unify pointer types with constants with value other than zero")
	     | (ArithmeticType _, ArithmeticType _) =>
		  unify_arithmetic_types typ1 typ2
	     | _ => raise UnifyError ((~1, ~1), "Can't unify these types")
	 end
		  

      and unify_op_types {is_assign} oper (expr1, typ1) (expr2,typ2) =
	 case oper of
	    (Mul | Div ) =>
	       (case (convert_to_general_types typ1, convert_to_general_types typ2) of
		   (ArithmeticType _, ArithmeticType _) => unify_arithmetic_types typ1 typ2
		 | _ => raise UnifyError ((~1, ~1),
					  "incorrect types of one or more of the operands to '*' or '/'; they must both be arithmetic types"))
	  | Mod =>
	       (case (convert_to_general_types typ1, convert_to_general_types typ2) of
		   (ArithmeticType {is_int=true, ...}, ArithmeticType {is_int=true, ...}) =>
		      unify_arithmetic_types typ1 typ2
		 | _ => raise UnifyError ((~1, ~1),
					  "incorrect types of one or more of the operands to '%'; they must both be integral types"))
	  | (Plus | Minus) =>
	       (case (convert_to_general_types typ1, convert_to_general_types typ2) of
		   (ArithmeticType _, ArithmeticType _) => unify_arithmetic_types typ1 typ2
		 | (PointerType _, ArithmeticType {is_int=true, ...}) => typ1
		 | (ArithmeticType {is_int=true, ...}, PointerType _) =>
		      if is_assign then
			 raise UnifyError
	    ((~1, ~1),
	     "incorrect types in assignment; the left hand side must be a pointer and the right hand side an integral value, not the reverse")
		      else 
			 typ2
		 | (PointerType _, PointerType _) =>
		      if oper = Plus then
			 raise UnifyError ((~1, ~1), "incorrect types of the operands to '+'; they cannot both be pointers")
		      else (unify_types {is_ptr_minus=true} (expr1, typ1) (expr2,typ2) (* just to make sure they're compatible pointer values *)
			    handle UnifyError _ =>
			       raise ParseError (getExprRegion expr1,
				 "incorrect types of the operands to '-'; if they are both pointers, they must point to the same type");
			    SyntacticType (NONE, Int, [Const], []))
		 | _ => raise UnifyError
			 ((~1, ~1), "incorrect types of one or more of the operands to binary '+' or binary '-'; they must both be arithmetic, one a pointer and one an integral value, or two pointers if the operator is binary '-'"))
	  | (Rshift | Lshift) =>
	       (case (convert_to_general_types typ1, convert_to_general_types typ2) of
		   (ArithmeticType {is_int=true, ...}, ArithmeticType {is_int=true, ...}) =>
		      promote_integer_type typ1
		 | _ => raise UnifyError ((~1, ~1),
					  "incorrect types of one or more of the operands to '<<' or '>>'; they must both be integral"))
	  | (Lt | Le | Gt | Ge | Eq | Ne) =>
	       let val out_typ = SyntacticType (NONE, Int, [Const], [])
	       in
		  case (convert_to_general_types typ1, convert_to_general_types typ2) of
		     (ArithmeticType _, ArithmeticType _) => out_typ
		   | (PointerType {is_void = is_void1}, PointerType {is_void = is_void2}) =>
			if equal_types {check_pointer_type_qualifiers=false} typ1 typ2 then
			   out_typ
			else if (is_void1 orelse is_void2) andalso (oper = Eq orelse oper = Ne) then
			   out_typ
			else
			   raise UnifyError
		      ((~1, ~1),
		       "incorrect types of the operands to '<', '<=', '>', '>=', '==' or '!='; if they are both pointers, they must have the same type")
		   | (PointerType _, ArithmeticType {is_int=true, is_const=true}) =>
		      (* assume the rhs is zero for now *)
		      if oper = Eq orelse oper = Ne then
			 out_typ
		      else
			 raise UnifyError ((~1, ~1),
					   "incorrect types of the operands to '<', '<=', '>' or '>='; both objects must be pointers (also a pointer cannot be compared with NULL except with the == or != operators)")
		   | _ => raise UnifyError
	      ((~1, ~1),
       "incorrect types of the operands to '<', '<=', '>' or '>='; they must either be both arithmetic types or pointers of the same type")
	       end
	  | (BitAnd | BitOr | BitXor) =>
	       (case (convert_to_general_types typ1, convert_to_general_types typ2) of
		   (ArithmeticType {is_int=true, ...}, ArithmeticType {is_int=true, ...}) =>
		      unify_arithmetic_types typ1 typ2
		 | _ => raise UnifyError ((~1, ~1), "incorrect types of the operands to '&', '|' or '^'; they must both be integral"))
	  | (And | Or) =>
	       (case (convert_to_general_types typ1, convert_to_general_types typ2) of
		   ( ((ArithmeticType _) | (PointerType _)),
		    ((ArithmeticType _) | (PointerType _)) ) =>  SyntacticType (NONE, Int, [Const], [])
		  | _ => raise UnifyError
		   ((~1, ~1), "incorrect types of the operands to '&&' or '||'; they must pointers or arithmetic values (in any combination)"))
	  | _ => raise Unsupported ((~1, ~1), "unexpected argument to unify_op_types; this is a bug in the parser")

      and cb_program (SOME func, sourceMap, tree) =
	 (case func (sourceMap, tree) of
	     SOME new_tree => new_tree
	   | NONE => tree)
	| cb_program (NONE, _, tree) = tree
      and cb_declaration (SOME func, sourceMap, tree) =
	 (case func (sourceMap, tree) of
	     SOME new_tree => new_tree
	   | NONE => tree)
	| cb_declaration (NONE, _, tree) = tree
      and cb_expression (SOME func, sourceMap, tree) =
	 (case func (sourceMap, tree) of
	     SOME new_tree => new_tree
	   | NONE => tree)
	| cb_expression (NONE, _, tree) = tree
      and cb_statement (SOME func, sourceMap, tree) =
	 (case func (sourceMap, tree) of
	     SOME new_tree => new_tree
	   | NONE => tree)
	| cb_statement (NONE, _, tree) = tree


      and check_argument_types_for_new_functions promoted_arg_types param_decls =
	 let
	    (* the parameters can only be ellipses or variable declarations, since functions and types can't be defined/declared in
	     * a function definition *)
	    fun iterate (cur_arg::rest_args) [EllipsisDecl _] = ()
	      | iterate [] [EllipsisDecl _] = ()
	      | iterate (cur_arg::rest_args) ((VarDecl {typ, ...}) :: rest_params) =
	       if equal_types {check_pointer_type_qualifiers=true} cur_arg typ then
		  iterate rest_args rest_params
	       else raise ParseError ((~1, ~1), "incorrect argument type")
	      | iterate [] [] = ()
	      | iterate (cur_arg::rest_args) [] = raise ParseError ((~1, ~1), "too many arguments provided in function call")
	      | iterate [] (cur_param::rest_params) = raise ParseError ((~1, ~1), "not enough arguments provided in function call")
	      | iterate _ _ = raise Unsupported ((~1, ~1), "unexpected types passed to function call checker; this is a bug in the parser")
	 in
	    iterate promoted_arg_types param_decls
	 end

   end
