(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

signature FLAGS =
    sig

	type flags

	val new : string -> flags
	val add : string * bool ref * flags -> unit
	val set : string * flags -> unit
	val clr : string * flags -> unit
	val get : string * flags -> bool
	val addFlag : flags -> {name : string,
				flag : bool ref,
				trueDesc : string,
				falseDesc : string} -> unit
	val newFlag : flags -> {name : string,
				default : bool,
				trueDesc : string,
				falseDesc : string} -> bool ref
	val defaults : flags -> unit
	val names : flags -> string list
	val options : flags -> (string * bool) list
	val printFlags : TextIO.outstream -> flags -> unit
    end

structure Flags : FLAGS=
    struct
	
	structure Map = BinaryMapFn(struct
					type ord_key = string
					val compare = String.compare
				    end)

	type flagInfo = {flag : bool ref,
			 default : bool,
			 desc : {trueDesc : string,
				 falseDesc : string} option}

	datatype flags = Flags of {name : string,
				   map : flagInfo Map.map ref}

	fun new name = Flags {name=name, map=ref (Map.empty)}

	fun add (name, flag, Flags {map,...}) =
	    map := Map.insert (!map, name, {flag=flag,
					    default= !flag,
					    desc=NONE})

	fun addFlag (Flags {map, ...}) {name, flag, trueDesc, falseDesc} =
	    map := Map.insert (!map, name, {flag=flag,
					    default= !flag,
					    desc=SOME {trueDesc=trueDesc,
						       falseDesc=falseDesc}})

	fun newFlag flags {name,trueDesc,falseDesc,default} =
	    let val flag = ref default
	    in
		addFlag flags {name=name,flag=flag,trueDesc=trueDesc,
			       falseDesc=falseDesc};
		flag
	    end

	fun set (flagName, Flags {map, name}) = 
	    case Map.find (!map, flagName) of
		SOME {flag, ...} => flag := true
	      | _ => print(concat ["WARNING: unknown flag ",
				   flagName,
				   " in ",
				   name,
				   "\n"])

	fun clr (flagName, Flags {name, map}) = 
	    case Map.find (!map, flagName) of
		SOME {flag, ...} => flag := false
	      | _ => print(concat ["WARNING: unknown flag ",
				   flagName,
				   " in ",
				   name,
				   "\n"])

	fun get (flagName, Flags {name, map}) = 
	    case Map.find (!map, flagName) of
		SOME {flag, ...} => !flag
	      | _ => (print(concat ["WARNING: unknown flag ",
				    flagName,
				    " in ",
				    name,
				    "\n"]);
		      raise Match)

	fun options (Flags {map,...}) =
	    List.map (fn (name, {flag,...}) => (name,!flag)) (Map.listItemsi (!map))

	fun names (Flags {map,...}) =
	    List.map #1 (Map.listItemsi (!map))

	fun defaults (Flags {map,...}) =
	    let fun reset (name, {flag,default,...}:flagInfo) =
		    flag := default
	    in
		Map.appi reset (!map)
	    end

	local
	    fun NStrs s 0 = ""
	      | NStrs s n = (s ^ (NStrs s (n - 1)))
	    fun padTo s n =
		s ^ (NStrs " " (Int.max (n - (size s), 0)))
	in
	    fun printFlags os (Flags {map, name}) =
		 let val maxName = Map.foldli
		                       (fn (s, _, n) => Int.max(size s, n))
				       0
				       (!map)
		     fun printFlag (name, {flag, default, desc}) =
			 TextIO.output (os,
					(concat (padTo name (maxName+1) ::
						 " = " ::
						 (if (!flag) then "true "
						  else "false") ::
						 (case desc of
						      SOME {trueDesc,
							    falseDesc} =>
						      ["  (",
						       if !flag then trueDesc
						       else falseDesc,
						       ")\n"]
						    | NONE => ["\n"]))))
		 in
		     TextIO.output (os, (concat [name,
						 ":\n"]));
		     Map.appi printFlag (!map)
		 end
	end
    end
