structure MemUsage:
    sig
	type report_func = int * int * int -> string

	type time = Time.time
	val second: time
	val tenthSecond: time
	val quarterSecond: time

	(* memUsage (file,report,timeOut) func -
         * Spawn a child process that monitors the memory usage
	 * of func. All data are printed to the file; each line
	 * is printed using report. The monitor waits, roughly,
	 * timeOut between each report.
         *)
	val memUsage: (string * report_func option * time) -> (unit -> unit) -> unit

	(* sendString string -
         * Send the string to the monitor process. This process
         * will then add it to the ouput.
         *)
	val sendString: string -> unit

        (* wasteSomeTime () -
         * Sample function to test the monitor process.
         *)
	val wasteSomeTime: unit -> unit
    end =
    struct

	fun pidName () =
	    let val pid = Posix.ProcEnv.getpid ()
	    in
		SysWord.fmt StringCvt.DEC (Posix.Process.pidToWord pid)
	    end

	type report_func = int*int*int -> string

	type time = Time.time
	val second = Time.fromMilliseconds 1000
	val longTime = Time.fromMilliseconds 10000
	val tenthSecond = Time.fromMilliseconds 100 
	val quarterSecond = Time.fromMilliseconds 250

	fun getString socket timeOut stream =
	    let val socketPoll = OS.IO.pollIn(Socket.pollDesc socket)
		fun socketReady infoDesc =
		    let open OS.IO 
		    in
			if compare(pollToIODesc socketPoll, pollToIODesc (infoToPollDesc infoDesc)) = EQUAL
			then let val (_,string) = Sockets.recv_from socket
			     in  TextIO.output(stream, string ^ "\n")
			     end
			else ()
		    end
	    in
		 case OS.IO.poll ([socketPoll], SOME timeOut) of
		     [] => ()
		 |   infoList => List.app socketReady infoList
	    end

	fun showOutput NONE (count, size, rss) =
	      concat [Int.toString count, " ", 
		      "Size= ", Int.toString size, ", ",
		      "RSS= ", Int.toString rss]
          | showOutput (SOME report) (count, size, rss) = report(count,size,rss)

	local val count = ref 0
	in
	    fun monitorMemoryUsage showInput stream report pid =
		let val {size, rss} = Info.getInfo (pid)
		in
		    TextIO.output(stream, (showOutput report (!count, size, rss)) ^ "\n")
		    before count := 1 + !count;
		    TextIO.flushOut stream;
		    showInput stream; (* Will make the thing sleep until input
				       * arrives or until the timeOut has elapsed.
				       *)
		    monitorMemoryUsage showInput stream report pid
		end
	end (* local *)

	fun addHandler (signal, handler) =
	    let open Signals
		val oldHandler = inqHandler signal
		fun signalHandler (signal, times, next) =
		    (handler ();
		     case oldHandler of
			 HANDLER oldHandler =>  
			     (* Process already provided handler; remove our handler
			      * and call the old one. *)
			     (setHandler (sigINT, HANDLER oldHandler);
			      oldHandler(signal, times, next)
			      )
		     |   action =>
			     (* Process did *not* provide handler; just remove our
			      * handler and proceed. *)
			     (setHandler (Signals.sigINT, action);
			      next)
		   )
	    in
		setHandler (signal, HANDLER signalHandler)
	    end

	local val client_socket = ref NONE
	in

	fun sendString string = 
	   (case !client_socket of
		NONE => raise Fail("Client socket never bound!")
            |   SOME sock => Sockets.send_to (sock, string)
           )

	fun memUsage (file,report,timeOut) func =
	    let val myPid = pidName ()
		fun killChild pid =
		    (print "parent: killing child\n";
		     Posix.Process.kill (Posix.Process.K_PROC pid, Posix.Signal.term);
		     Posix.Process.wait ();
		     print "parent: done killing\n")
		val (server_sock,server_addr) = Sockets.server_init (*6060*)6062 (* just some random port number *)
		val server_port = Sockets.get_port (server_sock, server_addr)
	    in
		print ("Forking child (parent = " ^ myPid ^ ")\n");
		case Posix.Process.fork () of
		    SOME pid => (* We're in the parent process *)
			let val _ = Posix.Process.sleep second
			 (* val _ = Posix.Process.sleep second *)
			    val client_sock = Sockets.client_init' (server_port)
			    val _ = client_socket := SOME client_sock
			    fun cleanUp () =
				(print "parent: Closing client socket\n"; 
				 Sockets.client_close client_sock;
				 killChild pid)
			in
			    Socket.close server_sock;
			    addHandler (Signals.sigINT, cleanUp);
			    func ();
			    cleanUp()
			end
		  | NONE => (* We're in the child process *)
			let val os = TextIO.openOut file
			    val (client_sock, client_port) = Sockets.server_connect server_sock
			    fun cleanUp () =
				(print "child: Closing sockets\n"; 
				 Sockets.server_close (server_sock, client_sock);
				 print "child: Closing output stream\n"; 
				 TextIO.closeOut os)
			in 
			    addHandler (Signals.sigTERM, cleanUp);
			    monitorMemoryUsage (getString client_sock timeOut) os report myPid
			end
	    end

	end (* local *)


	fun wasteSomeTime () =
	    let fun loop n =
		(if Int.mod(n, 1000)=0 then sendString ("Loop " ^ Int.toString n)
		 else ();
		 Posix.Process.sleep second;
		 loop (n+1))
	    in
		loop 0
	    end

    end

