(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

(* 
 *  Derived from code distributed with the SML/NJ compiler.
 *
 *  Added interface function "location_string"
 *
 *  Manuel Fahndrich
 *)
 
(* I can imagine at least three implementations: one that doesn't           *)
(* support resynchronization, one that supports resynchronization only at   *)
(* column 1, and one that supports arbitrary resynchronization.             *)
(*                                                                          *)
(*                                                                          *)
(* \section{Implementation}                                                 *)
(* This implementation supports arbitary resynchronization.                 *)
(*                                                                          *)
(* <sourcemap.sml>=                                                         *)
(* sourcemap.sml *)
(* <RCS log>=                                                               *)
(*
 * changed ErrorMsg to use SourceMap to get source locations; only the
 * formatting is done internally
 *
 * added SourceMap structure
 *
 * .sig and .sml for sourcemap, source, and errormsg are derived from .nw
 * files.  to extract, try
 *   for base in sourcemap source errormsg
 *   do
 *     for suffix in sml sig
 *     do
 *       $cmd -L'(*#line %L "%F"*)' -R$base.$suffix $base.nw > $base.$suffix
 *     done
 *   done
 * where
 *   cmd=notangle
 * or
 *   cmd="nountangle -ml"
 *
 * At some point, it may be desirable to move noweb support into CM
 * *)
structure SourceMap : SOURCE_MAP = struct
  (* A character position is an integer.  A region is delimited by the        *)
  (* position of the start character and one beyond the end.                  *)
  (* It might help to think of Icon-style positions, which fall between       *)
  (* characters.                                                              *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  type charpos = int
  type 'a pair = 'a * 'a
  type region = charpos pair
  val nullRegion : region = (0,0)
  type sourceloc = {fileName:string, line:int, column:int}
  (* The empty region is conventional.                                        *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  fun span ((0,0), r) = r
    | span (r, (0,0)) = r
    | span ((l1, h1), (l2, h2)) = if l1 < h2 then (l1, h2) else (l2, h1)
  (* The representation is a pair of lists.                                   *)
  (* [[linePos]] records line numbers for newlines \emph{and}                 *)
  (* resynchronization.                                                       *)
  (* [[resynchPos]] records file name and column for resynchronization.       *)
  (* The representation satisfies these invariants:                           *)
  (* \begin{itemize}                                                          *)
  (* \item                                                                    *)
  (* The lists are never empty (initialization is treated as a resynchronization). *)
  (* \item                                                                    *)
  (* Positions decrease as we walk down the lists.                            *)
  (* \item                                                                    *)
  (* The last element in each list contains the smallest valid position.      *)
  (* \item                                                                    *)
  (* For every element in [[resynchPos]], there is a corresponding element in *)
  (* [[linePos]] with the same position.                                      *)
  (* \end{itemize}                                                            *)
  (* We could get even more clever and store file names only when they        *)
  (* differ, but it doesn't seem worth it---we would have to get very         *)
  (* clever about tracking column numbers and resynchronizations.             *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  type sourcemap = {resynchPos: (charpos * string * int) list ref,
                    linePos:    (charpos * int)          list ref}

  fun newmap (pos, {fileName, line, column}: sourceloc) : sourcemap =
    {resynchPos = ref [(pos, fileName, column)], linePos = ref [(pos, line)]}

  fun resynch ({resynchPos, linePos}: sourcemap) (pos, {fileName, line, column}) =
    let val curFile = #2 (hd (!resynchPos))
        fun thefile (SOME file) = if file = curFile then curFile else file
                                     (* pathetic attempt at hash-consing *)
          | thefile NONE        = #2 (hd (!resynchPos))
        fun thecol NONE     = 1
          | thecol (SOME c) = c
    in  resynchPos := (pos, thefile fileName, thecol column) :: !resynchPos;
        linePos := (pos, line) :: !linePos
    end
  (* Since [[pos]] is the position of the newline, the next line doesn't      *)
  (* start until the succeeding position.                                     *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  fun newline ({resynchPos, linePos}: sourcemap) pos =
    let val (_, line) = hd (!linePos)
    in  linePos := (pos+1, line+1) :: !linePos
    end

  fun lastChange({linePos, ...}: sourcemap) = #1 (hd (!linePos))
  (* A generally useful thing to do is to remove from the lists the initial   *)
  (* sequences of tuples                                                      *)
  (* whose positions satisfy some predicate:                                  *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  fun remove p ({resynchPos,linePos}: sourcemap) =
    let fun strip (l as (pos, _   )::rest) = if p pos then strip  rest else l
          | strip [] = []
        fun strip'(l as (pos, _, _)::rest) = if p pos then strip' rest else l
          | strip'[] = []
    in  (strip'(!resynchPos), strip (!linePos))
    end
  (* We find file and line number by linear search.                           *)
  (* The first position less than [[p]] is what we want.                      *)
  (* The initial column depends on whether we resynchronized.                 *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  fun column ((pos, file, col), (pos', line), p) =
    if pos = pos' then p - pos + col else p - pos' + 1

  fun filepos smap p : sourceloc =
      let val (files, lines) = remove (fn pos : int => pos > p) smap
          val xx as (_, file, _) = hd files
          val yy as (_, line)    = hd lines
      in  {fileName = file, line = line, column = column(xx, yy, p)}
      end
  (* Searching regions is a bit trickier, since we track file and line        *)
  (* simultaneously.  We exploit the invariant that every file entry has a    *)
  (* corresponding line entry.                                                *)
  (* We also exploit that only file entries correspond to new regions.        *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  fun fileregion smap (lo, hi) =
    if (lo,hi) = nullRegion then [] else
    let exception Impossible
        fun gather((p, file, col)::files, (p', line)::lines, region_end, answers) =
          if p' <= lo then (* last item *)
            ({fileName=file, line=line, column=column((p, file, col), (p', line), lo)}, 
             region_end) :: answers
          else
            if p < p' then
              gather((p, file, col)::files, lines, region_end, answers)
            else (* p = p'; new region *)
              gather(files, lines, end_of (p, hd files, hd lines), 
                   ({fileName = file, line = line, column = col}, region_end) :: answers)
          | gather _ = raise Impossible
        and end_of(lastpos, xx as (p, file, col), yy as (p', line)) = 
               {fileName=file, line=line, column=column(xx, yy, lastpos)}
        val (files, lines) = remove (fn pos : int => pos >= hi andalso pos > lo) smap
        val _ = if null files orelse null lines then raise Impossible else ()
        val answer = gather(files, lines, end_of(hi, hd files, hd lines), [])
        fun validate(({fileName=f,  line=l,  column=c}, 
                      {fileName=f', line=l', column=c'}) :: rest) = 
              if f = f' andalso (l' > l orelse (l' = l andalso c' >= c)) then
                validate rest 
              else 
                raise Impossible
          | validate [] = ()
    in  validate answer; answer
    end
  (* [[validate]] checks the invariant that single regions occupy a           *)
  (* single source file and that coordinates are nondecreasing.               *)
  (* We have to be careful not to remove the entry for [[lo]] when            *)
  (* [[pos = hi = lo]].                                                       *)
  (*                                                                          *)
  (*                                                                          *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  fun positions({resynchPos,linePos}: sourcemap) (src:sourceloc) =
    let exception Unimplemented
    in  raise Unimplemented
    end
  (* When discarding old positions, we have to be careful to maintain the     *)
  (* last part of the invariant.                                              *)
  (*                                                                          *)
  (* <toplevel>=                                                              *)
  fun forgetOldPositions ({resynchPos, linePos} : sourcemap) =
    let val r as (p,  file, col) = hd (!resynchPos)
        val l as (p', line)      = hd (!linePos)
    in  linePos := [l];
        resynchPos := [if p = p' then r else (p', file, 1)]
    end
  (* <toplevel>=                                                              *)
  fun newlineCount smap (lo, hi) =
    let val (hifiles, hilines) = remove (fn pos : int => pos >= hi andalso pos > lo) smap
        val (lofiles, lolines) = remove (fn pos : int =>                   pos > lo) smap
    in  length hilines - length hifiles - (length lolines - length lofiles)
    end

    fun location_string sourceMap (p1,p2) =
      let fun shortpoint ({line, column,...}:sourceloc, l) = 
             Int.toString line :: "." :: Int.toString column :: l
          fun showpoint (p as {fileName,...}:sourceloc, l) = 
             (*OS.Path.file *) fileName :: ":" :: shortpoint (p, l)
          fun allfiles(f, (src:sourceloc, _)::l) = 
                f = #fileName src andalso allfiles(f, l)
            | allfiles(f, []) = true
          fun lastpos [(_, hi)] = hi
            | lastpos (h::t) = lastpos t
            | lastpos [] = (print "lastpos botch in location_string\n";
			    raise (Fail "location_string"))
      in  concat (
            case fileregion sourceMap (p1, p2)
              of [(lo, hi)] => 
                    if p1+1 >= p2 then showpoint (lo, [])
                    else showpoint(lo, "-" :: shortpoint(hi, []))
               | (lo, _) :: rest =>
                    if allfiles(#fileName lo, rest) then
                      showpoint(lo, "..." :: shortpoint(lastpos rest, []))
                    else
                      showpoint(lo, "..." :: showpoint (lastpos rest, []))
               | [] => ["<somefile>", ":<nullRegion>"]
          )
      end

end

(*
 * $Log: sourcemap.sml,v $
 * Revision 1.1  2000/10/10 05:51:37  daw
 * *** empty log message ***
 *
 * Revision 1.4  1998/06/12 20:28:19  jfoster
 * Added copyright
 *
 * Revision 1.3  1998/06/04 06:18:11  manuel
 * Added comment to sourcemap
 * Added options function to flags.sml
 *
 * Revision 1.2  1998-04-23 09:40:38-07  manuel
 * - Prints full path name
 *
 * Revision 1.1.1.1  1998-03-25 09:16:29-08  manuel
 * Directory containing support ML modules that don't depend on anything else.
 *
 * Revision 1.2  1998-01-12 13:49:53-08  manuel
 * Added location_string to sourcemap
 *
 * Revision 1.1.1.1  1997-12-09 09:51:19-08  manuel
 * Program analysis support tools.
 *
 * Revision 1.1.1.1  1997/01/14  01:38:50  george
 *   Version 109.24
 *
 *)
