(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)


structure Util : 
    sig
	(* return the new value *)
	val inc : int ref -> int
	val dec : int ref -> int

	(* Profiling and Timing Support *)
	val measure  : string -> (unit -> 'a) -> 'a
	val time_it  : string -> (unit -> 'a) -> 'a
	val time_usr : string -> (unit -> 'a) -> 'a
	val overTime : Time.time -> string -> (unit -> 'a) -> bool
	val overTime2: Time.time -> string -> (unit -> 'a) -> 'a

	type profiler
	val newProf   : string -> profiler
	val doProf    : profiler -> (unit -> 'a) -> 'a
	val showProf  : profiler -> unit
	val resetProf : profiler -> unit
	val timeProf  : profiler -> {gc:Time.time,
				     usr:Time.time,
				     sys:Time.time}

	(* Formatting Support *)
	val fraction      : {N:int,D:int} -> real
	val realString2   : real -> string
	val percent       : {Pct:int,Of:int} -> real
	val percentString : real -> string

	(* Setup Support *)
	exception NotImplemented
	val notImp     : string -> 'a -> 'b
	val warnNotImp : string -> 'a -> unit

	(* Join lists *)
	datatype 'a jlist =
	    Nil 
	  | Single of 'a
	  | Join of 'a jlist * 'a jlist
	  | List of 'a list

	val join     : 'a jlist * 'a jlist -> 'a jlist
	val jcons    : 'a * 'a jlist -> 'a jlist
	val jlist    : 'a list -> 'a jlist
	val jmap     : ('a -> 'b) -> 'a jlist -> 'b jlist
	val jflatmap : ('a -> 'b) -> 'a jlist -> 'b list
	val japp     : ('a -> unit) -> 'a jlist -> unit
	val jlength  : 'a jlist -> int
	val jconcat  : 'a jlist list -> 'a jlist
	val jfind    : ('a -> bool) -> 'a jlist -> 'a option
	val jexists  : ('a -> bool) -> 'a jlist -> bool
	val jnull    : 'a jlist -> bool
	val jhd      : 'a jlist -> 'a
	val jmapPartial
	             : ('a -> 'b option) -> 'a jlist -> 'b jlist
	val jflatmapPartial
                     : ('a -> 'b option) -> 'a jlist -> 'b list
	val jflatten : 'a jlist -> 'a list
	val jfoldl   : ('a * 'b -> 'b) -> 'b -> 'a jlist -> 'b

	(* Pointer equality and association lists *)
	val ptreql : 'a * 'a -> bool
	val member : 'a * 'a list -> bool
	val assoc  : 'a * ('a * 'b) list -> 'b option

	(* closure operation over list references.
	 * 
	 *   <upto> must be a shared sublist of the list ref. The
	 *   function close is called on all elements on the list ref
	 *   up to the sublist <upto>. Calls to close may add more elements
	 *   onto the list ref, resulting in more calls to close. The process
	 *   ends when all elements on the final list ref (up to <upto>)
	 *   have been passed to close.
	 *   
	 *   <upto> may be nil.
	 *
	 *)
	val close : 'a list ref -> {upto : 'a list, close : 'a -> unit} -> unit

	(* Escape characters for string between double quotes *)
	val escape_chars : string -> string

        (* Exception History *)
	val showHistory : exn -> unit

    end =
    struct

	fun percent {Pct,Of} =
	    (Real.fromInt Pct * 100.00) / Real.fromInt Of

	fun fraction {N,D} =
	    (Real.fromInt N) / Real.fromInt D

	val cvt = StringCvt.FIX (SOME 2)
	val percentString = Real.fmt cvt
	val realString2   = percentString

	fun escape_chars s = 
	    String.translate (fn #"\"" => "\\\""
	                       | #"\\" => "\\\\"
			       | #"\t" => "\\t"
			       | #"\n" => "\\n"
			       | c => String.str c) s


	fun inc r = let val n = !r + 1 in r:=n; n end
	fun dec r = let val n = !r - 1 in r:=n; n end

	fun measure s f =
	    let open Timer Time
		val tm = startCPUTimer ()
		fun compute_time () =
		    let val {gc,sys,usr} = checkCPUTimer tm
			val total = usr
		    in
			print (s ^ " " ^ toString total ^
			       "s (usr="^toString usr^"s gc="^toString gc^"s)\n")
		    end
	    in		
		(f () before compute_time ()) 
		handle exn => (compute_time (); raise exn)
	    end

	fun time_it s f =
	    let open Timer Time
		val tm = startCPUTimer ()
		val _ = print ("[Time " ^ s ^ " ")
		val _ = TextIO.flushOut TextIO.stdOut
		val r = f ()
		val {gc,sys,usr} = checkCPUTimer tm
		val total = sys + usr
	    in
		print (toString total ^"]\n");
		r
	    end

	fun time_usr s f =
	    let open Timer Time
		val tm = startCPUTimer ()
		val _ = print ("[Time " ^ s ^ " ")
		val _ = TextIO.flushOut TextIO.stdOut
		val r = f ()
		val {gc,sys,usr} = checkCPUTimer tm
		val total = usr
	    in
		print (toString total ^"]\n");
		r
	    end


	fun overTime threshold s f =
	    let open Timer Time
		val tm = startCPUTimer ()
		val r = f ()
		val {gc,sys,usr} = checkCPUTimer tm
		val total = sys+usr
		val over = total > threshold
	    in
		if over then
		    print ("[" ^ s ^ " "^toString total ^"]\n")
		else ();
		over
	    end

	fun overTime2 threshold s f =
	    let open Timer Time
		val tm = startCPUTimer ()
		val r = f ()
		val {gc,sys,usr} = checkCPUTimer tm
		val total = sys+usr
	    in
		if total > threshold then
		    print ("[" ^ s ^ " "^toString total ^"]\n")
		else ();
		r
	    end

	local
	    open Time Timer
	in
	    type timerec = {gc : time, usr : time, sys : time}

	    type profiler = timerec ref * string

	    val zero_time = zeroTime

	    fun resetProf (r,_) = (r := {gc=zero_time,
					  sys=zero_time,
					  usr=zero_time})

	    fun timeProf (ref t,_) = t

	    fun emptyProfiler () = ref {gc=zero_time,
					sys=zero_time,
					usr=zero_time}

	    fun newProf name = 
		let val p = (emptyProfiler (),
			     name)
		in
		    p
		end


	    fun add_time ({gc=gc1,usr=usr1,sys=sys1},
			  {gc=gc2,usr=usr2,sys=sys2}) =
		{gc=gc1+gc2,usr=usr1+usr2,sys=sys1+sys2}

	    fun sub_time ({gc=gc1,usr=usr1,sys=sys1},
			  {gc=gc2,usr=usr2,sys=sys2}) =
		{gc=gc1-gc2,usr=usr1-usr2,sys=sys1-sys2}


	    (* The profStack is used to attribute execution time to a single
	     * profiler only. We substract the profile time of all nested prof calls
	     * If execution time is to be attributed to multiple profilers, then the
	     * profStack should be made first class and passed to doProf.
	     *)
	    val profStack = ref ([] : timerec ref list)

	    fun doProf (r as ref old,_) f =
		let val nestProf = emptyProfiler ()
		    val _ = profStack := nestProf :: !profStack
		    val t = startCPUTimer ()
		    val a = f ()
		    val newNested = checkCPUTimer t
		    val new = sub_time (newNested, !nestProf) (* sub out nested prof time *)
		in
		    profStack := List.tl (!profStack);
		    case !profStack of
			[] => ()		(* bottom of stack *)

		      | (next::_) =>		(* add our nest time to next *)
			    next := add_time (!next, newNested)
                    ;
		    r := add_time (old, new);	(* accumulate time *)
		    a
		end

	    fun showProf (ref {gc,sys,usr}, name) =
		let val total = sys + usr
		    val ts = Time.toString total
		    val us = Time.toString usr
		    val ss = Time.toString sys
		    val gs = Time.toString gc
		in
		    print ("["^name^" (usr) "^us^" (sys) "^ss^
			   " (gc) "^gs^" (total) "^ts^"]\n")
		end

	end				(* local *)


	exception NotImplemented

	fun notImp s _ = (print ("Not Implemented: "^s^"\n");

			  raise NotImplemented)

	fun warnNotImp s _ = (print ("WARNING: Not Implemented: "^s^"\n"))

	datatype 'a jlist =
	    Nil 
	  | Single of 'a
	  | Join of 'a jlist * 'a jlist
	  | List of 'a list

	fun join (Nil,b) = b
	  | join (a,Nil) = a
	  | join (Single e, List l) = List (e::l)
	  | join (Single e1, Single e2) = List [e1,e2]
	  | join (a, List []) = a
	  | join (List [], b) = b
	  | join (a,b) = Join (a,b)

	fun jcons (e,Nil) = Single e
	  | jcons (e,Single e') = List [e,e']
	  | jcons (e,List l) = List (e::l)
	  | jcons (e, l as Join _) = Join (Single e, l)

	fun jlist [] = Nil
	  | jlist l  = List l

	fun jmap f Nil = Nil
	  | jmap f (List l) = List (List.map f l)
	  | jmap f (Single t) = Single (f t)
	  | jmap f (Join (a,b)) = Join (jmap f a, jmap f b)

	fun japp f Nil = ()
	  | japp f (List l) = List.app f l
	  | japp f (Single t) = f t
	  | japp f (Join (a,b)) = (japp f a; japp f b)

	fun jflatmap f Nil = []
	  | jflatmap f (List l) = (List.map f l)
	  | jflatmap f (Single t) = [(f t)]
	  | jflatmap f (Join (a,b)) = (op @) (jflatmap f a, jflatmap f b)

	fun jlength Nil = 0
	  | jlength (List l) = List.length l
	  | jlength (Single _) = 1
	  | jlength (Join (a,b)) = jlength a + jlength b

	fun jconcat [] = Nil
	  | jconcat (j::rest) = join (j, jconcat rest)

        exception NotFound

	fun find f =
	    let fun fi [] = raise NotFound
		  | fi (a::b) =
		    if f a then a else fi b
	    in
		fi
	    end

	fun jfind f jl =
	    let fun jf Nil = raise NotFound
		  | jf (List l) = find f l
		  | jf (Single a) = if f a then a else raise NotFound
		  | jf (Join (a,b)) = 
		      jf a handle NotFound => jf b
	    in
		SOME (jf jl) handle NotFound => NONE
	    end

	fun jexists f Nil = false
	  | jexists f (List l) = List.exists f l
	  | jexists f (Single a) = f a
	  | jexists f (Join (a,b)) = (jexists f a) orelse (jexists f b)

	fun jnull Nil = true
	  | jnull (List l) = List.null l
	  | jnull (Single _) = false
	  | jnull (Join (a,b)) = (jnull a) andalso (jnull b)

	fun jhd Nil = raise Empty
	  | jhd (List l) = hd l
	  | jhd (Single a) = a
	  | jhd (Join (a,b)) = jhd a handle Empty => jhd b

	fun jmapPartial _ Nil = Nil
	  | jmapPartial f (List l) = List (List.mapPartial f l)
	  | jmapPartial f (Single a) = (case f a of
					    SOME b => Single b
					  | NONE => Nil)
	  | jmapPartial f (Join (a,b)) = Join (jmapPartial f a,
					       jmapPartial f b)

	fun jflatmapPartial _ Nil = []
	  | jflatmapPartial f (List l) = List.mapPartial f l
	  | jflatmapPartial f (Single a) = (case f a of
						SOME b => [b]
					      | NONE => [])
	  | jflatmapPartial f (Join (a,b)) = (jflatmapPartial f a) @
	                                     (jflatmapPartial f b)

	fun jflatten Nil = []
	  | jflatten (List l) = l
	  | jflatten (Single a) = [a]
	  | jflatten (Join (a,b)) = ((jflatten a) @ (jflatten b))

	fun jfoldl _ z Nil = z
	  | jfoldl f z (List l) = foldl f z l
	  | jfoldl f z (Single a) = f (a, z)
	  | jfoldl f z (Join (a,b)) = jfoldl f (jfoldl f z a) b

	fun ptreql (a,b) = (Unsafe.cast a : Word.word) =
	                   (Unsafe.cast b : Word.word)

	fun member (i,[]) = false
	  | member (i,(a::rest)) = ptreql (i,a) orelse member (i,rest)

	fun assoc (i,[]) = NONE
	  | assoc (i,(j,e)::rest) = 
	    if ptreql (i,j) then SOME e else assoc (i,rest)


	fun close (listref:'a list ref) {upto:'a list, close=f:'a -> unit} =
	    let fun iterate (from, to) =
		let fun apply l =
		    if ptreql (l, to) then
			let val current_list = !listref
			in
			    if ptreql (current_list, from) then ()
			    else
				iterate (current_list, from)
			end
		    else
			let val next::rest = l
			in
			    f next;
			    apply rest
			end
		in
		    apply from
		end
	    in
		iterate (!listref, upto)
	    end


	(* Exception histories *)

	fun showhist' [s] = print(concat["  raised at: ", s, "\n"])
	  | showhist' (s::r) = (showhist' r; 
			       print (concat["             ", s, "\n"]))
	  | showhist' [] = ()

	fun showHistory exn =
	    let val loclist = SMLofNJ.exnHistory exn
	    in
		showhist' loclist
	    end




    end


