(* There's a serious architectural flaw with the way we convert
   set constraints into range constraints, or generate set constraints,
   or something.  For instance, consider the set constraint
   S - S >= 0, where S is a set of integers.  This gets converted
   into the the range l..h, and then S - S is represented as the
   range l-h..h-l, which already represents a loss of accuracy.
   Then if we consider the linear model for S - S >= 0, we see
   that it is l-h >= 0, i.e. l >= h, which is clearly inaccurate
   except when l=h (in which case the idea of value-range arithmetic
   collapses into constant propagation).  I'm not sure how to fix this.
   I think maybe one fix is to consider subsets of Z x Z, not just
   subsets of Z; a Int() term is the latter, whereas a Str() term is
   the former. *)
      

structure Constraint : CONSTRAINT =
  struct
    exception Err

    type label = int
    val curLabel = ref 0
    fun getFreshLabel () = let val l = !curLabel in curLabel := l + 1; l end
    val curConstraints = ref ""
    val names = ref(IntBinaryMap.empty) : string IntBinaryMap.map ref

    datatype var = Var of int (* int = unique label *)

    (* T([{mul=5,var=LoVar(0)},{mul=1,var=HiVar(3)}],-1)
       means 5*L0 + H3 - 1. *)
    datatype term = T of {coeff:int, var:var} list * int
    datatype endpoint = Int of int | Infinity | NegativeInfinity
    type range = endpoint * endpoint
    type query = term * term * (range * range * string -> unit)
    val curQueries = ref [] : query list ref

    fun myIntStr i = if i<0 then "-"^(myIntStr (~i)) else Int.toString i

    fun varToStr (Var(i) )= "S"^(myIntStr i)
    fun cvlistToStr [] = ""
      | cvlistToStr ({coeff,var}::xs) = 
            (if coeff=1 then "" else (myIntStr coeff)^"*")
            ^ (varToStr var)
            ^ (if xs=[] then "" else " + "^(cvlistToStr xs))
    fun endptToStr (Int i) = myIntStr i
      | endptToStr Infinity = "+Infinity"
      | endptToStr NegativeInfinity = "-Infinity"
    fun rangeToStr (x,y) = (endptToStr x)^".."^(endptToStr y)
    fun toString (T(vlist,const)) = 
          let val vs = cvlistToStr vlist
              val cs = myIntStr const
          in
              if vs="" then cs
              else if const=0 then vs
              else vs ^ " + " ^ cs
          end

    fun constrain s =
          (curConstraints := ((!curConstraints)^s^"\n"); ()
           (* if String.isPrefix "query " s then () else
           print ("Debug: new constraint: "^s^"\n") *))

    fun constantTerm i = T([],i)
    fun freshTerm nmopt =
          let val lbl = getFreshLabel ()
          in
              case nmopt of
                  SOME n => names := IntBinaryMap.insert(!names, lbl, n)
                | NONE   => ();
              T([{coeff=1,var=Var(lbl)}], 0)
          end
    fun getFreshInt () = (print "Err: getFreshInt\n"; raise Err)
    val globaltop = ref NONE : term option ref
    fun top () =
          case !globaltop of
              SOME t => t
            | NONE =>
          let val t = freshTerm (SOME "top")
          in
              (* print ("Debug: new top "^(toString t)^"\n"); *)
              constrain ("top subset "^(toString t));
              globaltop := SOME(t);
              t
          end

    fun reset () = (curLabel := 0; curConstraints := ""; curQueries := [];
                    globaltop := NONE; names := IntBinaryMap.empty)

    fun makeSingleton _ = (print "Err: makeSingleton\n"; raise Err)

    (* Naive O(n^2) algorithm; better is to use sorted lists.
       Actually, this is unnecessary, since setsolver already does
       this; but hey, it makes for prettier debugging output. *)
    fun addLists (xl,[]) = xl
      | addLists (xl,y::ys) =
          let fun addVar [] v = [v]
                | addVar ((z as {coeff=zc,var=zv})::zs)
                         (v as {coeff=vc,var=vv}) =
                    if zv = vv andalso
                       ( (zc>=0 andalso vc>=0) orelse (zc<=0 andalso vc<=0) )
                    then {coeff=zc+vc,var=vv}::zs
                    else z::(addVar zs v)
          in  addLists(addVar xl y,ys)
          end
    fun addTerms (T(xl,xc),T(yl,yc)) = T(addLists(xl,yl),xc+yc)
    fun negate (T(l,c)) =
          let fun negateOne {coeff=cc,var=vv} = {coeff=(~cc),var=vv}
          in  T(map negateOne l, ~c)
          end
    fun subTerms (x,y) = addTerms (x,negate y)
    fun constMulTerm (c,T(ylist,yconst)) =
          let fun mulOne {coeff=cc,var=vv} = {coeff=(c*cc),var=vv}
          in  T(map mulOne ylist, c*yconst)
          end handle overflow => top ()
          

    (* Only implemented when one of the terms is a constant range. *)
    fun mulTerms (T([],c),y) = constMulTerm (c,y)
      | mulTerms (x,T([],c)) = constMulTerm (c,x)
      | mulTerms (_,_) = top () (* FIX *)

    fun op / _ = top () (* FIX *)
    fun intersect _ = (print "Err: intersect\n"; raise Err)
    fun eqTerms (x,T([],yc)) =
          constrain ((toString x)^" = "^(myIntStr yc))
      | eqTerms (T(xv,xc),T(yv,yc)) =
          constrain ((toString (T(xv,xc-yc)))^" = "^(toString (T(yv,0))))
    (* XXX: is `S subset T+U' always equiv. to `S-U subset T'??? *)
    fun subset (_,T([],yc)) = (print "Err: subset\n"; raise Err)
      | subset (x,T(yv::yvs,yc)) =
          let val lhs = subTerms (x,T(yvs,yc))
              val l = (toString lhs)^" subset "^(toString (T([yv],0)))
              val _ = if yvs=[] then () else print ("Warning!: changing `"^(toString x)^" subset "^(toString (T(yv::yvs,yc)))^"' to `"^l^"'.\n")
          in  constrain l
          end
    fun termGe (x,y) = constrain ((toString x)^" >= "^(toString y))
    fun queryge (x as T(xl,xc), y as T(yl,yc), errorfn) =
          let fun qu {var,coeff} = constrain ("query "^(varToStr var))
          in
              curQueries := (x,y,errorfn)::(!curQueries);
              map qu xl; map qu yl;
              ()
          end

    fun union (x,y) =
          let val z = freshTerm NONE
          in
              subset (x,z);
              subset (y,z);
              z
          end
    fun min (x,y) =
          let val z = freshTerm NONE
              val x' = freshTerm NONE
              val y' = freshTerm NONE
          in
              (*
              print ("Debug: "^(toString z)^" := min {"^(toString x)
                     ^", "^(toString y)^"}\n");
              *)
              subset (x,x');
              subset (y,y');
              constrain ("min "^(toString x')^", "^(toString y')
                         ^" subset "^(toString z));
              z
          end

    val execFailed = "Couldn't exec ssolver; check your PATH?\n"
    fun externalSolver system =
          let val solver = "ssolver"
              val p = Unix.execute(solver, [])
                      handle e => (print execFailed;
                                   OS.Process.exit OS.Process.failure)
              val (rd,wr) = Unix.streamsOf p
          in
              TextIO.output (wr,system);
              TextIO.closeOut wr;
              let val s = (TextIO.inputAll rd)
              in  Unix.reap p; s
              end
          end;
    (* setsolver's response for infeasible linear programming models *)
    val infeasible = "This problem is infeasible\n"
    val feasible = "Feasible.\n"

    (* Interpret setsolver's output, and turn it into an integer
       if there is a bound on the value of the objective function,
       or NONE otherwise. *)
    fun getBounds output =
          let val (_,suf) = Substring.position feasible (Substring.all output)
              fun toEndpoint "+Infinity" = Infinity
                | toEndpoint "-Infinity" = NegativeInfinity
                | toEndpoint s = (case (Int.fromString s) of
                                      SOME i => Int(i)
                                    | NONE   => (print "Parse\n"; raise Err))
              fun toVar s =
                    let val varLbl =
                              case Substring.getc s of
                                  SOME(#"S",l) => l
                                  | _ => (print "Err: toVar\n"; raise Err)
                        val intOpt = Int.fromString (Substring.string varLbl)
                    in
                        case intOpt of
                            SOME lbl => lbl
                          | NONE => (print "Err: toVar\n"; raise Err)
                    end
              fun toRange s =
                    let val (lhs,x) = Substring.position ".." s
                        val rhs = Substring.slice (x, 2, NONE)
                               handle _ => (print "Err: toRange\n"; raise Err)
                    in  (toEndpoint (Substring.string lhs),
                         toEndpoint (Substring.string rhs))
                    end
              fun toDepends s =
                    if (Substring.size s) <= 1 then [] else
                    let val s' = Substring.slice (s, 1, NONE) (* skip space *)
                               handle _ => (print "Err: toDepends\n"; raise Err)
                        val (varName,rhs) = Substring.position " " s'
                        val lbl = toVar varName
                    in  lbl::(toDepends rhs)
                    end
              fun parseIt s =
                    if (Substring.size s) = 0
                    then (IntBinaryMap.empty : range IntBinaryMap.map,
                          IntBinaryMap.empty : int list IntBinaryMap.map) else
                    let val (t,u) = Substring.position "\n" s
                        val (varName,t') = Substring.position " = " t
                        val lbl = toVar varName
                        val (t'',d) = Substring.position " <-" t'
                        val range = toRange (Substring.slice (t'', 3, NONE))
                               handle _ => (print "Err: parseIt1\n"; raise Err)
                        val (vmap,dmap) =
                              parseIt (Substring.slice (u, 1, NONE))
                               handle _ => (print "Err: parseIt2\n"; raise Err)
                        val vmap' = IntBinaryMap.insert(vmap, lbl, range)

                    in
                        (vmap',
                         if (Substring.size d) = 0 then dmap
                         else IntBinaryMap.insert(dmap, lbl,
                           toDepends (Substring.slice (d, 3, NONE)))
                               handle _ => (print "Err: parseIt3\n"; raise Err)
                        )
                    end
          in
              if (Substring.size suf) < (size feasible)
              then (print "Error: unexpected response:\n";
                    print output; raise Err)
              else (parseIt (Substring.slice (suf, size(feasible), NONE))
                           handle _ => (print "Err: parseIt4\n"; raise Err))
          end

      local 
        (* d = dependencies map,
           s = set of already-seen labels (to prevent cycles),
           i = label to expand *)
        fun get (d,s,i) =
              if IntBinarySet.member(s,i) then IntBinarySet.empty else
              case IntBinaryMap.find(d,i) of
                   NONE => IntBinarySet.empty
                 | SOME l =>
                     let val s' = IntBinarySet.add(s,i)
                         fun f x = case IntBinaryMap.find(!names,x) of
                                       NONE   => get (d,s',x)
                                     | SOME _ => IntBinarySet.singleton(x)
                         fun g (x,y) = IntBinarySet.union(y, f x)
                     in
                         foldl g IntBinarySet.empty l
                     end
        fun fix (i,d) =
              let (* val _ = print ("Debug: fix S"^(myIntStr i)^"... ") *)
                  val l = IntBinarySet.listItems(get (d,IntBinarySet.empty,i))
                  (*
                  fun f y = " S"^(myIntStr y)
                  val _ = print (" get := "^(concat (map f l))^"\n")
                  *)
              in  IntBinaryMap.insert(d, i, l)
              end
      in
      fun fixDepends d = foldl fix d (map #1 (IntBinaryMap.listItemsi d))
      end

      fun endptAdd ((Infinity,NegativeInfinity)|(NegativeInfinity,Infinity)) =
              (print "Over-under flow.\n"; raise Err)
        | endptAdd ((Infinity,_)|(_,Infinity)) = Infinity
        | endptAdd ((NegativeInfinity,_)|(_,NegativeInfinity)) =
              NegativeInfinity
        | endptAdd (Int i, Int j) = Int(i+j)
      (* Note: endptMul is only used with non-negative values of i *)
      fun endptMul (i, Int j) = (Int(i*j)
           handle overflow => if i=0 then (print "0 * +/-Inf !?\n"; raise Err)
             else if i<0 then (print "Impossible: negative i!?\n"; raise Err)
             else if j>0 then Infinity else if j<0 then NegativeInfinity
             else (print "Impossible: j=0, but got overflow!?\n"; raise Err))
        | endptMul (i, inf as (Infinity|NegativeInfinity)) =
            if i>0 then inf else (print "0 * +/-Inf !?\n"; raise Err)
      fun endptNeg (Int i) = Int(~i)
        | endptNeg Infinity = NegativeInfinity
        | endptNeg NegativeInfinity = Infinity
      fun rangeAdd (   ((Infinity,NegativeInfinity),_)
                     | (_,(Infinity,NegativeInfinity)) ) =
            (Infinity,NegativeInfinity)
        | rangeAdd ((xl,xh):range,(yl,yh):range) =
            (endptAdd (xl,yl), endptAdd (xh,yh))
      fun rangeMul (cf,(xl,xh):range) =
            if cf<0 then rangeMul(~cf,(endptNeg xh, endptNeg xl))
            else (endptMul (cf,xl), endptMul (cf,xh))
      fun endptGe (Infinity,_) = true
        | endptGe (NegativeInfinity,NegativeInfinity) = true
        | endptGe (NegativeInfinity,(Infinity|Int(_))) = false
        | endptGe (Int x,Infinity) = false
        | endptGe (Int x,NegativeInfinity) = true
        | endptGe (Int x,Int y) = x >= y
    fun nspaces i = if i>0 then " "^(nspaces (i-1)) else if i=0 then ""
                    else (print "Err: nspaces\n"; raise Err)
    fun allspaces s = (List.filter (fn c => not (c = (#" "))) (explode s)) = []

    fun checkQueries _ [] = ()
      | checkQueries (vars : range IntBinaryMap.map,
                      depends : int list IntBinaryMap.map)
                     ((x,y,errfn)::restoflist) =
          let
              fun findVar (Var l) =
                    case IntBinaryMap.find(vars, l) of
                        NONE => (print ("findVar failed for "
                                         ^(varToStr (Var l))^"\n"); raise Err)
                      | SOME r => r
              fun evalCoeffVar ({coeff,var}) =
                    rangeMul (coeff, findVar var)
              fun evalTerm (T(l,c)) =
                    foldl rangeAdd (Int c,Int c) (map evalCoeffVar l)
                    handle _ => (NegativeInfinity,Infinity)
              val (xr as (xlow,_),yr as (_,yhigh)) = (evalTerm x, evalTerm y)
              val ok = endptGe (xlow,yhigh)
              (*
              val _ = print ("Debug: checking `"^(toString x)^" >= "^
                             (toString y)^"': "^
                             (rangeToStr xr)^", "^(rangeToStr yr)^": seems "^
                             (if ok then "ok" else "not ok")^"\n")
              *)
              fun buildDep m prefix lbl =
                    let val nm = case IntBinaryMap.find(!names, lbl) of
                                       NONE => "???"
                                     | SOME nn => nn
                        val prefix'  = prefix^" <- "^nm
                        val prefix'' = prefix^"    "^(nspaces (size nm))
                        fun fin p = if allspaces p then "" else p^"\n"
                        fun f p i = buildDep (IntBinarySet.add(m,i)) p i
                    in
                        if IntBinarySet.member(m,lbl) then fin prefix' else
                        case IntBinaryMap.find(depends, lbl) of
                            NONE => fin prefix'
                          | SOME [] => fin prefix'
                          | SOME (d::ds) =>
                               concat ((f prefix' d)::(map (f prefix'') ds))
                    end
              fun dep (T(l,_),T(l',_)) =
                    let fun g ({var=Var(lbl),coeff=_}) = lbl
                        val s = IntBinarySet.addList(IntBinarySet.empty,
                                       map g (l@l'))
                        val l' = IntBinarySet.listItems(s)
                        val f = buildDep (IntBinarySet.empty) " "
                    in  concat (map f l')
                    end
          in  if ok then () else errfn (xr,yr, dep (x,y));
              checkQueries (vars, depends) restoflist
          end

    val bugMsg = (
          "FAILED: couldn't find ssolver.\n"^ 
          "  Most likely, you ran into a known limitation of the prototype:\n"^
          "  You have to run boon from the same directory as the source.\n"^
          "  Try that and see if it fixes the problem.\n"^
          "  (Sorry about the known bug; I'll fix it someday.)\n"
        )

    (* Solve the constraint system, handling "query constraints" too. *)
    fun solve () =
          let
              fun fail msg output e =
                    (print ("Failure: "^msg^"\n");
                     print ("Input = {\n"^(!curConstraints)^"}\n");
                     print ("Output = {\n"^output^"}\n");
                     raise e)
              (* val _ = print ("Debug: input = \n"^(!curConstraints)) *)
              val allOut = externalSolver (!curConstraints)
                           handle e => fail "Solver failed" "" e
              (* val _ = print ("Debug: output = \n"^allOut) *)
              val isFeas = if allOut=infeasible then false
                           else if String.isPrefix feasible allOut then true
                           else if allOut=execFailed then
                                (print bugMsg; raise Err)
                           else fail "Odd response" allOut Err
          in
              if isFeas
              then let val (vars,depends) =
                         getBounds allOut
                         handle e => fail "getBounds" allOut e
                       val depends' = fixDepends depends
                       (*
                       fun f y = " S"^(myIntStr y)
                       fun p (x,ys) =
                             print ((f x)^": "^(concat (map f ys))^"\n")
                       fun p' (x,y) = print ((f x)^": "^y^"\n")
                       fun q d s = (print ("Debug: "^s^" =\n");
                                  app p (IntBinaryMap.listItemsi(d)))
                       fun q' n s = (print ("Debug: "^s^" =\n");
                                     app p' (IntBinaryMap.listItemsi(n)))
                       val _ = q depends "depends"
                       val _ = q' (!names) "names"
                       val _ = q depends' "depends'"
                       *)
                   in
                       checkQueries (vars,depends') (!curQueries)
                   end
              else fail "Constraint system infeasible!?" allOut Err
          end


    val op = = eqTerms
    val op + = addTerms
    val op - = subTerms
    val op * = mulTerms
    val op >= = termGe
  end
