/* Have to make an assumption about the semantics of `S + 5' when `S'
   is empty.  I will just assume throughout that all sets (`S') mentioned
   are non-empty; this resolves the definitional problems, I think.
   Two notes on familiar identities:
   1. S + S = 2S when S is a range (but not for general sets).
   2. 2S - S is not the same as S (unless S is singleton).
      Therefore, we keep track of positive and negative multiples separately.
*/

/* Known bug: 4*S = T fails.  It actually should not be supported
   in the first place...   (Sometimes 4*S subset T works.) */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <values.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>


static void debug_check_graph(void);
static void debug_dumpalleqns(void);

/* poor man's varargs. :-) */
void fatal_hook(void) { ; } /* a handy breakpoint for debugging */
#define fatal(f, rest...) do { char buf[1024]; sprintf(buf, f, ##rest); \
                            fprintf(stderr, "Error: %s\n", buf); \
			    fatal_hook(); \
                            debug_dumpalleqns(); \
                            exit(100); } \
                          while (0)
int debugging = 0;
#define debug(f, rest...) do { \
                        if (debugging) fprintf(stderr, f, ##rest); \
                        } while (0)
void warn_hook(void) { ; } /* a handy breakpoint for debugging */
#define warn(f, rest...) do { \
                        fprintf(stderr, f, ##rest); \
                        warn_hook(); \
                        } while (0)
static void *mymalloc(size_t len) {
        void *p = malloc(len);
        if (!p)
                fatal("Out of memory.");
        return p;
}
static void *myrealloc(void *old, size_t len) {
        void *p = realloc(old, len);
        if (!p)
                fatal("Out of memory.");
        return p;
}



/* A signed int type, with +Infinity and -Infinity. */
/* (Fixme: replace with a `struct { int; }' someday: better typechecking.) */
typedef int intinf;
#define INF ((intinf)MAXSHORT)
#define NEGINF ((intinf)(-MAXSHORT))

static intinf inf(long int x) {
        if (x <= NEGINF) {
                debug("Warning: inf() possible underflow (%ld)\n", x);
                return NEGINF;
        } else if (x >= INF) {
                debug("Warning: inf() possible overflow (%ld)\n", x);
                return INF;
        } else
                return (intinf)x;
}
static const intinf infzero = (intinf)0;
static char *infstr(intinf x) {
        static char buf[40];
        if (x <= NEGINF)
                sprintf(buf, "-Infinity");
        else if (x >= INF)
                sprintf(buf, "+Infinity");
        else
                sprintf(buf, "%d", (int)x);
        return buf;
}
/* Returns 0 if x=0, 2 if x=+Inf, -2 if x=-Inf, else sign(x). */
static int infsign(intinf x) {
        if (x >= INF)
                return 2;
        else if (x <= NEGINF)
                return -2;
        else if (x > infzero)
                return 1;
        else if (x < infzero)
                return -1;
        else
                return 0;
}
#define INFADD0_FATAL	(INF+0xBAD0000)
static intinf infadd0(intinf x, intinf y, intinf default_ifbad) {
        if (x < INF && x > NEGINF && y < INF && y > NEGINF)
                return inf( ((long int)x) + ((long int)y) );
        switch(infsign(x)+infsign(y)) {
                case  1: case  2: case  3: case  4: return INF;
                case -1: case -2: case -3: case -4: return NEGINF;
                case  0: if (default_ifbad == INFADD0_FATAL)
				fatal("INF + NEGINF");
                         else
			 	return default_ifbad;
                default: fatal("impossible infadd().");
        }
}
static intinf infadd(intinf x, intinf y) {
	return infadd0(x, y, INFADD0_FATAL);
}
static intinf infnegate(intinf x) {
        if (x >= INF)
                return NEGINF;
        else if (x <= NEGINF)
                return INF;
        assert((-x) > NEGINF && (-x) < INF);
        return -x;
}
static intinf infsub(intinf x, intinf y) {
        return infadd(x, infnegate(y));
}
static intinf infmul(intinf x, intinf y) {
        if (x < INF && x > NEGINF && y < INF && y > NEGINF)
                return inf( ((long int)x) * ((long int)y) );
        switch(infsign(x)*infsign(y)) {
                case  2: case  4: return INF;
                case -2: case -4: return NEGINF;
                case  0: debug("Warning: 0 * +/- Infinity.\n");
		         return 0;
                default: fatal("impossible infmul().");
        }
}
static intinf infdiv(intinf x, intinf y) {
        if (!y && !x)
                fatal("0 / 0");
        else if (!y)
                return (x>0) ? INF : NEGINF;
        else if (x < INF && x > NEGINF && y < INF && y > NEGINF) {
                long int d = ((long int)x) / ((long int)y);
                if (d*((long int)y) != ((long int)x))
                        fatal("%d / %d: not evenly divisible.", x, y);
                return inf(d);
        } else if ((x >= INF || x <= NEGINF) && (y >= INF || y <= NEGINF))
                fatal("+/-Infinity / +/-Infinity");
        else if (x >= INF && y < INF && y > NEGINF)
                return (y>0) ? INF : NEGINF;
        else if (x <= NEGINF && y < INF && y > NEGINF)
                return (y>0) ? NEGINF : INF;
        else
                return infzero;
}

/* The range of integers from `lo' to `hi' (inclusive). */
typedef struct {
        intinf lo;
        intinf hi;
} range_t;

static const range_t zerorange = {(intinf)0, (intinf)0};
static const range_t emptyrange = {(intinf)INF, (intinf)NEGINF};

static char *rangestr(range_t x) {
        static char buf[80];
        sprintf(buf, "%s..", infstr(x.lo));
        strcat(buf, infstr(x.hi));
        return buf;
}
static range_t rangesingleton(intinf n) {
        range_t r;
        r.lo = r.hi = n;
        if (r.lo == INF && r.lo == INF) {
                debug("Debug: rangesingleton() = {INF,INF}.\n");
        } else if (r.lo == NEGINF && r.hi == NEGINF) {
                debug("Debug: rangesingleton() = {NEGINF,NEGINF}.\n");
        }
        return r;
}
static int rangeempty(range_t x) {
        return (x.lo > x.hi);
}
static range_t rangeadd(range_t x, range_t y) {
        range_t new;
        if (rangeempty(x) || rangeempty(y))
                return emptyrange;
        new.lo = infadd0(x.lo, y.lo, NEGINF);
        new.hi = infadd0(x.hi, y.hi, INF);
        return new;
}
static range_t rangemul(intinf x, range_t y) {
        range_t new;
        if (rangeempty(y))
                return emptyrange;
        if (x == infzero) {
                new = zerorange;
                /*
                if (y.lo <= NEGINF || y.hi >= INF)
                        debug("Warning: substituting 0 for 0 * %s.\n",
                                rangestr(y));
                */
        } else if (x > 0) {
                new.lo = infmul(x, y.lo);
                new.hi = infmul(x, y.hi);
        } else {
		assert(x < 0);
                new.lo = infmul(x, y.hi);
                new.hi = infmul(x, y.lo);
        }
        return new;
}
static range_t rangesub(range_t x, range_t y) {
        return rangeadd(x, rangemul(inf(-1L), y));
}
static int rangesubset(range_t x, range_t y) {
        return (x.lo >= y.lo || y.lo <= NEGINF)
            && (x.hi <= y.hi || y.hi >= INF);
}
static range_t rangeunion(range_t x, range_t y) {
        range_t new;
        if (rangeempty(x) || rangeempty(y))
                return emptyrange;
        new.lo = (x.lo > y.lo) ? y.lo : x.lo;
        new.hi = (x.hi < y.hi) ? y.hi : x.hi;
        return new;
}
static range_t rangemin(range_t x, range_t y) {
        range_t new;
        if (rangeempty(x) || rangeempty(y))
                return emptyrange;
        new.lo = (x.lo > y.lo) ? y.lo : x.lo;
        new.hi = (x.hi > y.hi) ? y.hi : x.hi;
        return new;
}
static int rangeeq(range_t x, range_t y) {
        if (rangeempty(x) || rangeempty(y))
                return rangeempty(x) == rangeempty(y);
        return x.lo == y.lo && x.hi == y.hi;
}

/* A list of equations. */
typedef struct eqnlist_s {
        struct eqn_s *e;
        struct eqnlist_s *next;
} eqnlist_t;

/* Represents the affine function $S \mapsto -zS+aS+b$. */
typedef struct {
        intinf z;
        intinf a;
        range_t b;
} affine_t;

static const affine_t affineidentity =
   {(intinf)0, (intinf)1, {(intinf)0, (intinf)0}};

/* Returns $f(x)$. */
static range_t affineapp(affine_t f, range_t x) {
        return rangeadd(rangesub(rangemul(f.a, x), rangemul(f.z, x)), f.b);
}

static int affineeq(affine_t f, affine_t g) {
        return f.z == g.z && f.a == g.a && rangeeq(f.b, g.b);
}

/* For debugging only. */
static range_t randomrange(void) {
        range_t rng;
        signed char l, h;
        switch(random()&3) {
                case 0: l=h=0; break;
                case 1: l=h=random(); break;
                case 2: case 3:
                        do { l = random(); h = random(); } while (l > h);
                        break;
                default: assert(0); exit(1);
        }
        rng.lo = inf(l); rng.hi = inf(h);
        return rng;
}


/* Returns $g \circ f$, i.e. the composition of f and g (applying f first). */
static affine_t affinecomp(affine_t f, affine_t g) {
        intinf z = infadd(infmul(f.a, g.z), infmul(f.z, g.a));
        intinf a = infadd(infmul(f.a, g.a), infmul(f.z, g.z));
        range_t b =
          rangeadd(rangesub(rangemul(g.a, f.b), rangemul(g.z, f.b)), g.b);
        affine_t gf = {z, a, b};
        { // test it
                range_t x = randomrange();
                range_t y = affineapp(gf, x);
                range_t yy = affineapp(g, affineapp(f, x));
                assert(rangeeq(y, yy));
        }
        return gf;
}


/* Given $h=g \circ f$ and $f$, returns $g$. */
static affine_t affinedecomp(affine_t h, affine_t f) {
        intinf div, d, y, dy;
        affine_t g;

        assert(!rangeempty(f.b) && !rangeempty(h.b));
        assert(f.a >= infzero && f.z >= infzero
                && h.a >= infzero && h.z >= infzero);
        assert((f.a || f.z) && (h.a || h.z));

/* Oh my, this is tricky!  Watch out for the special semantics
   of -zx+ax (which is NOT equivalent to (a-z)x).
   Algorithm:
   If F:x -> -zx+ax+b..c, G:x -> -yx+dx+e..f, H:x -> -wx+gx+h..i, then
   y = (aw-gz)/(a^2-z^2), d = (ag-wz)/(a^2-z^2),
   e = h + cy - bd, f = i + by - cd. */

        div = infsub(infmul(f.a,f.a), infmul(f.z,f.z));
        y = infdiv(infsub(infmul(f.a,h.z), infmul(f.z,h.a)), div);
        d = infdiv(infsub(infmul(f.a,h.a), infmul(f.z,h.z)), div);
        g.z = y; g.a = d;
        dy = infsub(d, y);
        g.b.lo = infsub(infadd(h.b.lo, infmul(y, f.b.hi)), infmul(d, f.b.lo));
        g.b.hi = infsub(infadd(h.b.hi, infmul(y, f.b.lo)), infmul(d, f.b.hi));
        assert(!rangeempty(g.b));

        { // test it
                affine_t hh = affinecomp(f, g);
                assert(affineeq(h, hh));
        }
        return g;
}

static affine_t randomaffine(void) {
        affine_t f;
        unsigned char z, a;
        do {
                switch(random()&3) {
                        case 0: z = 0; break;
                        case 1: z = 1; break;
                        case 2: z = 2; break;
                        case 3: do { z = random(); } while (z > 16); break;
                        default: assert(0); exit(1);
                }
                switch(random()&3) {
                        case 0: a = 0; break;
                        case 1: a = 1; break;
                        case 2: a = 2; break;
                        case 3: do { a = random(); } while (a > 16); break;
                        default: assert(0); exit(1);
                }
        } while (a == z);
        f.z = inf((long)z); f.a = inf((long)a);
        f.b = randomrange();
        return f;
}
static void affinetest(void) {
        int i;
        affine_t f, g, h, gg;
        for (i=0; i<1000; i++) {
                f = randomaffine();
                g = randomaffine();
                h = affinecomp(f, g);
                gg = affinedecomp(h, f);
                assert(affineeq(g, gg));
        }
}

static char *affinestr(affine_t af) {
        static char buf[160];
        sprintf(buf, "$x \\mapsto -%s x", infstr(af.z));
        sprintf(buf, "%s + %s x", buf, infstr(af.a));
        sprintf(buf, "%s + %s$", buf, rangestr(af.b));
        return buf;
}

/* ALIVE:     currently a "live" set (still changing)
   SURVIVES:  will remain "live" in the next iteration
   FINALIZED: true when we're done changing this set,
   QUERIED:   we're asked to output the possible values of this set,
   RELEVANT:  relevant to some query,
   MARKED:    a generic ``marker'' for any of many graph searches,
   GRAY:      another marker for DFS
              (indicates an edge has been discovered but not yet finished),
*/
typedef enum {ALIVE=1, SURVIVES=2, FINALIZED=4, QUERIED=8, RELEVANT=16,
              MARKED=32, GRAY=64} setflags_t;

/* A set variable. */
typedef struct set_s {
        /* (In my algorithm, sets start out as empty, then grow as required
            by the eqn's.) */
        range_t r; /* current approximation */
        range_t hardbound; /* hard lower and upper bounds */

        setflags_t flags;

        eqnlist_t *affects; /* If s changes, what eqns does s affect? */
        eqnlist_t *affectedby; /* What eqns affect s? */

        char *name;

        struct set_s *nextset; /* For `allsets'. */

#define MAXDEPENDS 8
        struct set_s *depends[MAXDEPENDS]; /* Dependency tracking. */

        /* For the DFS; see findcycles(). */

        /* dfs_backptr_*: The DFS predecessor tree:
           _s contains the prev vertex,
           _e contains the edge used to get from _s to this. */
        struct set_s *dfs_backptr_s;
        struct eqn_s *dfs_backptr_e;
        /* runningaf: A running composition of the affine's on all the
           edges traversed from the root of the predecessor tree to me.
           i.e. we have the eqn `runningaf(root of pred. tree) subset this' */
        affine_t runningaf;
} set_t;

/* A term. */
typedef struct {
        intinf coeff;
        set_t *s;
} term_t;
/* A SORTED list of terms. */
typedef struct termlist_s {
        term_t *t;
        struct termlist_s *next;
} termlist_t;

/* A set constraint (``equation''). */
typedef struct eqn_s {
        /* Represents the set constraint
           `sum of the lhs terms  +  c   op  rhs term'. */
        termlist_t *lhs;
        range_t c; /* A constant additive term added to the left side. */
        enum {SUBSET, EQUALS, MIN_SUBSET} op;
           /* MIN_SUBSET represents `min {lhs terms, c}  subset  rhs term' */
        enum {DELETED=1} eflags;
        term_t rhs;
        termlist_t *extra_depends;
           /* deleted LHS-terms; retained for dependency tracking */
} eqn_t;

/* Generate a printable representation of e for debugging purposes. */
static char *debug_dumpeqn(eqn_t *e) {
        static char buf[1024], *q;
        termlist_t *p;
        buf[0] = '\0';
        assert(e && e->rhs.s && e->rhs.s->name);
        if (e->op == MIN_SUBSET)
                sprintf(buf, "min {");
        for (p=e->lhs; p; p=p->next) {
                assert(p->t && p->t->s && p->t->s->name);
                if (p->t->coeff != 1)
                        sprintf(buf, "%s%d*", buf, p->t->coeff);
                sprintf(buf, "%s%s%s", buf, p->t->s->name,
                   (e->op==MIN_SUBSET)?", ":" + ");
        }
        switch(e->op) {
                case SUBSET: q = "subset"; break;
                case EQUALS: q = "="; break;
                case MIN_SUBSET: q = "} subset"; break;
                default: q = "???"; break;
        }
        sprintf(buf, "%s%s %s %s%s", buf, rangestr(e->c), q,
                e->rhs.s->name, (e->eflags & DELETED) ? " [deleted]" : "");
        if (e->extra_depends) {
                sprintf(buf, "%s [extra_depends:", buf);
                for (p=e->extra_depends; p; p=p->next) {
                        assert(p->t && p->t->s && p->t->s->name);
                        assert(p->t->coeff == inf(1L));
                        sprintf(buf, "%s %s", buf, p->t->s->name);
                }
                sprintf(buf, "%s]", buf);
        }
        return buf;
}

/* Delete _all_ instances of `e' from a list of equations.
   If not able delete at least `min' instances, fail.  */
static eqnlist_t *deleqn0(eqnlist_t *list, eqn_t *e, int min) {
        int ndeletions = 0;
        eqnlist_t *p, *q;
        assert(e && (min == 0 || min == 1));
        while (list && list->e == e) {
                q = list;
                list = list->next;
                free(q);
                ndeletions++;
        }
        for (p=list; p; p=p->next)
                while (p->next && p->next->e == e) {
                        q = p->next;
                        p->next = q->next;
                        free(q);
                        ndeletions++;
                }
        if (ndeletions < min)
                fatal("deleqn() failed: deleted %d, wanted at least %d.",
                        ndeletions, min);
        else if (ndeletions != 1)
                debug("Debug: deleqn() del'ing %d instances of %s\n",
                        ndeletions, debug_dumpeqn(e));
        return list;
}
static eqnlist_t *deleqn(eqnlist_t *list, eqn_t *e) {
        return deleqn0(list, e, 1);
}

/* Prepend `e' to a list of equations.
   (Won't create duplicates!) */
static eqnlist_t *inseqn(eqnlist_t *list, eqn_t *e) {
        eqnlist_t *l, *node;
        assert(!(e->eflags & DELETED));
        for (l=list; l; l=l->next)
                if (l->e == e)
                        return list; // no dup's
        node = mymalloc(sizeof(eqnlist_t));
        node->next = list;
        node->e = e;
        return node;
}
/* Duplicate an equation list (shallow copy: dup just `next' fields). */
static eqnlist_t *dupeqn(eqnlist_t *orig) {
        eqnlist_t *new = NULL, *p;
        for (p=orig; p; p=p->next)
                new = inseqn(new, p->e);
        return new;
}

/* Delete `t' from a list of terms (only once). */
static termlist_t *delterm(termlist_t *list, term_t *t) {
        termlist_t *p, *q;
        assert(t && list);
        if (list->t == t) {
                p = list->next;
                assert(p != list); // no cycles (not a circularly linked list)
                free(list->t);
                list->t = (term_t *)0xDEAD01;
                free(list);
                return p;
        }
        for (p=list; p->next; p=p->next)
                if (p->next->t == t) {
                        q = p->next;
                        p->next = q->next;
                        free(q->t);
                        q->t = (term_t *)0xDEAD02;
                        free(q);
                        return list;
                }
        fatal("delterm() failed.");
}
static int maybecoalesce(term_t *old, term_t *new) {
        assert(old && new && old->s == new->s);
        if (    (old->coeff >= infzero && new->coeff >= infzero)
             || (old->coeff <= infzero && new->coeff <= infzero) ) {
                debug("Debug maybecoalesce(): (%s", infstr(old->coeff));
                debug("+%s)*%s = ", infstr(new->coeff), new->s->name);
                debug("%s*%s (head)\n", infstr(infadd(old->coeff, new->coeff)),
                        old->s->name);
                old->coeff = infadd(old->coeff, new->coeff);
                free(new);
                return 1;
        } else {
                debug("Debug maybecoalesce(): not coalescing %s*%s and ",
                        infstr(old->coeff), old->s->name);
                debug("%s*%s.\n", infstr(new->coeff), new->s->name);
                return 0;
        }
}
/* Prepend `t' to a list of terms. */
/* Note: we promise not to end up with something like `0*x' or `x + 3*x'.
   But we could easily end up with `5*x + -2*x' (NOT the same as `3*x'!!). */
static termlist_t *insterm(termlist_t *list, term_t *t) {
        termlist_t *node = mymalloc(sizeof(termlist_t)), *p;
        node->next = list;
        node->t = t;
        assert(t && t->s);
        assert(!list || (list->t && list->t->s));
        if (t->coeff == infzero) {
                free(node);
                return list;
        }
        if (!list || t->s < list->t->s)
                return node;
        if (t->s == list->t->s && maybecoalesce(list->t, t)) {
                free(node);
                return list;
        }
        for (p=list; p->next; p=p->next) {
                assert(p->next && p->next->t
                       && p->next->t->s && p->next->t->coeff);
                if (t->s < p->next->t->s) {
                        node->next = p->next;
                        p->next = node;
                        return list;
                } else if (t->s == p->next->t->s
                                && maybecoalesce(p->next->t, t)) {
                        free(node);
                        return list;
                }
        }
        node->next = NULL;
        p->next = node;
        return list;
        fatal("insterm() failed.");
}
/* Duplicate a list of terms (shallow copy: dup just `next' fields). */
static termlist_t *dupterm(termlist_t *orig) {
        termlist_t *new = NULL, *p;
        for (p=orig; p; p=p->next)
                new = insterm(new, p->t);
        return new;
}


/* Aww hell, the whole damn constraint system is infeasible. */
static void infeasible(void) {
        printf("This problem is infeasible\n");
        exit(1);
}

/* Update `y' to include `x'.  Return 1 if `y' changed, else 0. */
static int rangeinrange(range_t x, range_t *y) {
        int changed = 0;
        assert(y);
        if (x.lo < y->lo) {
                assert(x.lo >= NEGINF);
                y->lo = x.lo;
                changed = 1;
        }
        if (x.hi > y->hi) {
                assert(x.hi <= INF);
                y->hi = x.hi;
                changed = 1;
        }
        return changed;
}
/* s->r just changed; make sure it doesn't violate the hard bounds. */
static void checkhardbounds(set_t *s) {
        assert(s);
        if (s->r.lo < s->hardbound.lo) {
                fprintf(stderr, "Error: %s lo=%s, ", s->name, infstr(s->r.lo));
                fprintf(stderr, " hardlo=%s.\n", infstr(s->hardbound.lo));
                infeasible();
        }
        if (s->r.hi > s->hardbound.hi) {
                fprintf(stderr, "Error: %s hi=%s, ", s->name, infstr(s->r.hi));
                fprintf(stderr, " hardhi=%s.\n", infstr(s->hardbound.hi));
                infeasible();
        }
}

/* Force the set `s' to include the range `r',
   as though we had written the constraint `r subset s'.
   Also touch `s' if it changes.
   Returns 1 if `s' changed, 0 otherwise. */
static int rangeinset(range_t r, set_t *s) {
        assert(s);
        debug("Debug:  force `%s in %s'", rangestr(r), s->name);
        if (rangeinrange(r, &(s->r))) {
                debug(" (now includes %s).\n", rangestr(s->r));
                s->flags |= SURVIVES;
                checkhardbounds(s);
                return 1;
        }
        debug(".\n");
        return 0;
}

static void adddepends_set(set_t *s, set_t *t) {
        int i;
        for (i=0; i<MAXDEPENDS && s->depends[i]; i++)
                if (s->depends[i] == t)
                        return;
        if (i < MAXDEPENDS) {
                s->depends[i] = t;
                if (++i < MAXDEPENDS)
                        s->depends[i] = NULL;
                return;
        }
        debug("adddepends_set(%s): no room for %s.\n", s->name, t->name);
}

/* Adds to the existing dependency information, rather than overwriting. */
static void adddepends(set_t *s, eqn_t *e) {
        termlist_t *t;
        assert(s && e);
        for (t=e->lhs; t; t=t->next) {
                assert(t->t && t->t->s && t->t->s->name);
                adddepends_set(s, t->t->s);
        }
        for (t=e->extra_depends; t; t=t->next) {
                assert(t->t && t->t->s && t->t->s->name);
                adddepends_set(s, t->t->s);
        }
}

/* Overwrites the dependency information so `s' depends on
   the sets on the LHS of `e'. */
static void overwritedepends(set_t *s, eqn_t *e) {
        assert(s && e);
        debug("Debug: overwritedepends(%s, %s)\n", s->name, debug_dumpeqn(e));
        if (!e->lhs && !e->extra_depends)
                return; // short-circuit: don't overwrite old value
        s->depends[0] = NULL;
        adddepends(s, e);
}
/* Same, but for a set, instead of an equation. */
static void overwritedepends_set(set_t *s, set_t *t) {
        assert(s && t);
        s->depends[0] = t;
        if (MAXDEPENDS > 1)
                s->depends[1] = NULL;
}

static void killeqn(eqn_t *e);

/* Simplify an equation by folding in 'finalized' sets. */
static void simplify(eqn_t *e, int modify_graph) {
        termlist_t *p, *next;
        set_t *s;

        assert(e && e->rhs.s && e->rhs.coeff == inf(1L));
        assert(e->op == SUBSET || e->op == MIN_SUBSET);
        assert(!(e->eflags & DELETED));
        // debug("Debug simplify(): before: %s\n", debug_dumpeqn(e));

        p = e->lhs;
        while (p) {
                assert(p->t && p->t->s);
                next = p->next; // p might be deleted below
                s = p->t->s;
                if (s->flags & FINALIZED || p->t->coeff == infzero) {
                        range_t r;
                        debug("Debug: folding `%s' into %s.\n",
                                s->name, debug_dumpeqn(e));

                        /* If any set on LHS is empty, e tells us nothing. */
                        if (rangeempty(s->r)) {
                                if (modify_graph) {
                                        debug("Debug: %s empty: kill `%s'.\n",
                                                s->name, debug_dumpeqn(e));
                                        killeqn(e);
                                }
                                return;
                        }

                        assert(p->t->coeff != infzero); /* Avoid 0 * INF. */
                        r = rangemul(p->t->coeff, s->r);

                        /* Fold r into the constant term of e. */
                        if (e->op == SUBSET)
                                e->c = rangeadd(e->c, r);
                        else if (e->op == MIN_SUBSET)
                                e->c = rangemin(e->c, r);
                        else
                                fatal("This can't happen (e->op=%d).", e->op);
                        if (modify_graph) {
                                /*
                                debug("Debug: del `%s' from `%s'->affects.\n",
                                        debug_dumpeqn(e), s->name);
                                */
                                s->affects = deleqn0(s->affects, e, 0);
                        }
                        /*
                        debug("Debug: deleting `%s' from `%s'->lhs.\n",
                                s->name, debug_dumpeqn(e));
                        */
                        e->lhs = delterm(e->lhs, p->t);
                        { term_t *tp = mymalloc(sizeof(term_t));
                          tp->coeff = inf(1L);
                          tp->s = s;
                          e->extra_depends = insterm(e->extra_depends, tp);
                        }
                        // debug("Debug: now e = `%s'.\n", debug_dumpeqn(e));
                }
                p = next;
        }
        // debug("Debug simplify(): after: %s\n", debug_dumpeqn(e));
}

static void finalize(set_t *s);


/* Prune this equation if it is no longer relevant.
   (i.e. if there are no more set variables on the left-hand-side)
   Returns 1 if pruned, else 0.  */
static int prune(eqn_t *e) {
        set_t *rhs;
        assert(e && e->rhs.s);
        if (e->lhs)
                return 0; /* LHS not empty; no pruning needed. */

        // debug("Debug: pruning %s.\n", debug_dumpeqn(e));

        /* Ok, this equation is irrelevant.
           It should already have been deleted out of every s->affects,
           and the constant term has already been propagated to the RHS, so
           all we have to do is delete it from the relevant s->affectedby. 
           and make sure the constant term gets propagated to the RHS. */
        rhs = e->rhs.s;
        killeqn(e); /* Delete it. (XXX: Ok?) */
        /* ^^^ killeqn(e) deletes e from rhs->affectedby. */

        /* Oh, and it's possible that this might have been the last
           eqn to affect rhs; if so, rhs will never be
           changed again, and can be marked FINALIZED. */
        if (!rhs->affectedby)
                finalize(rhs);

        return 1;
}

static void simplifyprune(eqn_t *e) {
        // debug("Debug: simplifyprune %s.\n", debug_dumpeqn(e));
        assert(e && e->rhs.s);
        assert(e->op == SUBSET || e->op == MIN_SUBSET);
        assert(!(e->eflags & DELETED));
        simplify(e, 1);
        if (!e->lhs && !(e->eflags & DELETED)) {
                if (rangeinset(e->c, e->rhs.s))
                        overwritedepends(e->rhs.s, e);
                (void)prune(e);
        }
        /* Check: `e' might have been deleted by simplify, which might have
           eliminated the last eqn affected `e->rhs.s'.  If this happens, we
           can finalize `e->rhs.s' right away. */
        assert(e->rhs.s);
        if (!e->rhs.s->affectedby && !(e->rhs.s->flags & FINALIZED)) {
                debug("Debug:  nothing affects %s, finalizing.\n",
                        e->rhs.s->name);
                finalize(e->rhs.s);
        }
}

static void finalize(set_t *s) {
        eqnlist_t *ep, *next;
        assert(s);
        debug("Debug: finalizing %s.\n", s->name);
        if (s->flags & FINALIZED) {
                debug("Debug finalize(): already finalized!\n");
                return;
        }
        s->flags |= FINALIZED;

        /*
        debug("Debug:  `%s'->affects = %s, ...\n", s->name,
                s->affects ? debug_dumpeqn(s->affects->e) : "(null)");
        */

        /* Simplify (and maybe prune?) other eqns. */
        for (ep=s->affects; ep; ep=next) {
                assert(ep && ep->e && !(ep->e->eflags & DELETED));
                next = ep->next;
                /*
                debug("Debug:  `%s'->affects=%X, ep=%X, ep->e=%X.\n",
                        s->name, (int)s->affects, (int)ep, (int)ep->e);
                debug("Debug:   finalize(%s) -> simplifyprune %s\n",
                        s->name, debug_dumpeqn(ep->e));
                */
                simplifyprune(ep->e);
        }
        // debug("Debug: finished loop (s=`%s')\n", s->name);
}


#define EQNSTBLSIZ  (1<<20)
static int neqns = 0, eqnssiz = EQNSTBLSIZ;
// Can't make it dynamic! (how to resize without leaving dangling ptrs?)
static eqn_t the_eqns[EQNSTBLSIZ];
static eqn_t *eqns = &the_eqns[0];
static eqn_t *neweqn(void) {
        eqn_t *e;
        if (neqns >= eqnssiz) {
                fatal("More than %d eqns, bump up eqnssiz and recompile.\n",
                        eqnssiz);
                /*  Following creates a huge bug.  Don't use it!
                eqnssiz = (eqnssiz<1024) ? 1024 : (2*eqnssiz);
		debug("Debug: realloc'ing eqns to %d elements...\n", eqnssiz);
                eqns = myrealloc(eqns, eqnssiz*sizeof(eqn_t));
                */
        }
        e = &eqns[neqns++];
        e->c = zerorange;
        e->lhs = NULL;
        e->rhs.s = NULL; e->rhs.coeff = inf(1L);
        e->op = SUBSET;
        e->eflags = 0;
        e->extra_depends = NULL;
        return e;
}

/* Delete this equation from the global list of equations. */
static void killeqn(eqn_t *e) {
        termlist_t *t;
        debug("Debug: killeqn(%s)\n", debug_dumpeqn(e));
        if (e->eflags & DELETED) {
                debug("Debug:  already dead.\n");
                return;
        }
        e->eflags |= DELETED;
        for (t=e->lhs; t; t=t->next) {
                assert(t && t->t && t->t->s && t->t->coeff != infzero);
                t->t->s->affects = deleqn0(t->t->s->affects, e, 0);
        }
        assert(e->rhs.s && e->rhs.coeff != infzero);
        e->rhs.s->affectedby = deleqn(e->rhs.s->affectedby, e);
        /* e->rhs.s = NULL ??  probably bad, because of aftermath() */
        /* XXX: Can't free, because aftermath() checks all equations.
           And anyway, equations aren't malloc'ed. */
}

/* Show _all_ the current equations. */
static void debug_dumpalleqns(void) {
        int i;
        debug("Debug: -- dumping all constraints:\n");
        for (i=0; i<neqns; i++)
                debug("   %s\n", debug_dumpeqn(&eqns[i]));
        debug("Debug: -- end dump.\n");
}

/* Checks whether `e' is a duplicate of another equation already
   in the graph.  (This works whether `e' is currently in the graph
   or not.)
   Output: 1 if `e' is redundant, 0 otherwise.
   (May modify an eqn in the graph slightly, if there is another
   equation with the same terms as `e' but with a different constant
   term, but this shouldn't matter to anyone.) */
static int isdup(eqn_t *e) {
        eqnlist_t *ep;
        termlist_t *t;

        /* XXX: Uses that lhs termlists are kept in sorted order.
           (FIX: is this still true???) */

        assert(!(e->eflags & DELETED));
        // debug("Debug:  isdup(%s)\n", debug_dumpeqn(e));
        for (ep=e->rhs.s->affectedby; ep; ep=ep->next) {
                termlist_t *t2;
                assert(ep->e);
                if (ep->e == e)
                        continue;
                // debug("Debug:   checking against %s\n", debug_dumpeqn(ep->e));
                for (t=ep->e->lhs,t2=e->lhs; t && t2; t=t->next,t2=t2->next) {
                        assert(t->t && t->t->s && t2->t && t2->t->s);
                        if (t->t->coeff != t2->t->coeff
                                       || t->t->s != t2->t->s)
                                break;
                }
                if (t || t2)
                        continue;

                /* `op' mismatches can cause subtle bugs. */
                if (e->op == EQUALS && ep->e->op == SUBSET)
                        return 0;
                else if (e->op == SUBSET && ep->e->op == EQUALS)
                        return rangesubset(e->c, ep->e->c);
                assert(e->op == ep->e->op);

                if (e->op == MIN_SUBSET)
                        return 1; /* Shortcut: Don't do the union below. */

                /* Found another equation with the same terms, don't need
                   to add a new one.  But if they have different constant
                   terms, we must take their union, since both have to hold. */
                // debug("Debug:  already exists, rangeunion()\n");
                /* XXX: should really free e->lhs here. */
                assert(e->op == SUBSET || e->op == EQUALS);
                if (rangeinrange(e->c, &(ep->e->c)) && ep->e->lhs) {
                        assert(ep->e->lhs->t && ep->e->lhs->t->s);
                        ep->e->lhs->t->s->flags |= SURVIVES;
                }
                return 1;
        }
        return 0;
}


/* Input: an equation `tmp', that was allocated locally in the caller
   (NOT with `neweqn()'!!), and that is NOT referenced in the graph.
   Operation: We add `tmp' to the graph (updating all the necessary
   affects and affectedby lists), _unless_ it would create a duplicate.
   Returns: 0 if it was a duplicate, 1 if it added new information
   to the graph.
   It's the caller's responsibility to "touch" a LHS set if this equation
   is added. */
static int addeqn_unlessdup(eqn_t *tmp) {
        eqn_t *prm; /* permanent */
        termlist_t *t;

        /* See if there is already another equation with the same
           terms as `tmp'. */
        assert(tmp && tmp->rhs.s);
        assert(!(tmp->eflags & DELETED));
        debug("Debug: addeqn_unlessdup(%s)\n", debug_dumpeqn(tmp));
        if (isdup(tmp))
                return 0;

        /* This is a totally new equation, so add it for real. */
        prm = neweqn();
        *prm = *tmp;
        prm->rhs.s->affectedby = inseqn(prm->rhs.s->affectedby, prm);
        for (t=prm->lhs; t; t=t->next) {
                assert(t->t && t->t->s);
                t->t->s->affects = inseqn(t->t->s->affects, prm);
        }
        debug("Debug:  it's new, added %s.\n", debug_dumpeqn(prm));
        return 1;
}


/* A linked list of all the sets in existence, linked by s->nextset fields. */
static set_t *allsets = NULL;


/* Finds cycles among the "live" sets.
   (A set is "live" if it is in one of the queues, or if it is
    reachable from another live set.)
   We ignore eqns `S + T + ... + c subset Z' where c is empty
   (probably shouldn't happen, but who knows).

   Every cycle lets us deduce new information about the sets involved in it.
   There are three cases:
   <1> The sum of the hi values around the cycle is > 0.
   <2> The sum of the lo values around the cycle is < 0.
   <3> Every edge on the cycle has lo=hi, and moreover
       the sum of those values is = 0.
   (To see that these cases enumerate all the possibilities,
    note that lo<hi for all edges on the cycle, and take sums.
    Thus, if you're not in <1> or <2>, we have
    0 <= sum(lo's) <= sum(hi's) <= 0, i.e. sum(lo's) = sum(hi's) = 0
    and we're in <3>.)
   Each of these cases gives some new information, as follows:
   <1> Deduce that every set in the cycle has the form {n,n+1,n+2,...}
   <2> Deduce that every set in the cycle has the form {n,n-1,n-2,...}
   <3> Deduce that all the sets in the cycle can be unified.
   (Proof: <1>, <2> are trivial.  <3> requires some thought, but it's ok.)

   Given this principle, the algorithm is very simple.
   We just do a depth-first search; each back-edge gives a cycle
   that can be eliminated.

   Many nice properties can be proven:
   1. If findcycles() is called, at least one cycle will be found.
      (Proof: run() only calls findcycles() if there are no nodes of
       in-degree zero, so at least one cycle exists.)
   2. Suppose we call findcycles() repeatedly.  Then each call to
      findcycles() removes at least one new cycle forever (with
      multiplicities, i.e. if a cycle falls into two cases simultaneously,
      it may be discovered two times, but it the second time will be
      classified into a different case than the first time), and
      no cycle will be removed twice.
      (Proof: First, if there are no new cycles to discover, the
      iteration terminates immediately.  Second, no cycle can be
      discovered twice, because no "flow" can be pushed through it
      and thus none of its sets will get "touched".  If a set S
      is not reachable from any "touched" set, then S can be finalized.
      Third, if there is a new cycle, the DFS will find it.)

   We use ALIVE to mark currently "live" sets, SURVIVES to mark
   sets which will be "live" in the next iteration, MARKED to indicate
   sets which have already been traversed by the DFS.

   Returns 0 when there are no further "live" sets to traverse,
   1 otherwise.
*/
static void dfsvisit(set_t *s, int lvl);
static int findcycles(void) {
        set_t *s;
        int rv = 0, i;

        for (s=allsets; s; s=s->nextset) {
                setflags_t f = s->flags;
                if (f & SURVIVES) {
                        s->flags = (f|ALIVE) & ~(SURVIVES|MARKED|GRAY);
                        rv = 1;
                } else {
                        s->flags = f & ~(ALIVE|SURVIVES|MARKED|GRAY);
                }
                s->dfs_backptr_s = NULL;
                s->dfs_backptr_e = NULL;
                s->runningaf = affineidentity;
        }

        /* Do a DFS. */
        debug("Debug findcycles(): doing DFS\n");
        i = 0;
        for (s=allsets; s; s=s->nextset) {
                if ((s->flags & (ALIVE|MARKED)) == ALIVE) {
                        dfsvisit(s, 0);
                }
        }

        for (i=0; i<neqns; i++)
                if (!(eqns[i].eflags & DELETED))
                        simplifyprune(&eqns[i]);

        return rv;
}

/* Returns the label on the edge from `s' to `e->rhs.s' via `e'.
   (By ``edge label'' we mean the affine function $f$ such that
   `e' induces `f(s) subset e->rhs.s'.) */
static affine_t edgelbl(set_t *s, eqn_t *e) {
        termlist_t *tp;
        affine_t af;
        assert(e->op == SUBSET);
        assert(e->rhs.coeff == inf(1L));
        af.z = af.a = infzero;
        af.b = e->c;
        for (tp=e->lhs; tp; tp=tp->next) {
                assert(tp->t && tp->t->s);
                assert(tp->t->coeff != infzero);
                if (tp->t->s == s) {
                        if (tp->t->coeff < 0) {
                                assert(af.z == infzero);
                                af.z = infnegate(tp->t->coeff);
                        } else {
                                assert(af.a == infzero);
                                af.a = tp->t->coeff;
                        }
                        continue;
                }
                af.b = rangeadd(af.b, rangemul(tp->t->coeff, tp->t->s->r));
        }
        assert(af.a != infzero || af.z != infzero);
        /*
        debug("Debug: edgelbl(%s->%s by %s) = %s\n",
                s->name, e->rhs.s->name, debug_dumpeqn(e), affinestr(af));
        */
        return af;
}

/* Given a MIN_SUBSET eqn `e', flow from `s' (on LHS of `e') to `e->rhs.s'. */
static void minflow(set_t *s, eqn_t *e) {
        termlist_t *tp;
        range_t r;
        assert(e->op == MIN_SUBSET);
        assert(e->rhs.coeff == inf(1L));
        r = e->c;
        for (tp=e->lhs; tp; tp=tp->next) {
                assert(tp->t && tp->t->s);
                assert(tp->t->coeff != infzero);
                r = rangemin(r, rangemul(tp->t->coeff, tp->t->s->r));
        }
        debug("Debug: minflow(%s,%s): ", s->name, debug_dumpeqn(e));
        debug("flowing %s...\n", rangestr(r));
        if (rangeinset(r, e->rhs.s)) {
                assert(!(e->rhs.s->flags & FINALIZED));
                overwritedepends(e->rhs.s, e);
        }
}

static void handlecycle(set_t *s, eqn_t *e);

static void dfsvisit(set_t *s, int lvl) {
        eqnlist_t *ep, *nextep;

        assert(s && s->name && !(s->flags & (MARKED|GRAY)));
        debug("Debug dfs(): visit `%s' (depth %d)\n", s->name, lvl);
        if (rangeempty(s->r)) {
                /* Shortcut -- abort, because we can't modify anyone else. */
                s->flags |= MARKED;
                return;
        }
        s->flags |= MARKED|GRAY;
        for (ep=s->affects; ep; ep=nextep) {
                eqn_t *e = ep->e;
                set_t *rhs = e->rhs.s;
                affine_t af;

                nextep = ep->next;

                assert(e && e->rhs.s && rhs);
                if (e->op == MIN_SUBSET)
                        minflow(s, e);
                if (e->op != SUBSET) {
                        debug("Debug: visit() skipping `%s' (not SUBSET).\n",
                                debug_dumpeqn(e));
                        continue;
                }
                af = edgelbl(s, e);

                if (rangeinset(affineapp(af, s->r), rhs)) {
                        if ((rhs->flags & (MARKED|GRAY)) == (MARKED|GRAY))
                                adddepends(rhs, e);
                        else
                                overwritedepends(rhs, e);
                } else
                        continue; /* Not modified, so _DO NOT_ visit it!! */

                /* We modified the RHS of this edge, so it will be
                   "live" in the next iteration.
                   Also, we visit it, if we haven't already. */
                assert(!(rhs->flags & FINALIZED));
                if (rhs->flags & MARKED) {
                        /* We already visited it.  Is it a cycle? */
                        if (rhs->flags & GRAY)
                                handlecycle(s, e); /* back-edge => cycle */
                        continue;
                } else {
                        /* We haven't visited it yet, so recurse. */
                        rhs->dfs_backptr_s = s;
                        rhs->dfs_backptr_e = e;
                        rhs->runningaf = affinecomp(s->runningaf, af);
                        dfsvisit(rhs, lvl+1);
                }
        }
        s->flags &= ~GRAY; /* Done with this vertex. */
}

/* We found a cycle.  It starts at e->rhs.s, then follows the DFS graph,
   then traverses s (as the next-to-last vertex), and the final edge in
   the cycle is e.  The cycle can be recovered (in reverse order) by
   retracing the `dfs_backptr_*' edges, starting at `s'.

   See above for what information can be deduced from each cycle. */
/* Pictorially:

   __
  /  \
 /    \
|      !  <-- the . in the ! is lastedge->rhs.s
 \    %|
  \__, |
     ^ |   the % is lastedge
     |
    the , is lastset
*/
/*
 This is too cool.  I get to use a fixed point theorem!
 Thm. Suppose $\{d,\ldots,e\} \in S$ (for $d\le e$),
      and $a \times S + \{b,\ldots,c\} \subset S$ (for $a>0$ and $b\le c$).
      (Everything in sight is over the integers.)
      Then there are two cases:
      1. $ad+b < d$, in which case $-\infty \in S$.
      2. $ae+c > e$, in which case $+\infty \in S$.
      These two cases exhaust the possibilities
      (but are not necessarily mutually exclusive),
      unless $a=1$ and $b=c=0$.
 Prf. WLOG, consider case 1, and suppose there is a least element $x$ of $S$.
      Then $ax+b \in S$, so $ax+b \ge x$; but also $x\le d$,
      so $(a-1)x+b \le (a-1)d+b < 0$ and $ax+b < x$, contradiction.
      On to the last statement of the theorem.
      First, if $a=1$, $b<c$, and $d<e$,
      then $(ae+c)-(ad+b)>e-d$, so 
      $\{ad+b,\ldots,ae+c\}$ cannot be contained in $\{d,\ldots,e\}$.
      Second, if $a>1$ and $d<e$, then again $(ae+c)-(ad+b)\ge a(e-d)>e-d$.
      Third, suppose $a>1$, $d=e$, and $b<c$: if $ad+b \ge d$
      then $ad+c \ge d+c-b > d$, and similarly if $ad+c \le d$
      then $ad+b \le d+b-c < d$.
      Finally, if $a=1$, $b=c \ne 0$, and $d=e$, then
      either $ad+b<d$ (if $b<0$) or $ae+c>e$ (if $c>0$).
      QED.
 Corollary.
      Suppose we get the relation $a \times S + \{b,\ldots,c\} \subset S$
      from a cycle, and we already know $\{d,\ldots,e\} \subset S$.
      Then we can extract all useful information from the relation in
      constant time.
 Prf. Suppose first $a>0$ and not all of $a=1,b=c=0$ hold;
      then the previous theorem solves the problem.
      Next, suppose that $a=1,b=c=0$; I claim that all the sets on
      the cycle can be unified (up to addition by a constant).
      Note that every relation on the cycle is of the form $U+\{x\} \subset V$.
      Let $T$ be any set on the cycle (other than $S$); then we have
      $S+x \subset T$ and $T-x \subset S$ for some $x$, and thus
      $T \subset S+x$, which implies $T=S+x$.
      Third, if $a=0$, the case is trivial.
      Finally, if $a<0$, we note that we may traverse the cycle
      twice to get a relation of the form
      $a^2 \times S + \{(a+1)b,\ldots,(a+1)c\} \subset S$.
      (XXX: need to analyze this case more fully.)
*/
/*
 Actually, we need to extend the theorem a bit.
 The proof actually simplifies nicely in this case!
 Thm.  Fix $a,z \ge 0$, $b\le c$, and $d\le e$.
      Define $S$ to be the minimal set satisfying $\{d,\ldots,e\} \in S$ and
      (*) $-z \times S + a \times S + \{b,\ldots,c\} \subset S$.
      (Everything is over the integers.)
      Let $d' = -\infty$ if $ad-ze+b<d$, otherwise $d$,
      and define $e'$ to be $+\infty$ if $ae-ze+c>e$ else $e$.
      Then
      1. $-\infty \in S$ if $ad-ze+b < d$, and
      2. $+\infty \in S$ if $ae-zd+c > e$,
      or $S=\{d,\ldots,e\}$ if neither clause applies.
 Prf. Let $f(d) = ad-ze+b$.  If $f(d)<d$, then $f(f(d))<f(d)$ (since $a>0$),
      and the first case is proved by induction.
      The proof of the second case goes similarly: consider $g(e)=ae-zd+c$.
      Finally, if $ad-ze+b \ge d$ and $ae-zd+c \le e$ then
      $S'=\{d,\ldots,e\} \subset S$ satisfies (*) and using the minimality
      of $S$ we conclude that $S=S'$.
 Corollary.
      Suppose we get a relation of the form (*) from a cycle, and we
      already know $\{d,\ldots,e\} \subset S$.  Then we can extract all
      useful information from (*) in constant time.
 Prf. If neither (1) nor (2) apply, we are done: $S=\{d,\ldots,e\}$.
      If both (1) and (2) apply, we are done: $S=\Z$.
      If (1) but not (2) applies, we learn that
      $\{-\infty,\ldots,e\} \subset S$ and re-apply the theorem a second
      time replacing $d$ by $-\infty$; the second time, either (2) will
      apply (in which case we are done: $S=\Z$) or neither clause will apply
      (we are also done: $S=\{-\infty,\ldots,e\}$).
      Finally, if (2) but not (1) applies, by symmetry a second application
      of the theorem will suffice.
*/

static void unifycycle(set_t *lastset, eqn_t *lastedge, int cyclelen);
static void handlecycle(set_t *lastset, eqn_t *lastedge) {
        set_t *s, *prev, *t=lastedge->rhs.s;
        eqn_t *e;
        affine_t cycleaf = affineidentity;
        affine_t cycle2af = affineidentity;
        range_t newr;
        int changed;
        int cyclelen = 0;

        assert(lastset && lastedge && lastedge->rhs.s);
        debug("Debug: handling cycle: %s", lastedge->rhs.s->name);
        /* Backtrace the cycle to calculate the affine function for the
           whole cycle (primarily for debugging, since the `cycle2af'
           computation below is more efficient). */
        for (s=lastedge->rhs.s,prev=lastset,e=lastedge;
             prev;
             s=prev,prev=s->dfs_backptr_s,e=s->dfs_backptr_e) {
                /* The edge from `prev' -> `s' by way of `e' is in the cycle. */
                affine_t edgeaf;
                assert(s && prev && e && e->rhs.s == s);
                assert(e->rhs.coeff == inf(1L) && e->op == SUBSET);
                debug(" <- %s", prev->name);
                cyclelen++;
                edgeaf = edgelbl(prev, e);
                cycleaf = affinecomp(cycleaf, edgeaf);
                if (prev == lastedge->rhs.s)
                        break;
        }
        assert(prev && prev == lastedge->rhs.s && prev == t);
        debug(".\n");
        /* XXX: not safe, if `lastset->runningaf' has e.g. Inf in its .b term
        cycle2af = affinecomp(
                affinedecomp(lastset->runningaf, lastedge->rhs.s->runningaf),
                edgelbl(lastset, lastedge)
        );
        */
        debug("Debug handlecycle(): cycleaf=%s.\n", affinestr(cycleaf));
        /* XXX
        debug(" cycle2af=%s.\n", affinestr(cycle2af));
        assert(affineeq(cycleaf, cycle2af));
        */
        debug("Debug handlecycle(): cyclelen=%d.\n", cyclelen);
        if (cycleaf.z > infzero)
                warn("Warning!: CycleAF = %s: Negative coeff's not tested much.\n", affinestr(cycleaf));
        if (cycleaf.a == infzero && cycleaf.z == infzero)
                fatal("CycleAF = %s: Zero coeff shouldn't happen.", affinestr(cycleaf));
        newr = affineapp(cycleaf, t->r);
        /* Now use the fixed point theorem... */
        if (newr.lo < t->r.lo)
                newr.lo = NEGINF;
        if (newr.hi > t->r.hi)
                newr.hi = INF;
        changed = rangeinset(newr, t);
        // No need to change dependencies; that was already done in dfsvisit()
        /* FIX: if z != 0, then we might need to apply the theorem
           a second time here. This is not crucial for correctness
           -- after all, if the cycle is still live, it will be found
           in the next pass of findcycles() -- but it helps performance. */
        if (affineeq(cycleaf, affineidentity)) {
                assert(!changed);
                /* XXX: I think this is unnecessary now. */
                debug("Debug: skipping unifycycle() (ok?)..\n");
                // unifycycle(lastset, lastedge, cyclelen);
                /* FIX: do we need to do the unifycycle() after all?? */
        }
        /* FIX: any other cases where we might want to do unifycycle()?
                is it even necessary in the first place? */
}

static void unify(set_t *s, range_t c, set_t *t);
static void unifycycle(set_t *lastset, eqn_t *lastedge, int cyclelen) {
        set_t *s, *prev;
        set_t *t=lastedge->rhs.s;
        eqn_t *e;
        set_t *cyset[cyclelen-1]; // list of sets on the cycle to unify
        intinf cyconst[cyclelen-1]; // additive constant
        int i;
        intinf quum = infzero; /* Cumulative sum of `c' around the cycle. */

        debug("Debug unifycycle(): walking the cycle...\n");
        /* Backtrace the cycle, unifying each element of the cycle. */
        for (s=lastedge->rhs.s,prev=lastset,e=lastedge,i=0;
             prev && prev != t;
             s=prev,prev=s->dfs_backptr_s,e=s->dfs_backptr_e,i++) {
                /* The edge from `prev' -> `s' by way of `e' is in the cycle,
                   so unify `prev' and `t', effectively removing `prev'. */
                affine_t edgeaf;

                assert(s && prev && e && e->rhs.s == s);
                assert(e->rhs.coeff == inf(1L) && e->op == SUBSET);
                debug(" <- %s", prev->name);
                edgeaf = edgelbl(prev, e);
                assert(edgeaf.z == infzero && edgeaf.a == inf(1L)
                       && edgeaf.b.lo == edgeaf.b.hi);
                cyconst[i] = infsub(quum, edgeaf.b.lo);
                cyset[i] = prev;
        }
        debug(".\n");
        assert(i == cyclelen-1);
        assert(prev && prev == lastedge->rhs.s && prev == t);
        assert(cyconst[cyclelen-2] == infzero);
        debug("Debug unifycycle(): unifying...\n");
        for (i=0; i<cyclelen-1; i++) {
                range_t r;
                assert(cyset[i] && cyset[i]->name);
                r.lo = r.hi = cyconst[i];
                unify(t, r, cyset[i]);
        }
        debug("Debug unifycycle(): done.\n");
}





static void dfs_relevant(set_t *s) {
        eqnlist_t *e;
        termlist_t *t;
        assert(s && s->name);
        // debug("Debug dfs: visit `%s'\n", s->name);
        s->flags |= RELEVANT;
        for (e=s->affectedby; e; e=e->next)
                for (t=e->e->lhs; t; t=t->next)
                        if (!(t->t->s->flags & RELEVANT))
                                dfs_relevant(t->t->s);
}

/* Remove stuff in the system not relevant to the QUERIED sets.
   (For each equation `S + T + ... subset U', if U is relevant
   we say that S,T,...  are also relevant.  QUERIED sets are
   trivially relevant.  Thus, a depth-first search can recover
   all the relevant sets.) */
static void removeirrelevant(void) {
        set_t *s, *prev, *next;
        debug("Debug: -- removeirrelevant() --\n");
        for (s=allsets; s; s=s->nextset)
                s->flags &= ~RELEVANT;
        for (s=allsets; s; s=s->nextset)
                if (s->flags & QUERIED && !(s->flags & RELEVANT))
                        dfs_relevant(s);
        s = allsets; prev = NULL;
        while (s) {
                next = s->nextset;
                if (!(s->flags & RELEVANT)) {
                        eqnlist_t *e, *next2;
                        debug("Debug: `%s' irrelevant, deleting.\n",
                                s->name);
                        /* Delete s from the list of all sets. */
                        if (s == allsets) {
                                allsets = next;
                        } else {
                                assert(prev && prev->nextset);
                                prev->nextset = next;
                        }

                        /* Delete all equations mentioning s. */
                        for (e=s->affectedby; e; e=next2) {
                                next2 = e->next;
                                killeqn(e->e);
                        }
                        for (e=s->affects; e; e=next2) {
                                next2 = e->next;
                                killeqn(e->e);
                        }
                        /* XXX: Should free s->affects, s->affectedby lists. */
                        assert(!s->affects && !s->affectedby);

                        /* Free s. */
                        /*  XXX: no: other stuff (e.g. aftermath()) uses this
                        free(s->name);
                        free(s);
                        */
                } else {
                        prev = s;
                }
                s = next;
        }
        if (!allsets)
                debug("Debug: no more sets left!\n");
}


/* Process an identity of the form `s + c = t',
   where s,t are sets and c is a constant term.
   Algorithm: find all references to `t' and replace them with `s+c';
   then remember the relationship between `s' and `t' with a new equation.
   (Can't handle stuff like `s + t = u', because what if `u' appears
   on the RHS of another equations??)  */
static void unify(set_t *s, range_t c, set_t *t) {
        eqnlist_t *el, *nextel;
        eqn_t tmpe;
        term_t *trm;

        assert(s && t && !rangeempty(c));
        debug("Debug: unify(%s + %s = %s)\n",
                s->name, rangestr(c), t->name);

        if (s == t) {
                /* Special case, unify(s,c,s). */
                range_t r = s->r;
                debug("Debug: unifying to self (!?)...\n");
                /* Flow to self.
                   (XXX: Only works properly when caller is visit().
                    Might break for do_unifies().) */
                if (c.lo < infzero)
                        r.lo = NEGINF;
                if (c.hi > infzero)
                        r.hi = INF;
                if (rangeinset(r, s))
                        fatal("XXX: add dependency info here.");
                for (el=t->affectedby; el; el=nextel) {
                        assert(el && el->e && el->e->rhs.s == s);
                        nextel = el->next;
                        if (el->e->op != EQUALS)
                                continue;
                        assert(el->e->lhs && el->e->lhs->t);
                        assert(!el->e->lhs->next);
                        assert(el->e->lhs->t->coeff == inf(1L));
                        assert(el->e->rhs.coeff == inf(1L));
                        if (el->e->lhs->t->s != s)
                                continue;
                        debug("Debug: removing %s\n", debug_dumpeqn(el->e));
                        killeqn(el->e);
                }
                if (!t->affectedby) {
                        debug("Debug:  nothing affects %s, finalizing.\n",
                                t->name);
                        finalize(t);
                }
                return;
        }

        /* Find all equations with `t' on the LHS, and replace `t' with `s'. */
        for (el=t->affects; el; el=nextel) {
                termlist_t *tl, *nexttl;
                eqn_t *e = el->e;
                assert(e);
                nextel = el->next;
                assert(e->op == SUBSET || e->op == EQUALS);
                /*
                debug("Debug: unify() doing subst. %s -> %s in `%s'.\n",
                        t->name, s->name, debug_dumpeqn(e));
                */
                for (tl=e->lhs; tl; tl=nexttl) {
                        term_t *trm;
                        assert(tl->t && tl->t->s);
                        nexttl = tl->next;
                        if (tl->t->s != t)
                                continue;
                        trm = mymalloc(sizeof(term_t));
                        trm->coeff = tl->t->coeff;
                        trm->s = s;
                        e->lhs = delterm(e->lhs, tl->t);
                        t->affects = deleqn(t->affects, e);
                        e->lhs = insterm(e->lhs, trm);
                        s->affects = inseqn(s->affects, e);
                        assert(trm->coeff != infzero);
                        e->c = rangeadd(e->c, rangemul(trm->coeff, c));
                        if (isdup(e)) {
                                /* We made e a duplicate.  Delete it. */
                                killeqn(e);
                        }
                           
                        break;
                }
                assert(tl);
        }
        /* Should have eliminated all equations with `t' on LHS. */
        assert(!t->affects);

        /* Find all equations with `t' on the RHS, and substitute. */
        // debug("Debug: unify() doing RHS...\n");
        for (el=t->affectedby; el; el=nextel) {
                eqn_t *e = el->e;
                assert(e && e->rhs.s == t);
                nextel = el->next;
                assert(e->op == SUBSET || e->op == EQUALS);
                if (e->op == EQUALS && e->lhs && !e->lhs->next
                    && e->lhs->t->s == s) {
                        /* We're about to create an equation of the
                           form `s + c = s'.  Don't do it!  */
                        assert(e->lhs->t->coeff == inf(1L));
                        assert(e->rhs.coeff == inf(1L));
                        assert(e->rhs.s == t);
                        assert(e->lhs->t->s && e->rhs.s);
                        assert(e->c.lo == c.lo && e->c.hi == c.hi);
                        debug("Debug: removing %s\n", debug_dumpeqn(e));
                        killeqn(e);
                        continue;
                }
                e->rhs.s = s;
                assert(e->rhs.s && e->rhs.s->name);
                assert(e->rhs.coeff == inf(1L));
                e->c = rangesub(e->c, c);
                t->affectedby = deleqn(t->affectedby, e);
                s->affectedby = inseqn(s->affectedby, e);
                if (isdup(e)) {
                        /* We made e a duplicate.  Delete it. */
                        killeqn(e);
                        continue;
                }
        }
        /* Should have eliminated all equations with `t' on RHS. */
        assert(!t->affectedby);

        /* Flow from `t' to `s'. */
        if (rangeinset(rangesub(t->r,c), s))
                overwritedepends_set(s, t);

        /* Add a new equation `s + c subset t'.
           (Note that we do _not_ add `s + c = t'.
            This is safe, since now `t' only appears in one equation,
            namely `s + c = t', so `t' won't change again except
            due to `s''s effect, and we've already flowed from
            `t' to `s' once so the `t' -> `s' dependence is
            already accounted for.) */
        tmpe.c = c;
        trm = mymalloc(sizeof(term_t));
        trm->s = s; trm->coeff = inf(1L);
        tmpe.lhs = insterm(NULL, trm);
        assert(t && t->name);
        tmpe.rhs.s = t; tmpe.rhs.coeff = inf(1L);
        tmpe.op = SUBSET;
        tmpe.eflags = 0;
        tmpe.extra_depends = NULL;
        if (addeqn_unlessdup(&tmpe))
                s->flags |= SURVIVES;
        else
                free(trm);
        assert(t->affectedby);
}



#define SETTBLSIZ 50021
static set_t *settbl[SETTBLSIZ];


static void initsettbl(void) {
        int i;
        for (i=0; i<SETTBLSIZ; i++)
                settbl[i] = (set_t *)NULL;
}

static unsigned int hash(unsigned char *name, unsigned int salt) {
        unsigned int hash = (salt*salt) % 15000017;
        while (*name)
                hash = (((hash<<8) + (*name++)) ^ salt) % 15000017;
        return hash;
}


static set_t *newset(char *name) {
        set_t *s = mymalloc(sizeof(set_t));

        if (!isalpha(*name))
                fatal("Variable-name `%s' didn't start with a letter.", name);
        else if (strchr(name, ' '))
                fatal("Variable-name `%s' contained a space.", name);

        s->r.lo = INF;
        s->r.hi = NEGINF;
        s->hardbound.lo = NEGINF;
        s->hardbound.hi = INF;
        s->flags = 0;
        s->name = strdup(name);
        s->depends[0] = NULL;

        /* Insert s in the list of all sets. */
        s->nextset = allsets;
        allsets = s;

        /* Fill in s->affects, s->affectedby later. */
        s->affects = s->affectedby = NULL;
        return s;
}

static set_t *getset(char *name) {
        set_t *p;
        int i, h;
        for (i=0; i<100; i++) {
                h = hash(name, i) % SETTBLSIZ;
                p = settbl[h];
                assert(!p || p->name);
                if (p && p->name && strcmp(p->name, name) == 0)
                        return p;
                if (!p) {
                        p = newset(name);
                        settbl[h] = p;
                        return p;
                }
        }
        fatal("Hash table too full (name=`%s').", name);
}

/* Reads a term of the form `var' or `5*var' (but not `5'!). */
static void readterm(term_t *t, char *str) {
        char *p = strchr(str, '*'), *q;
        t->coeff = inf(1L);
        if (p) {
                *p++ = '\0';
                t->coeff = inf(strtol(str, &q, 10));
                if (q != p)
                        fatal("Couldn't parse coefficient of `%s*%s'.", str, p);
                str = p;
        }
        t->s = getset(str);
}

/* Reads the left-hand-side of an equation. */
static void readleft(char *str, eqn_t *e) {
        char *p, *q;
        intinf n;
        assert(str && e && !e->lhs && !e->c.lo && !e->c.hi);
        assert(e->op == SUBSET || e->op == EQUALS);
        while (str) {
                /* Separate out the current term, and remember where it ends. */
                p = strstr(str, " + ");
                if (p) {
                        *p = '\0'; p = p + strlen(" + ");
                }

                /* Parse the term. */
                n = inf(strtol(str, &q, 10));
                if (*q == '\0') {
                        e->c = rangeadd(e->c, rangesingleton(n));
                } else if (strcmp(q, "top") == 0) {
                        e->c.lo = NEGINF;
                        e->c.hi = INF;
                } else {
                        term_t *t = mymalloc(sizeof(term_t));
                        t->coeff = inf(1L);
                        if (*q == '*') {
                                t->coeff = n; q++;
                        } else if (q != str)
                                fatal("Trouble parsing term `%s'.", str);
                        t->s = getset(q);
                        if (t->coeff == infzero)
                                debug("readleft() saw 0*%s.\n", t->s->name);
                        e->lhs = insterm(e->lhs, t);
                }

                /* Advance to the next term. */
                str = p;
        }
}

static void readquery(char *line) {
        set_t *s = getset(line);
        assert(s);
        s->flags |= QUERIED;
}

static void readeqn(char *line) {
        char *p = NULL;
        eqn_t *e = NULL;
        if (strstr(line, "query ") == line) {
                readquery(line + strlen("query "));
                return;
        } else if (strstr(line, "min ") == line) {
                char *q;
                q = line + strlen("min ");
                if ((p = strstr(q, " subset ")) == NULL)
                        fatal("Trouble parsing `%s' (no subset).", line);
                e = neweqn();
                e->op = MIN_SUBSET;
                *p = '\0';
                p += strlen(" subset ");
                readterm(&(e->rhs), p);
                assert(e->rhs.s && e->rhs.s->name);
                e->lhs = NULL;
                e->c = rangesingleton(INF);
                while (q && *q) {
                        char *r;
                        intinf n;
                        p = strstr(q, ", ");
                        if (p)
                                *p = '\0';
                        n = inf(strtol(q, &r, 10));
                        if (*r == '\0') {
                                e->c = rangemin(e->c, rangesingleton(n));
                        } else {
                                term_t *t = mymalloc(sizeof(term_t));
                                termlist_t *tl;
                                readterm(t, q);
                                assert(t && t->s && t->s->name);
                                /* insterm() doesn't work exactly right
                                   for the MIN operation, when you have
                                   coeff = 0. */
                                assert(t->coeff != infzero);
                                for (tl=e->lhs; tl; tl=tl->next) {
                                        /* insterm() doesn't work exactly
                                           right for the MIN operation, when
                                           you have duplicate sets on lhs. */
                                        assert(tl->t->s != t->s);
                                }
                                e->lhs = insterm(e->lhs, t);
                        }
                        q = p ? p + strlen(", ") : NULL;
                }
                return;
        } else if ((p = strstr(line, " subset ")) != NULL) {
                e = neweqn();
                e->op = SUBSET;
                *p = '\0';
                p += strlen(" subset ");
        } else if ((p = strstr(line, " = ")) != NULL) {
                intinf n;
                char *q;
                *p = '\0';
                p += strlen(" = ");
                n = inf(strtol(p, &q, 10));
                if (*q == '\0') {
                        set_t *s = getset(line);
                        s->r = s->hardbound = rangesingleton(n);
                        return;
                } else if (q != p && *q != '*')
                        fatal("Trouble parsing `%s = %s'.", line, p);
                e = neweqn();
                e->op = EQUALS;
        } else if ((p = strstr(line, " >= ")) != NULL) {
                intinf lo;
                term_t t;
                char *q;
                *p = '\0';
                p += strlen(" >= ");
                lo = inf(strtol(p, &q, 10));
                if (*q != '\0')
                        fatal("Couldn't parse RHS of `%s >= %s'; "
                              "not an integer constant.", line, p);
                readterm(&t, line);
                if (t.coeff != inf(1L) || !(t.s))
                        fatal("Couldn't parse LHS of `%s >= %s'; "
                              "not a set variable.", line, p);
                if (lo > t.s->hardbound.lo)
                        t.s->hardbound.lo = lo;
                return;
        } else
                fatal("Trouble parsing eqn `%s'.", line);

        readleft(line, e);
        readterm(&(e->rhs), p);

        if (e->rhs.coeff != inf(1L))
                fatal("Coeff != 1 in `%s subset %s'.", line, p);
}

/* Given a system of equations (set constraints),
   eliminate all the EQUALS constraints, by unifying
   set variables if possible, or else by doing substitutions. */
static void do_unifies(void) {
        int i;
        debug("Debug: -- do_unifies() --\n");
        for (i=0; i<neqns; i++) {
                eqn_t *e = &eqns[i];
                if (e->op != EQUALS || e->eflags & DELETED)
                        continue;
                assert(e && e->lhs && e->lhs->t && e->lhs->t->s && e->rhs.s);
                assert(e->rhs.coeff == inf(1L) && e->lhs->t->coeff == inf(1L));
                assert(!e->lhs->next);
                assert(e->lhs->t->s != e->rhs.s); // XXX: avoid bug in unify()
                unify(e->lhs->t->s, e->c, e->rhs.s);
        }
}

static int ready_to_check_graph = 0;
static void debug_check_graph(void) {
        eqnlist_t *l;
        set_t *s;
        int i;
        if (!ready_to_check_graph) {
                debug("Debug:   Not ready to check graph, skipping.\n");
                return;
        }
        for (i=0; i<neqns; i++) {
                eqn_t *e = &eqns[i];
                termlist_t *t;
                assert(e);
                if ((e->eflags & DELETED) || !e->lhs)
                        continue;
                assert(e->rhs.s && e->rhs.s->name);
                for (l=e->rhs.s->affectedby; l; l=l->next) {
                        if (l->e == e)
                                break;
                }
                assert(e == &eqns[i]);
                if (!l) {
                        fprintf(stderr, "Eqn `%s' not in `%s'->affectedby.\n",
                                debug_dumpeqn(e), e->rhs.s->name);
                        assert(0);
                }
                for (t=e->lhs; t; t=t->next) {
                        assert(t->t && t->t->s && t->t->s->name);
                        assert(t->t->coeff != infzero);
                        s = t->t->s;
                        for (l=s->affects; l; l=l->next) {
                                if (l->e == e)
                                        break;
                        }
                        if (!l) {
                                fprintf(stderr,
                                  "Eqn `%s' not in `%s'->affects.\n",
                                  debug_dumpeqn(e), s->name);
                                assert(0);
                        }
                }
        }
        for (s=allsets; s; s=s->nextset) {
                assert(s && s->name);
                for (l=s->affectedby; l; l=l->next) {
                        assert(l->e && l->e->rhs.s);
                        if (l->e->rhs.s != s) {
                                fprintf(stderr,
                                   "`%s'-affectedby includes eqn `%s'.\n",
                                   s->name, debug_dumpeqn(l->e));
                                assert(0);
                        }
                }
                for (l=s->affects; l; l=l->next) {
                        termlist_t *t;
                        assert(l->e);
                        for (t=l->e->lhs; t; t=t->next) {
                                assert(t->t && t->t->s && t->t->s->name);
                                if (t->t->s == s)
                                        break;
                        }
                        if (!t) {
                                fprintf(stderr,
                                   "`%s'-affects includes eqn `%s'.\n",
                                   s->name, debug_dumpeqn(l->e));
                                assert(0);
                        }
                }
        }
        //debug("Debug:    Graph is ok.\n");
}

/* Initialize the s->affects, s->affectedby lists. */
static void seteffects(void) {
        termlist_t *p;
        int i;
        debug("Debug: -- seteffects() --\n");
        for (i=0; i<neqns; i++) {
                assert(eqns[i].rhs.s && !(eqns[i].eflags & DELETED));
                eqns[i].rhs.s->affectedby =
                    inseqn(eqns[i].rhs.s->affectedby, &eqns[i]);

                for (p=eqns[i].lhs; p; p=p->next) {
                        assert(p->t && p->t->s);
                        p->t->s->affects = inseqn(p->t->s->affects, &eqns[i]);
                }
        }
        ready_to_check_graph = 1;
}

/* Process equations which would otherwise have been overlooked. */
static void orphans(void) {
        set_t *s;
        int i;

        debug("Debug: -- orphans() --\n");
        for (s=allsets; s; s=s->nextset)
                if (s->affectedby == NULL && !(s->flags & FINALIZED))
                        finalize(s);
        for (i=0; i<neqns; i++) {
                if (!eqns[i].lhs && !(eqns[i].eflags & DELETED)) {
                        /* This equation doesn't appear in any affectedby
                           list, so if we don't do it now, it'll never get
                           considered. */
                        /* XXX: is it ok that eqns[i].rhs might have been
                           already marked FINALIZED? */
                        eqn_t *e = &eqns[i];
                        assert(e && e->rhs.s);
                        if (rangeinset(e->c, e->rhs.s))
                                overwritedepends(e->rhs.s, e);
                        (void)prune(e);
                }
        }
}

/* Get ready to start flowing by initializing the workqueue to contain
   all the FINALIZED sets. */
static void touch_finalized(void) {
        set_t *s;
        debug("Debug: -- touch_finalized() --\n");
        for (s=allsets; s; s=s->nextset)
                if (s->flags & FINALIZED) {
                        /* Simplify (and maybe prune?) other eqns. */
                        s->flags |= SURVIVES;
                }
}

/* Find equations `LHS + c subset RHS' where c is an empty range:
   this seems highly unnatural! */
static void findemptyc(void) {
        int i;
        debug("Debug: -- findemptyc() --\n");
        for (i=0; i<neqns; i++) {
                eqn_t *e = &eqns[i];
                if (e->op == MIN_SUBSET)
                        continue;
                if (rangeempty(e->c) && (debugging || !(e->eflags & DELETED))) {
                        warn("Warning: empty constant term?: %s\n",
                                debug_dumpeqn(e));
                }
        }
}

static void readsystem(void) {
        char line[1024], *p;
        while (fgets(line, sizeof(line), stdin)) {
                p = strchr(line, '\n');
                if (p)
                        *p = '\0';
                if (*line == '\0')
                        continue;
                readeqn(line);
        }

        /* Order is very important here. */
        seteffects();
        debug_check_graph();
        removeirrelevant();
        findemptyc();
        do_unifies();
        debug_check_graph();
        orphans();
        touch_finalized();
        debug_check_graph();
}

/* Check whether the current solution respects the constraints. */
static int isasolution(void) {
        int i;
        set_t *s;
        for (s=allsets; s; s=s->nextset) {
                if (!rangeempty(s->r)
                         && (s->hardbound.lo > s->r.lo
                             || s->r.lo > s->r.hi
                             || s->r.hi > s->hardbound.hi) ) {
                        debug("Debug: `%s'->r = %s, ", s->name, rangestr(s->r));
                        debug("`%s'->hardbound = %s.\n",
                                s->name, rangestr(s->hardbound));
                        return 0;
                }
        }
        for (i=0; i<neqns; i++) {
                eqn_t *e = &eqns[i];
                termlist_t *p;
                range_t lhs = e->c;
                range_t rhs, r;
                if (e->eflags & DELETED)
                        continue;
                assert(e->rhs.s);
                rhs = e->rhs.s->r;
                assert(e->rhs.coeff == inf(1L));
                assert(e->op == SUBSET || e->op == MIN_SUBSET
                    || e->op == EQUALS);
                for (p=e->lhs; p; p=p->next) {
                        assert(p->t->coeff != infzero);
                        r = rangemul(p->t->coeff, p->t->s->r);
                        if (e->op == MIN_SUBSET)
                                lhs = rangemin(lhs, r);
                        else
                                lhs = rangeadd(lhs, r);
                }
                if (((lhs.lo != rhs.lo || lhs.hi != rhs.hi) && e->op == EQUALS)
                 || ((rhs.lo > lhs.lo || rhs.hi < lhs.hi)
                             && (e->op == SUBSET || e->op == MIN_SUBSET))) {
                        debug("Debug: eqn `%s' violated!\n", debug_dumpeqn(e));
                        debug("Debug: lhs=%s, ", rangestr(lhs));
                        debug("rhs=%s.\n", rangestr(rhs));
                        return 0;
                }
        }
        return 1;
}

static void aftermath(void) {
        set_t *s;

        debug_check_graph();
        if (!isasolution())
                fatal("Putative solution didn't check out.");
        printf("Feasible.\n");
        for (s=allsets; s; s=s->nextset) {
                if (s->flags & QUERIED) {
                        int i;
                        printf("%s = %s",  s->name, rangestr(s->r));
                        if (s->depends[0]) {
                                printf(" <-");
                                for (i=0; i<MAXDEPENDS && s->depends[i]; i++)
                                        printf(" %s", s->depends[i]->name);
                        }
                        printf("\n");
                }
        }
}

static void showstatus(int iter) {
        int nsets = 0, nlivesets = 0, nliveeqns = 0, i;
        set_t *s;
        for (s=allsets; s; nsets++,s=s->nextset)
                if (!(s->flags & FINALIZED)) {
                        assert(s->flags & RELEVANT);
                        /*
                        if ((s->flags & (ALIVE|SURVIVES)) == 0)
                                debug("Debug:  %s->flags = %X\n",
                                                s->name, s->flags);
                        */
                        nlivesets++;
                }
        for (i=0; i<neqns; i++)
                if (!(eqns[i].eflags & DELETED))
                        nliveeqns++;
        debug("Debug: iter %d: (%d of %d sets / %d of %d eqns still live)\n",
                iter, nlivesets, nsets, nliveeqns, neqns);
}

static void run(void) {
        int iter;
        debug("-- Let's run!\n");
        iter = 0;
        do {
                debug_check_graph();
                showstatus(++iter);
        } while (findcycles());
}

static void again(void) {
        set_t *s;
        debug("Debug: again(), touching");
        for (s=allsets; s; s=s->nextset)
                if (!(s->flags & FINALIZED)) {
                        debug(" `%s'", s->name);
                        s->flags |= SURVIVES;
                }
        debug(".\n");
}

#ifndef MINLONG
#ifdef LONG_MIN
#define MINLONG LONG_MIN
#else
#define MINLONG (-MAXLONG - 1L)
#endif
#endif


int main(int argc, char **argv) {
        if (argc > 1 && strcmp(argv[1], "-v")==0)
                debugging = 1;

        /* For infmul() */
        if (INF*INF > MAXLONG)
                fatal("Decrease INF.");
        else if (NEGINF*NEGINF < MINLONG)
                fatal("Increase NEGINF.");

        initsettbl();
        affinetest();


        readsystem();
        run();
#if 0
        debug("--- Debug: is%s a solution, before again(). ---\n",
                (isasolution() ? "" : " NOT"));
        again(); run(); /* just to double-check; shouldn't be necessary */
#endif
        aftermath();
        return 0;
}
