/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         decl.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of types / declarations.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#ifndef    DECL_H
#define    DECL_H

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "tree.h"
#include "symbol.h"
#include "callback.h"
#include "location.h"

class Constant;
class Expression;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

typedef unsigned long BaseTypeSpec;

const BaseTypeSpec BT_NoType       = 0x000000;  // no type provided
const BaseTypeSpec BT_Void         = 0x000001;  // explicitly no type
const BaseTypeSpec BT_Char         = 0x000002;
const BaseTypeSpec BT_Short        = 0x000004;
const BaseTypeSpec BT_Int          = 0x000008;
const BaseTypeSpec BT_Long         = 0x000010;
const BaseTypeSpec BT_LongLong     = 0x000020;  // a likely C9X addition
const BaseTypeSpec BT_Float        = 0x000040;
const BaseTypeSpec BT_Double       = 0x000080;
const BaseTypeSpec BT_Ellipsis     = 0x000100;

const BaseTypeSpec BT_Struct       = 0x000200;
const BaseTypeSpec BT_Union        = 0x000400;
const BaseTypeSpec BT_Enum         = 0x000800;
const BaseTypeSpec BT_TypeDef      = 0x001000;
const BaseTypeSpec BT_UserType     = 0x002000;
const BaseTypeSpec BT_BaseMask     = 0x002FFF;

// Sign indicator
const BaseTypeSpec BT_Signed       = 0x010000;
const BaseTypeSpec BT_UnSigned     = 0x020000;
const BaseTypeSpec BT_SignMask     = 0x030000;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
typedef unsigned long TypeQual;

const TypeQual TQ_None         = 0x0000;
const TypeQual TQ_Const        = 0x0001;
const TypeQual TQ_Volatile     = 0x0002;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum StorageType
{
    ST_None=0,
    ST_Auto,
    ST_Register,
    ST_Static,
    ST_Extern
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum TypeType
{
    TT_Base,        // a simple base type, T
    TT_Pointer,     // pointer to T
    TT_Array,       // an array of T
    TT_BitField,    // a bitfield
    TT_Function,    // <args> -> <result>

    TT_Struct,     
    TT_Union,
    TT_Enum
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Type
{
  public:
    Type(TypeType _type=TT_Base);
    virtual ~Type();

    virtual int     precedence() const { return 16; }
    virtual Type*   dup() const =0;    // deep-copy

    virtual void extend(Type *extension) { assert(0); }
    virtual void print(ostream& out, Symbol *name, int level) const;
    virtual void findExpr( fnExprCallback cb ) {}

    TypeType        type;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Decl
{
  public:
    Decl();
    ~Decl();

    bool    isTypedef() const { return (baseType & BT_TypeDef) != 0; }
    bool    isStatic() const { return (storage == ST_Static); }

    void    clear();

    Decl*   halfdup() const;    // only the basic stuff.
    Decl*   dup() const;        // deep-copy

    void    copy(const Decl& decl);    // shallow copy

    void    print(ostream& out, bool showBase, int level=0) const;

    void    findExpr( fnExprCallback cb );


    StorageType     storage;
    TypeQual        qualifier;
    BaseTypeSpec    baseType;

    Type            *form;
    Symbol          *name;
    Symbol          *tag;        // tag for struct/union/enum

    Symbol          *typeName;   // typedef name for a UserType

    Expression      *initializer;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class PtrType : public Type
{
  public:
    PtrType();
    ~PtrType();

    int    precedence() const { return 15; }

    Type*  dup() const;    // deep-copy

    void extend(Type *extension);
    void print(ostream& out, Symbol *name, int level) const;

    void findExpr( fnExprCallback cb );

    Type           *subType;
    TypeQual        qualifier;
};

class ArrayType : public Type
{
  public:
    ArrayType();
    ~ArrayType();

    Type*   dup() const;    // deep-copy

    void extend(Type *extension);
    void print(ostream& out, Symbol *name, int level) const;

    void findExpr( fnExprCallback cb );

    Type           *subType;

    Expression     *size;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class BitFieldType : public Type
{
  public:
    BitFieldType();
    ~BitFieldType();

    Type*   dup() const;    // deep-copy

    void print(ostream& out, Symbol *name, int level) const;
    void findExpr( fnExprCallback cb );

    Expression     *size;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class FunctionType : public Type
{
  public:
    FunctionType();
    ~FunctionType();

    Type*   dup() const;    // deep-copy

    void extend(Type *extension);
    void print(ostream& out, Symbol *name, int level) const;

    void addArg(Decl *arg);
    void getParams(treeNode *node);

    void fixParam(Decl *arg);

    void findExpr( fnExprCallback cb );

    int             nArgs;
    int             size;
    Decl          **args;

    Type           *subType;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class StructType : public Type
{
  public:
    StructType();
    ~StructType();

    bool    isUnion() const { return (type == TT_Union); }

    Type*   dup() const;    // deep-copy

    void print(ostream& out, Symbol *name, int level) const;

    void findExpr( fnExprCallback cb );

    void addComponent(Decl *comp);
    void getStructComponents(treeNode *node);

    Decl* findComponent( Symbol *fieldName );

    Symbol         *tag;

    int             size;          // size of the array.
    int             nComponents;
    Decl          **components;

  private:
    Decl   *lastDecl();
};

class EnumType : public Type
{
  public:
    EnumType();
    ~EnumType();

    Type*   dup() const;    // deep-copy

    void print(ostream& out, Symbol *name, int level) const;
    void findExpr( fnExprCallback cb );

    void addElement( Symbol *nme, Expression *val = NULL );
    void getEnumElements(treeNode *node);

    Symbol         *tag;

    int             size;       // size of the arrays.
    int             nElements;
    Symbol        **names;
    Expression    **values;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

void         buildBaseType(Decl *decl, treeNode *node);

Type        *buildType(treeNode *node, Symbol*& typeName);
Decl        *buildDecl(Decl *decl, treeNode *node);

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif  /* DECL_H */

