/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         symbol.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 15, 1998
    o+
    o+     A symbol class.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#ifndef    SYMBOL_H
#define    SYMBOL_H

#include <stdlib.h>
#include <iostream.h>

#include "tree.h"
#include "nametbl.h"

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o 

    SymTab:  A tree of scope tables,
             one for each scope that declares something, or has a
             child scope that declares something:

         Level
           1              external
                         /        \
           2          file        file 
                      scope       scope 
                                /       \
           3              prototype    function
                            scope        scope
                                        /     \
           4                         block    block
                                     scope    scope    
                                                 \
           5                                     block
                                                 scope    
                                                    \
                                                   (etc.)

    At any particular point you can see all the symbols 
    (variables, types, functions, etc) declared above you in the tree.

    The scope tables are recreated in a lazy fashion, so entering
    and exiting scopes that don't declare new symbols is cheap.

    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
   

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

/* Different kinds of entries in the symbol table. */

enum SymEntryType
{
    /* These things are all in the same namespace. */
    TypedefEntry,          /* type definition. */
    FuncdefEntry,          /* function definition. */
    VarDeclEntry,          /* variable declaration. */
    EnumConstEntry,        /* enum constant. */

    /* These things are in seperate namespaces. */
    LabelEntry,            /* label definition. */
    TagEntry,              /* struct/union/enum tags */

    /* There is a separate namespace for each struct/union. */
    ComponentEntry         /* components of a struct/union. */
};

/* These might be convienent. */
const int  CURRENT_SCOPE  = 0;    /* the default */
const int  EXTERN_SCOPE   = 1;
const int  FILE_SCOPE     = 2;
const int  FUNCTION_SCOPE = 3;
const int  BLOCK_SCOPE    = 4;    /* really, 4 or more. */

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

class StructType;
class Decl;
class Expression;
class Statement;

class SymEntry
{
  public:
    SymEntry(SymEntryType _type);
    SymEntry(SymEntryType _type, StrItem *sym, treeNode *tn);
   ~SymEntry();

    bool    IsTypeDef() const { return (type == TypedefEntry); }
    bool    IsFuncDef() const { return (type == FuncdefEntry); }
    bool    IsVarDecl() const { return (type == VarDeclEntry); }
    bool    IsEnumConst() const { return (type == EnumConstEntry); }

    SymEntryType    SymType() { return type; }

    void    Show(ostream& out) const;

  public:
    SymEntryType    type;    /* what kind of entry this is. */
    StrItem        *name;    /* The name of the thing. */

    // Pointer to it's definition/declaration.
    union
    {
        Statement*      uFuncDef;
        Decl*           uVarDecl;        /* Also used by typedef */
        Expression*     uEnumValue;

        Statement*      uLabelPosition;

        Decl*           uStructDef;
        Decl*           uComponent;
    };

    /* ptr to it's definition/declaration. (old-style) */
    treeNode       *def;    

    /* For a struct/union component, this points to the definition
       of the container. */
    StructType*     Container;        /* Todo */

    /* This would probably be a good place to add some attributes, 
       but none come to mind at the moment. */

    SymEntry       *next;
};

SymEntry  *mk_typedef   (StrItem *sym, treeNode *);
SymEntry  *mk_funcdef   (StrItem *sym, treeNode *);
SymEntry  *mk_vardecl   (StrItem *sym, treeNode *);
SymEntry  *mk_enum_const(StrItem *sym, treeNode *);
SymEntry  *mk_label     (StrItem *sym, treeNode *);
SymEntry  *mk_tag       (StrItem *sym, treeNode *);
SymEntry  *mk_component (StrItem *sym, treeNode *);

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#define    INIT_HASHTAB_SIZE    (5)

class HashTbl
{
  public:
    HashTbl(int initSize = INIT_HASHTAB_SIZE);
   ~HashTbl();

    SymEntry   *Lookup( StrItem* );
    SymEntry   *Insert( SymEntry* );

    void Show(ostream& out) const;

  public:
    int            tsize;     /* The current size of the table. */
    int            nent;      /* The number of entries being stored. */
    SymEntry     **tab;       /* The table. */
};

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#define    INIT_CHILD_SIZE     4

class ScopeTbl
{
  public:
    ScopeTbl( ScopeTbl *mom = NULL, int initSize = INIT_CHILD_SIZE );
   ~ScopeTbl();

    SymEntry   *Lookup( StrItem* );
    SymEntry   *Find( StrItem* );
    SymEntry   *Insert( SymEntry* );

    void        Show(ostream& out) const;

    void        PostOrderTblDelete();
    bool        ChildInsert(ScopeTbl *kid);

  public:
    int               nsyms; /* The num of syms declared at this scope. */
    int               level; /* This ScopeTbl's scoping level. */

    HashTbl          *htab;  /* A hash table - to store the symbols. */

    ScopeTbl         *parent; /* The scope enclosing us, if any. */

    /* A doubling array of scopes we enclose. */
    int               nchild;
    int               size;
    ScopeTbl        **children;
};

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

class SymTbl
{
  public:
    SymTbl();
   ~SymTbl();

    SymEntry *Lookup( StrItem* );
    SymEntry *LookupAt( StrItem *, int level );
    SymEntry *Insert( SymEntry* );
    SymEntry *InsertUp( SymEntry* );
    bool      InsertAt( SymEntry*, int level );

    int       EnterScope();
    void      ExitScope();

    void      Show(ostream& out) const;

  public:
    ScopeTbl        *root;    /* The top scopetab - external scope. */

    int              clevel;  /* The current level. */
    ScopeTbl        *current; /* The current scopetab, or one of its
                                 ancestors, if it doesn't exist yet. */
};


/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

class Context
{
  public:
    Context();
   ~Context();

    int     EnterScope();
    void    ExitScope();
    void    ExitScopes( int newlev );

  public:
    SymTbl     *labels;    /* Statement labels. */
    SymTbl     *tags;      /* Struct/Union/Enum tags. */
    SymTbl     *syms;      /* Vars, Types, Functions, etc. */
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Symbol
{
  public:
    Symbol();
   ~Symbol();

    Symbol *dup() const;

    uint    hash() const { return name->hash(); }

    StrItem      *name;
    SymEntry     *entry;

  friend ostream& operator<< ( ostream& out, const Symbol& sym );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

Symbol *buildSymbol(leafNode *leaf);

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif  /* SYMBOL_H */

