(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

functor SymTab(structure IdKey : ORD_KEY
	       type info
	       type scopeInfo
	       val checkInsert : IdKey.ord_key->{old : info option,
						 new :info}->
		   scopeInfo->bool) : SYMTAB =
    struct

	(*************************************************************

	 Types and exceptions

	 *************************************************************)

	type identifier = IdKey.ord_key
	type info = info
	type scopeInfo = scopeInfo

	structure IdMap = BinaryMapFn(IdKey)

	datatype scope = Scope of {parent : scope option,
				   symbols : info IdMap.map ref,
				   info : scopeInfo}
	(* A scope that's not updatable *)
	  | FixedScope of {parent : scope option,
			   symbols : info IdMap.map,
			   info : scopeInfo}

	datatype symbolTable = ST of {currentScope : scope option ref}

	exception ScopeNotFound    (* There is no current scope *)

	(*************************************************************

	 Functions

	 *************************************************************)

	(* Make a new symbol table *)
	fun mkSymTab () = ST {currentScope = ref NONE}

	(* Reset a symbol table *)
	fun resetSymTab (ST {currentScope}) = currentScope := NONE

	(* Enter a new scope *)
	fun enterScope (ST {currentScope}) inf =
	    let val newScope = Scope {parent = !currentScope,
				      symbols = ref IdMap.empty,
				      info = inf}
	    in
		currentScope := SOME newScope
	    end

	(* Leave the current scope *)
	fun exitScope (ST {currentScope = ref NONE}) = raise ScopeNotFound
	  | exitScope (ST {currentScope}) =
	    let val (Scope {parent, ...}) = valOf (!currentScope)
	    in
		currentScope := parent
	    end

	(* Get the current scope *)
	fun getScope (ST {currentScope = ref (SOME s)}) = s
	  | getScope _ = raise ScopeNotFound

	(* Get the current scope, but make all levels except the root
	   non-updatable. *)
	fun getCScope (ST {currentScope = ref (SOME s)}) =
	    let fun toFixedScope (Scope {parent=SOME p, symbols, info}) =
		FixedScope {parent=SOME (toFixedScope p),
			    symbols=(!symbols),
			    info=info}
		  | toFixedScope s = s
	    in
		toFixedScope s
	    end

	(* Set the current scope *)
	fun setScope (ST {currentScope}) s =
	    currentScope := SOME s

	(* Get the information associated with a scope *)
	fun getScopeInfo (Scope {info,...}) = info
	  | getScopeInfo (FixedScope {info,...}) = info

	(* Given a scope, return its parent *)
	fun parentScope (Scope {parent, ...}) = parent
	  | parentScope (FixedScope {parent, ...}) = parent

	(* Apply a function to all the information in a scope *)
	fun mapScope f (Scope {parent, symbols, info}) =
	    Scope {parent=parent,
		   symbols=ref (IdMap.map f (!symbols)),
		   info=info}

	(* Given a scope, return the number of scopes above it
	   in the symbol table *)
	fun scopeDepth (Scope {parent=NONE, ...}) = 0
	  | scopeDepth (Scope {parent=SOME p, ...}) = 1 + (scopeDepth p)
	  | scopeDepth (FixedScope {parent=NONE, ...}) = 0
	  | scopeDepth (FixedScope {parent=SOME p, ...}) = 1 + (scopeDepth p)

	(* Return true iff the scope is the outermost (global) scope *)
	fun isGlobalScope (Scope {parent=NONE, ...}) = true
	  | isGlobalScope (FixedScope {parent=NONE, ...}) = true
	  | isGlobalScope _ = false

	(* Insert a symbol *)
	fun insert (ST {currentScope = ref
			(SOME (Scope {symbols,info,...}))}) id inf =
	    if checkInsert id
		          {old=(IdMap.find(!symbols, id)),
			   new=inf}
			  info then
		symbols := IdMap.insert(!symbols, id, inf)
	    else
		()
	  | insert _ _ _ = raise ScopeNotFound

	(* Lookup a symbol in a particular scope without changing
	 the current scope. *)
	fun findInScope (s as FixedScope {symbols, parent, ...}) id =
	    (case IdMap.find(symbols, id) of
		 SOME i => SOME {info=i, scope=s}
	       | NONE => (case parent of
			      SOME p => findInScope p id
			    | NONE => NONE))
	  | findInScope (s as Scope {symbols, parent, info}) id =
	     (case IdMap.find(!symbols, id) of
		  SOME i => SOME {info=i, scope=s}
		| NONE => (case parent of
			       SOME p => findInScope p id
			     | NONE => NONE))

	(* Lookup a symbol *)
	fun find (ST {currentScope = ref scope}) id =
	    case scope of
		SOME sc => findInScope sc id
	      | NONE => NONE
    end
