(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

signature SymbolTable =
    sig
	structure AST : AST
	val insertId : AST.identifier -> AST.syntacticType -> unit
	val insertStruct : AST.identifier -> AST.syntacticType -> unit
	val insertEnum : AST.identifier -> AST.syntacticType -> unit
	val getIdType : AST.identifier -> AST.syntacticType option
	val idIsTypedefName : AST.identifier -> bool
	val pushScope : unit -> unit
	val pushStructScope : unit -> unit
	val popScope : unit -> unit
	(* push last popped scope back on along with next new scope *)
	val recoverParameterScopeOnNextPush : unit -> unit
	val reset : unit -> unit
	    
	type scope
	val getScope : unit -> scope
	val scopeLookup : scope -> AST.lookupFun
	val mkLookupFun : unit -> AST.lookupFun
    end

functor SymbolTableFn(structure AST : AST) : SymbolTable =
    struct

	structure AST = AST
	structure Unparse = UnparseFn(structure AST = AST)

	datatype scopeType = Block | Struct
	datatype scopeInfo =
	    SCInfo of {scopeType : scopeType,
		       scopeInfo : AST.scopeInfo option}
	datatype symInfo =
	    SymInfo of {typ : AST.syntacticType,
			typLookup : AST.lookupFun,
			isParam : bool,
			symInfo : AST.symInfo option ref}
	val debug = false

	fun scopeTypeToStr Block = "Block"
	  | scopeTypeToStr Struct = "Struct"

	(* hasMoreInfo t1 t2 is true if type t1 contains more
	   information than type t2; type t1 must be a refinement
	   of t2.  This is only an issue with function types, and this
	   code is heuristic. *)
	fun hasMoreInfo (AST.SyntacticType(SOME AST.Static, _, _,
				    (AST.Function _|AST.OldFunction _)::_))
	    (AST.SyntacticType(_, _, _, (AST.Function _|AST.OldFunction _)::_))
	    = true
	  | hasMoreInfo (AST.SyntacticType(_, _, _,
				      (AST.Function _|AST.OldFunction _)::_))
	    (AST.SyntacticType(_, _, _,
			 (AST.Function []|AST.OldFunction ([],_))::_)) = true
	  | hasMoreInfo _ _ = false

	(* Never insert declarations when in a struct scope *)
	fun checkInsertId id _ (SCInfo {scopeType=Struct, ...}) =
	    (if debug then
		 print (concat ["Not inserting (Struct) ",
				AST.idToString id,
				"\n"])
	     else ();
		 false)
	  | checkInsertId id {old=SOME (SymInfo {typ=typ as
				   (AST.SyntacticType(sc,_,_,_)), ...}),
			      new=SymInfo {typ=typ' as
				   AST.SyntacticType(sc',_,_,_), ...}} _ =
		 if (sc = SOME AST.Typedef andalso sc' <> SOME AST.Typedef)
		     orelse
		    (sc <> SOME AST.Typedef andalso sc' = SOME AST.Typedef)
		     then
			 (print (concat [AST.idToString id,
					 " redeclared as different kind",
					 " of symbol\n"]);
			  false)
		 else if hasMoreInfo typ' typ then
		     (if debug then
			  print (concat ["Reinserting (Block) ",
					 AST.idToString id,
					 " (old=",
					 Unparse.syntacticType "" typ,
					 ", new=",
					 Unparse.syntacticType "" typ',
					 ")\n"])
		      else ();
		      true)
		 else
		     (if debug then
			  print (concat ["Not reinserting ",
					 AST.idToString id,
					 " (old=",
					 Unparse.syntacticType "" typ,
					 ", new=",
					 Unparse.syntacticType "" typ',
					 ")\n"])
		      else ();
		      false)
	  | checkInsertId id _ _ =
		     (if debug then
			  print (concat ["Inserting (Block) ",
					 AST.idToString id,
					 "\n"])
		      else ();
			  true)
			  
	structure IdST = SymTab(structure IdKey = AST.IdKey
				type info = symInfo
				type scopeInfo = scopeInfo
				val checkInsert = checkInsertId)
	    
	structure StructST = SymTab(structure IdKey = AST.IdKey
				    type info = AST.syntacticType*
					        AST.lookupFun
				    type scopeInfo = unit
				    fun checkInsert  _ _ _ = true)
	    
	    
	(* Maintain two separate namespaces, one for
	   identifiers, one for struct and union names *)
	fun emptySymTab () =
	    let val ids = IdST.mkSymTab ()
		val structs = StructST.mkSymTab ()
	    in
		IdST.enterScope ids (SCInfo {scopeType=Block,
				     scopeInfo = SOME
					 (AST.ExtraInfo.rootScopeInfo ())});
		StructST.enterScope structs ();
		{ids=ids, structs=structs}
	    end
	fun idsOfSymTab {ids,structs} = ids
	fun structsOfSymTab {ids,structs} = structs
	fun resetSymTab st =
	    (IdST.resetSymTab (idsOfSymTab st);
	     StructST.resetSymTab (structsOfSymTab st);
	     IdST.enterScope (idsOfSymTab st)
	                     (SCInfo {scopeType=Block,
				      scopeInfo = SOME
				          (AST.ExtraInfo.rootScopeInfo ())});
	     StructST.enterScope (structsOfSymTab st) ())
	    
	    
	type scope = IdST.scope*StructST.scope
	type lookupFun = AST.namespace -> AST.identifier ->
	    {typ : AST.syntacticType,
	     scopeInfo : AST.scopeInfo option,
	     symInfo : AST.symInfo option ref} option
	val symbols = emptySymTab ()
	val lastScope = ref (NONE : scope option)
	val paramsScope = ref (NONE : scope option)
	fun reset () = (resetSymTab symbols;
			lastScope := NONE;
			paramsScope := NONE)
	    
	(* Get the current scope.  This should probably call
	   getCScope to get a scope that can't be updated, but
	   that uses too much memory. *)
	fun getScope () =
	    let val scIds = IdST.getScope (idsOfSymTab symbols)
		val scStructs = StructST.getScope (structsOfSymTab
						    symbols)
	    in
		(scIds, scStructs)
	    end

	fun scopeLookup (idScope, _) AST.VarNames id =
	    (case IdST.findInScope idScope id of
		 SOME {info=SymInfo {typ, typLookup, isParam, symInfo},
		       scope} =>
		 let val SCInfo {scopeType, scopeInfo, ...} =
		                       IdST.getScopeInfo scope
		 in
		     SOME (AST.SymInfo {typ = typ,
					typLookup = typLookup,
					isParam = isParam,
					isGlobal = IdST.isGlobalScope scope,
					scopeInfo = scopeInfo,
					symInfo = symInfo})
		 end
	       | NONE => NONE)
	  | scopeLookup (_, structScope) AST.StructNames id =
	    case StructST.findInScope structScope id of
		SOME {info=(typ, typLookup), scope} => 
		    SOME (AST.SymInfo {typ = typ,
				       typLookup = typLookup,
				       isParam = false,
				       isGlobal = false,
				       scopeInfo = NONE,
				       symInfo = ref NONE})
	      | NONE => NONE

	fun mkLookupFun () = scopeLookup (getScope ())

	fun insertId id typ =
	    IdST.insert (idsOfSymTab symbols) id (SymInfo {typ = typ,
							   typLookup =
							   mkLookupFun (),
							   isParam = false,
							   symInfo = ref NONE})
			    
	fun insertStruct id agAndDeclList =
	    StructST.insert (structsOfSymTab symbols) id (agAndDeclList,
							  mkLookupFun ())
	    
	(* Insert a value of an enumeration.  Behaves oddly -- the names
	   of enumeration values declared inside of a struct are at the
	   same lexical scope as the structure. *)
	fun insertEnum id typ =
	    let val currentIdScope = IdST.getScope (idsOfSymTab symbols)
		val structDepth = StructST.scopeDepth (StructST.getScope
						   (structsOfSymTab symbols))
		fun getEnumScope curScope =
		    if IdST.scopeDepth curScope = structDepth then
			curScope
		    else
			case IdST.parentScope curScope of
			    SOME parent => getEnumScope parent
			  | NONE => raise Fail "No parent scope"
		val idScopeOfEnum = getEnumScope (IdST.getScope
						  (idsOfSymTab symbols))
	    in
		IdST.setScope (idsOfSymTab symbols) idScopeOfEnum;
		insertId id typ;
		IdST.setScope (idsOfSymTab symbols) currentIdScope
	    end

	fun getIdType id =
	    case IdST.find (idsOfSymTab symbols) id of
		SOME {info=SymInfo {typ, ...},
		      scope} => SOME typ
	      | NONE => NONE

	fun idIsTypedefName id =
	    case getIdType id of
		SOME (AST.SyntacticType (SOME AST.Typedef, _, _, _)) => true
	      | _ => false
	    
	fun pushScope () =
	    case !paramsScope of
		SOME (ids, structs) =>
		    (if debug then
			 print "pushScope: entering function-scope\n"
		     else ();
			 paramsScope := NONE;
			 IdST.setScope (idsOfSymTab symbols) ids;
			     StructST.setScope (structsOfSymTab symbols)
			     structs)
	      | NONE =>
			 (if debug then
			      print "pushScope: Block scope\n"
			  else ();
			      IdST.enterScope (idsOfSymTab symbols)
			      (SCInfo {scopeType=Block,
				       scopeInfo=SOME
				       (AST.ExtraInfo.enterScope ())});
			      StructST.enterScope
			      (structsOfSymTab symbols) ())
			      
	(* Structs are weird:
	 1. Nested structs do not have nested scopes.
	 2. Field names of structs do not shadow typedef names
	 (they're in their own namespace).
	 Thus pushStructScope simply makes
	 insertId temporarily do nothing, and doesn't
	 add on another nested struct scope. *)
	fun pushStructScope () =
	    IdST.enterScope (idsOfSymTab symbols) (SCInfo
						   {scopeType=Struct,
						    scopeInfo=NONE})
	    
	fun popScope () =
	    let val scIds = IdST.getScope (idsOfSymTab symbols)
		val scStructs = StructST.getScope (structsOfSymTab symbols)
	    in
		lastScope := SOME (scIds, scStructs);
		IdST.exitScope (idsOfSymTab symbols);
		(* If we last entered a Struct scope, we don't
		 pop off the struct symbol table, since we
		 didn't add one. *)
		case IdST.getScopeInfo scIds of
		    SCInfo {scopeType=Block,...} => StructST.exitScope
			(structsOfSymTab symbols)
		  | _ => ()
	    end

	(* Mark all the identifiers in a scope as parameters, and
           change any array types to pointer types (since parameters
	   that are arrays actually have a pointer type). *)
	fun fixParams (SymInfo {typ, typLookup, symInfo, ...}) =
	    let val newType = (case typ of
				   AST.SyntacticType (sc, ts, tq,
						      (AST.Array _::rest)) =>
				   AST.SyntacticType (sc, ts, tq,
						      (AST.Pointer []::rest))
				 | nonArrayType => nonArrayType)
	    in
		SymInfo {typ=newType,
			 typLookup=typLookup,
			 isParam=true,
			 symInfo=symInfo}
	    end

	fun recoverParameterScopeOnNextPush () =
	    case !lastScope of
		SOME (idsc, strsc) =>
		    (paramsScope := SOME (IdST.mapScope fixParams idsc,
					  strsc);
		     lastScope := NONE)
	      | _ =>
	       print "ERROR recoverParameterScopeOnNextPush: no such scope\n"

    end
