(* a PAM-3 interface to the type checker for testing purposes *)

structure pam_struct : PAM_STRUCT = struct
   type region_type = typeCheckCParser.AST.region

   fun get_region_range region = region
end

structure pam_typecheck = PAM_3(pam_struct)

structure pam_traverser : sig

   structure AST : AST
   val table : (string, (AST.region * AST.syntacticType * (pam_typecheck.markup option)) list ref) HashTable.hash_table

   val traverse_ast_and_view_results : typeCheck.ast_type -> string list -> typeCheck.ast_type

end = struct

   structure AST = typeCheckCParser.AST

   exception hashException

   val table_of_contents = (pam_typecheck.VirtualWindow "File Summary")

   val table = HashTable.mkTable (HashString.hashString, (fn (str1, str2) => OS.FileSys.fullPath str1 =
							                     OS.FileSys.fullPath str2))
      (100, hashException)
      : (string, (AST.region * AST.syntacticType * (pam_typecheck.markup option)) list ref) HashTable.hash_table

   fun expression_handler (sourceMap, expr) =
      case expr of
	 ((AST.Variable (_, astinfo)) |
	  (AST.Constant (_, astinfo)) |
	  (AST.Conditional (_, _, _, astinfo)) |
	  (AST.Application (_, _, astinfo, _)) |
	  (AST.Operation (_, _, astinfo)) |
	  (AST.Assignment (_, _, _, astinfo)) |
	  (AST.Comma (_, _, astinfo)) |
	  (AST.Cast (_, _, astinfo)) |
	  (AST.SizeOfExpr (_, astinfo)) |
	  (AST.SizeOfTypeExpr (_, astinfo)) |
	  (AST.FieldDirect (_, _, astinfo)) |
	  (AST.FieldIndirect (_, _, astinfo)) |
	  (AST.Postfix (_, _, astinfo)) |
	  (AST.Index (_, _, astinfo)) |
	  (AST.Extension ( _, astinfo))) =>
	 let
	    val {region as (lo, hi), typeInfo, ...} = astinfo
	    val {fileName, ...} = SourceMap.filepos sourceMap lo
	    val entry = HashTable.find table fileName
	    val color =
	       case expr of
		  ( (AST.Variable _) | (AST.Constant _)) => SOME pam_typecheck.color_2
		| _ => NONE
	    val new_addition =
	       case typeInfo of
		  SOME synType =>
		     ((lo-1, hi-1), synType, color) 
		| NONE =>
		     raise typeCheck.Unsupported ((~1, ~1), "all expressions must have types")
	 in
	    (case entry of
		SOME nodes =>
		   nodes := new_addition :: (!nodes)
	      | NONE =>
		   HashTable.insert table (fileName, ref [new_addition]);
	     NONE)
	 end

   fun decl_handler (sourceMap, expr) =
      (case expr of
	  ((AST.VarDecl {region as (lo, hi), typ, ...}) |
	   (AST.FunDef {region as (lo, hi), typ, ...}) |
	   (AST.FunDecl {region as (lo, hi), typ, ...})) =>
	  let
	     val {fileName, ...} = SourceMap.filepos sourceMap lo
	     val entry = HashTable.find table fileName
	     val new_addition = ((lo-1, hi-1), typ, SOME pam_typecheck.color_4)
	  in
	     (case entry of
		 SOME nodes =>
		    nodes := new_addition :: (!nodes)
	       | NONE => 
		    HashTable.insert table (fileName, (ref [new_addition]));
	      NONE)
	  end
	 | _ => raise typeCheck.Unsupported ((~1, ~1), "illegal function call"))


   fun print_type synType env =
      let
	 fun class_str class =
	    case class of
	       AST.Register => "register "
	     | AST.Auto => "auto "
	     | AST.Static => "static "
	     | AST.Extern => "extern "
	     | AST.Typedef => "typedef ? "

	 and specifier_str specifier =
	    case specifier of
	       AST.Void => "void "
	     | AST.Char => "char "
	     | AST.UnsignedChar => "unsigned char "
	     | AST.ShortInt => "short int "
	     | AST.UnsignedShortInt => "unsigned short int "
	     | AST.Int => "int "
	     | AST.UnsignedInt => "unsigned int "       
	     | AST.LongInt => "long int "
	     | AST.LongLong => "long long "
	     | AST.UnsignedLongInt => "unsigned long int "
	     | AST.UnsignedLongLong => "unsigned long long "
	     | AST.Float => "float "
	     | AST.Double => "double "
	     | AST.LongDouble => "long double "     
	     | AST. NamedType (id) => (Lexeme.idToString id) ^ " "
	     | AST.Enum (SOME id, _) => "enum " ^ (Lexeme.idToString id)
	     | AST.Enum (NONE, _) => "enum <unnamed> "
	     | AST.Aggregate (AST.Struct, SOME id, _) => "struct " ^ (Lexeme.idToString id) ^ " "
	     | AST.Aggregate (AST.Union, SOME id, _) => "union " ^ (Lexeme.idToString id) ^ " "
	     | AST.Aggregate (AST.Struct, NONE, _) => "union <unnamed> "
	     | AST.Aggregate (AST.Union, NONE, _) => "union <unnamed> "
	     | AST.NamedAggregate (AST.Struct, id) => "struct " ^ (Lexeme.idToString id) ^ " "
	     | AST.NamedAggregate (AST.Union, id) => "Union " ^ (Lexeme.idToString id) ^ " "

	 and one_qualifier_str qual =
	    case qual of
	       AST.Const => "const "
	     | AST.Volatile => "volatile "
	     | AST.Inline => "inline "
	     | AST.NonConst => "nonconst "
	     | AST.Qualifier id => "qualifier(" ^ (Lexeme.idToString id) ^ ") "

	 and qualifiers_str quals =
	    foldl (fn (qual, str) => str ^ (one_qualifier_str qual)) "" quals

	 and region_identifers_str region_identifier_list =
	    foldl (fn ((region, id), str) => str ^ (Lexeme.idToString id)) "" region_identifier_list

	 and one_declaration_str decl =
	    case decl of
	       AST.EllipsisDecl _ => "..."
	     | AST.VarDecl {name, typ, ...} =>
		  (get_type_str typ) ^ (Lexeme.idToString name)
	     | _ => "illegal type definition"

	 and declarations_str [] = ""
	   | declarations_str [decl] = one_declaration_str decl
	   | declarations_str (a :: rest) =
	    (one_declaration_str a) ^ (foldl (fn (decl, str) => str ^ ", " ^ (one_declaration_str decl)) "" rest)

	 and constant_integer_str expr = ""

	 and one_declarator_str decl =
	    case decl of
	       AST.Pointer quals =>
		  "* " ^ (qualifiers_str quals)
	     | AST.Array expr =>
		  "[" ^ (constant_integer_str expr) ^ "] "
	     | AST.BitField expr =>
		  " : " ^ (constant_integer_str (SOME expr))
	     | AST.Function declarations =>
		  "(" ^ (declarations_str declarations) ^ ") "
	     | AST.OldFunction (region_identifier_list, SOME region) =>
		  "(" ^ (region_identifers_str region_identifier_list) ^ ", ...) "
	     | AST.OldFunction (region_identifier_list, NONE) => 
		  "(" ^ (region_identifers_str region_identifier_list) ^ ") "

	 and declarators_str declarators =
	    foldl (fn (decl, str) => str ^ (one_declarator_str decl)) "" declarators

	 and get_type_str (AST.SyntacticType (class, specifier, qualifiers, declarators)) =
	    (case class of
		SOME c => class_str c
	      | NONE => "") ^ (qualifiers_str qualifiers) ^ (specifier_str specifier) ^ (declarators_str declarators)
      in
	 (pam_typecheck.send_status (#1 (pam_typecheck.streams_of env), get_type_str synType);
	  SOME env)
      end

   fun send_by_filename filename env =
      let
	 val entry = HashTable.find table filename
	 fun send_entry ((region, synType, color), env) =
	    let
	       val message = pam_typecheck.Hyper {attrib = color, region=region, func = print_type synType}
	    in
	       pam_typecheck.send (pam_typecheck.RealWindow filename) message env
	    end
      in
	 if List.exists (fn str => (OS.FileSys.fullPath str) = filename) (pam_typecheck.get_info env) then
	    SOME env
	 else
	    case entry of
	       SOME nodes =>
		  let
		     val env' = ((pam_typecheck.change_window (pam_typecheck.RealWindow filename)) o
				 (pam_typecheck.set_window_down table_of_contents) o
				 (pam_typecheck.set_window_up (pam_typecheck.RealWindow filename))) (foldl send_entry env (!nodes))
		  in
		     SOME (pam_typecheck.set_info env' ((OS.FileSys.fullPath filename)
							:: (pam_typecheck.get_info env')))
		  end
	     | NONE => (pam_typecheck.send_status (#1 (pam_typecheck.streams_of env), "Error: cannot find indicated filename: " ^ filename);
			NONE)
      end

   val callbacks = typeCheck.Callbacks
      {(* the whole program *)
       program = NONE,
       (* statements *)
       return = NONE,
       expression = NONE,
       ifThen = NONE,
       ifThenElse = NONE,
       whileDo = NONE,
       doWhile = NONE,
       for = NONE,
       goTo = NONE,
       switch = NONE,
       caseStat = NONE,
       labeled = NONE,
       break = NONE,
       continue = NONE,
       compound = NONE,
       (* expressions *)
       variable = SOME expression_handler,
       constant = SOME expression_handler,
       conditional = SOME expression_handler,
       application = SOME expression_handler,
       operation = SOME expression_handler,
       assignment = SOME expression_handler,
       comma = SOME expression_handler,
       cast = SOME expression_handler,
       sizeOfExpr = SOME expression_handler,
       sizeOfTypeExpr = SOME expression_handler,
       fieldDirect = SOME expression_handler,
       fieldIndirect = SOME expression_handler,
       postfix = SOME expression_handler,
       index = SOME expression_handler,
       extension = SOME expression_handler,
       (* declarations *)
       ellipsisDecl = NONE,
       varDecl = SOME decl_handler,
       typeDef = NONE,
       funDef = SOME decl_handler,
       funDecl = SOME decl_handler}

   fun traverse_ast_and_view_results (ast as {sourceMap, ...}) filename_list =
      let 
	 val env =pam_typecheck.start_pam TextIO.stdOut TextIO.stdIn [] (fn str => fn env => env) (fn env => env)

	 fun connect_to_error_location filename (lo, hi) env =
	    let
	       val win = pam_typecheck.RealWindow filename
	       val new_env_option = send_by_filename filename env
	    in
	       case new_env_option of
		  SOME new_env =>
		     SOME (((pam_typecheck.goto_char win lo) o
			    (pam_typecheck.change_window (pam_typecheck.RealWindow filename)) o
			    (pam_typecheck.set_window_down table_of_contents) o
			    (pam_typecheck.set_window_up (pam_typecheck.RealWindow filename))) new_env)


		| NONE => SOME env
	    end

	 val ((ast, errors), env') =
	    (typeCheck.check_types ast callbacks, env)
	    handle ((typeCheck.ParseError (region, str)) | (typeCheck.Unsupported (region, str))) =>
	       let
		  val locs = SourceMap.fileregion sourceMap region
		  val ({fileName, line = line_begin, column = column_begin},
		       {line = line_end, column = column_end, ...}) = hd locs
		  val error_display = [pam_typecheck.Text {attrib = NONE,
							   text = "The following major error occurred during type checking:",
							   region = NONE},
				       pam_typecheck.Text {attrib = NONE, text = "\n\n", region = NONE},
				       pam_typecheck.HyperText {attrib = SOME pam_typecheck.color_3,
								text =
								fileName ^ ":" ^ (Int.toString line_begin) ^ "." ^
								Int.toString column_begin ^ "-" ^ (Int.toString line_end) ^ "." ^
								Int.toString column_end,
								region = NONE,
								func = connect_to_error_location fileName region},
				       pam_typecheck.Text {attrib = NONE, text = " " ^ str, region = NONE},
				       pam_typecheck.Text {attrib = NONE, text = "\n\n", region = NONE}]
	       in
		  ((ast, []), pam_typecheck.send_list table_of_contents error_display env)
	       end

	 fun error_producer ((region, str), cur_list) =
	    let
	       val locs = SourceMap.fileregion sourceMap region
	       val ({fileName, line = line_begin, column = column_begin},
		       {line = line_end, column = column_end, ...}) = hd locs
	    in
	       [pam_typecheck.HyperText {attrib = SOME pam_typecheck.color_3,
					 text =
					 fileName ^ ":" ^ (Int.toString line_begin) ^ "." ^
					 Int.toString column_begin ^ "-" ^ (Int.toString line_end) ^ "." ^
					 Int.toString column_end,
					 region = NONE,
					 func = connect_to_error_location fileName region},
		pam_typecheck.Text {attrib = NONE, text = " " ^ str, region = NONE},
		pam_typecheck.Text {attrib = NONE, text = "\n", region = NONE}] @ cur_list
	    end

	 val type_errors_message =
	    (foldr error_producer [] errors) @
	    [pam_typecheck.Text {attrib = NONE, text = "\n", region = NONE}]

	 val files_message = foldr (fn (str, cur_list) =>
				    [pam_typecheck.HyperText {attrib = SOME pam_typecheck.color_2,
							      text = str, region = NONE, func = send_by_filename str},
				     pam_typecheck.Text {attrib = NONE, text = "\n", region = NONE}] @
				    cur_list)
	                            []
				    filename_list

	 val messages =
	    (case type_errors_message of
		[] => []
	      | [el] => []
	      | _ =>
		   [pam_typecheck.Text {attrib = NONE,
					text =
					case length type_errors_message of
					   2 => "A type checking error occurred:"
					 | _ => "Type check errors occurred:",
					      region = NONE}, pam_typecheck.Text {attrib = NONE, text = "\n\n", region = NONE}] @
		   type_errors_message) @
	    [pam_typecheck.Text {attrib = NONE,
				 text =
				 case length filename_list of
				    1 => "Here is the file that was typechecked:"
				  | _ => "Here are the files that were typechecked:",
				 region = NONE}, pam_typecheck.Text {attrib = NONE, text = "\n\n", region = NONE}]
	    @ files_message
      in
	 ((pam_typecheck.loop_until_done o (pam_typecheck.change_window table_of_contents) o
	   (pam_typecheck.send_list table_of_contents messages)) env';
	  ast)
	 handle _ =>
	    (pam_typecheck.send_debug ((#1 (pam_typecheck.streams_of env')), "An exception occurred");
	     ast)
      end

end