(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

functor GraphFn(NodeKey : GRAPHNODE) : GRAPH =
    struct

	open URef
	open Util

	(*************************************************************

	 Data types and exceptions

	 *************************************************************)

	type key = NodeKey.node

	structure NodeMap = BinaryMapFn(NodeKey)

	datatype NodeMark = VISITED | NOTVISITED

	datatype node = Node of {mark : NodeMark ref,
				 dfn     : int ref,
				 lowlink : int ref, (* Hopcroft *)
				 info : NodeKey.ord_key,
				 aliases : NodeKey.ord_key jlist,
				 predecessors : node uref jlist ref,
				 successors : node uref jlist ref}

	type graph = node uref NodeMap.map ref

	exception noRoot        (* Call graph has more than
				   one or no root nodes *)

	(*************************************************************

	 Global variables

	 *************************************************************)

	val debug = ref(false)

	fun empty () = ref(NodeMap.empty) : graph

	fun reset CG = CG := NodeMap.empty

	(*************************************************************

	 Graph utilities

	 *************************************************************)

	fun nodeMarkOr NOTVISITED NOTVISITED = NOTVISITED
	  | nodeMarkOr _ _ = VISITED

	(* unifyNode n1 n2 makes a new node containing the info
	   of n1, the OR of the marks of n1 and n2, and the union
	   of the successor lists (possibly with duplicates).
	   dfn and lowlink are the minimum of the values in the nodes. *)
	fun unifyNode (Node{mark = ref mark1,
			    dfn = ref dfn1,
			    lowlink = ref lowlink1,
			    info = info1,
			    aliases = aliases1,
			    predecessors= ref predecessors1,
			    successors = ref successors1},
		       Node{mark = ref mark2,
			    dfn = ref dfn2,
			    lowlink = ref lowlink2,
			    info = info2,
			    aliases = aliases2,
			    predecessors = ref predecessors2,
			    successors = ref successors2}) =
	    (if !debug then
		 (print "Unifying ";
		  print (NodeKey.toString info1);
		  print " and ";
		  print (NodeKey.toString info2);
		  print "\n")
	     else
		  ();
	     Node{mark=ref(nodeMarkOr mark1 mark2),
		  info=info1,
		  aliases=join (aliases1, aliases2),
		  predecessors=ref(join(predecessors1,predecessors2)),
		  successors=ref(join (successors1,successors2)),
		  dfn=ref (Int.min (dfn1, dfn2)),
		  lowlink=ref (Int.min (lowlink1, lowlink2))})

	(* Given a node uref and a node uref list, return true iff
	   the node uref is in the list *)
	fun nodeEqual i1 n2 =
	    let val Node {info=i2,...} = !!n2
	    in
		(NodeKey.compare (i1,i2) = EQUAL)
	    end

	fun nodeMember n jl =
	    let val Node {info,...} = !!n
	    in
		case jfind (nodeEqual info) jl of
		    NONE => false
		  | SOME _ => true
	    end

	fun unMark CG =
	    let fun unMarkNode (Node{mark,...}) = mark := NOTVISITED
	    in
		NodeMap.app (unMarkNode o op!!) (!CG)
	    end

	(* Get the node corresponding to a info *)
	fun getNode CG n =
	    (case NodeMap.find(!CG, n) of
		 SOME m => m
	       | NONE =>
		     let val newNode = uRef(Node {mark = ref NOTVISITED,
						  info = n,
						  aliases = Single n,
						  predecessors = ref Nil,
						  successors = ref Nil,
						  dfn = ref 0,
						  lowlink = ref 0})
		     in
			 CG := NodeMap.insert(!CG, n, newNode);
			 newNode
		     end)

	(* Add node if it doesn't already exist *)
	fun addNode CG n =
	    (case NodeMap.find(!CG, n) of
		 SOME m => ()
	       | NONE =>
		     let val newNode = uRef(Node {mark = ref NOTVISITED,
						  info = n,
						  aliases = Single n,
						  predecessors = ref Nil,
						  successors = ref Nil,
						  dfn = ref 0,
						  lowlink = ref 0})
		     in
			 CG := NodeMap.insert(!CG, n, newNode)
		     end)

	fun mapSuccEdges (f : key*key->key*key)
	             (CG : graph) =
	    let val newCG = empty()
		fun traverseEdges (s as Node{info=sinfo, successors,...}) =
		    let fun traverseEdge (t as Node {info=tinfo,...}) =
			addEdge newCG (f (sinfo, tinfo))
		    in
			japp (traverseEdge o op!!) (!successors)
		    end
	    in
		NodeMap.app (traverseEdges o op!!) (!CG);
		newCG
	    end


	and appSuccEdges (f : key*key->unit)
	             (CG : graph) =
	    let fun traverseEdges (s as Node{info=sinfo, successors,...}) =
		let fun traverseEdge (t as Node {info=tinfo,...}) =
		    (f (sinfo, tinfo))
		in
		    japp (traverseEdge o op!!) (!successors)
		end
	    in
		NodeMap.app (traverseEdges o op!!) (!CG)
	    end

	(*************************************************************

	 Interface

	 *************************************************************)

	(* Add edge from s to t *)
	and addEdge CG (s, t) =
	    let val snode = getNode CG s
		val Node {successors, ...} = !!snode
		val tnode = getNode CG t
		val Node {predecessors, ...} = !!tnode
	    in
		(* Performance bottleneck! *)
		if not (nodeMember tnode (!successors)) then
		    (successors := jcons (tnode, (!successors));
		     predecessors := jcons (snode, (!predecessors)))
		else
		    ()
	    end

	fun getAliases nodeUref =
	    let val Node {aliases, ...} = !!nodeUref
	    in
		aliases
	    end

	(* Return the predecessors of a node *)
	fun predecessors CG n =
	    let val Node {predecessors, ...} = !!(getNode CG n)
	    in
		jfoldl (fn (n, l) => Join(getAliases n, l)) Nil (!predecessors)
	    end

	(* Return the sucessors of a node *)
	fun successors CG n =
	    let val Node {successors, ...} = !!(getNode CG n)
	    in
		jfoldl (fn (n, l) => Join(getAliases n, l)) Nil (!successors)
	    end

	(* Make two nodes equivalent *)
	fun mkEquiv CG (s, t) =
	    let val snode = getNode CG s
		val tnode = getNode CG t
	    in
		if not (equal (snode, tnode)) then
		    (unify unifyNode (snode, tnode); ())
		else
		    ()
	    end

	(* Return a list of root nodes in the graph, i.e.
	   nodes with no predecessors (not including self loops).
	   If the strongly connected components have been computed,
	   at least one such node must exist. *)
	fun findRoots CG =
	    let fun markSuccessors n =
		let val n' as Node {mark, info, successors,...} = !!n
		    fun markIfNotN m =
			let val m' as Node {mark, info,...} = !!m
			in
			    if equal (n, m) then ()
			    else mark := VISITED
			end
		in
		    japp markIfNotN (!successors)
		end
		fun accumRoots (n, rootList) =
		    let val n' as Node {mark,info,...} = !!n
		    in
			case (!mark) of
			    VISITED => rootList
			  | NOTVISITED =>
				(mark := VISITED;
				 info::rootList)
		    end
	    in
		unMark CG;
		NodeMap.app markSuccessors (!CG);
		NodeMap.foldl accumRoots [] (!CG)
	    end

	(*
	 *   Finds strongly connected components in the graph
	 *   corresponding to the constraints.
	 *   Alg. Aho Hopcroft Ullmann "The Design and Analysis of
	 *   Computer Algorithms", page 189
	 *
	 *)
	structure StackHash =
	    struct
		structure Key =
		    struct
			type hash_key = int
			val hashVal = Word.fromInt
			val sameKey = (op = : int * int -> bool)
		    end

		structure HT = HashTableFn(Key)

		datatype stack = Stack of {list : (node * int) list ref,
					   hash : unit HT.hash_table}

		fun emptyStack () = Stack {list=ref [],
					   hash=HT.mkTable (64, Match)}

		fun push (Stack {list,hash}) (n as (_,i)) =
		    (
		     list := n :: !list;
		     HT.insert hash (i,())
		    )

		fun onStack (Stack {list,hash}) i =
		    case HT.find hash i of
			SOME _ => true
		      | NONE => false

		fun popUntil (Stack {list,hash}) i =
		    let fun info (Node {aliases,...}) = aliases

			fun split ((v,j)::rest, acc) = 
			    (
			     HT.remove hash j;
			     if j=i then 
				 (rest, info v::acc)
			     else
				 split (rest, info v::acc)
			    )
		    
			val (rest, popped) = split (!list, [])
		    in
			list := rest;
			popped
		    end

	    end
	 
	fun DFSCApply CG f =
	    let val roots = findRoots CG
		val dfcount = ref 1
		val stack   = StackHash.emptyStack ()

		fun incDF () = (!dfcount before (dfcount := !dfcount + 1))

		val push = StackHash.push stack

		val onStack = StackHash.onStack stack

		val popUntil = StackHash.popUntil stack

		fun DFSVisit (v as Node {mark,dfn,lowlink=vlowlink,
					 aliases,successors,...}) =
		    let val _ = mark := VISITED
			val vdfn = incDF ()
			val _ = dfn  := vdfn
			val _ = vlowlink := vdfn
			val _ = push (v,vdfn)

			fun visitChild uw =
			    let val (w as Node {mark,dfn,lowlink=wlowlink,
						...}) = !!uw
			    in
				case !mark of
				    VISITED => 
					let val wdfn = !dfn
					in
					    if wdfn < vdfn andalso
					       onStack wdfn then
					       vlowlink := Int.min (wdfn,
								    !vlowlink)
					    else
						()
					end
				  | NOTVISITED => 
					let val _ = ()
					in
					    DFSVisit w;
					    vlowlink := Int.min (!vlowlink,
								 !wlowlink)
					end
			    end
		    in
			japp visitChild (!successors);
			if !vlowlink = vdfn then
			    let val component = popUntil vdfn
			    in
				f (jconcat component)
			    end
			else ()
		    end

		fun VisitIfUnmarked (v as Node {mark,...}) =
		    case !mark of
			VISITED => ()
		      | NOTVISITED => DFSVisit v

	    in
		unMark CG;
		app (fn root => VisitIfUnmarked (!!(getNode CG root)))
		    roots;
		(* Not all cycles may be reachable from roots!
		 * Try all nodes now!
		 *)
		NodeMap.app (VisitIfUnmarked o !!) (!CG)
	    end
					    
	(* Compute the strongly connected components of a graph.  Untested. *)
(*	fun collapseStronglyConnectedComponents CG =
	    let fun mkEquiv' (x, y) = (mkEquiv CG (x, y); x)
		fun collapse nodes =
		    case Util.jflatten nodes of
			[] => ()
		      | [_] => ()
		      | (x::xs) => (foldl mkEquiv' x xs; ())
	    in
		DFSCApply CG collapse
	    end*)

	(* Apply f to each node of the graph in final depth first order,
	   starting at the root; the function is applied after all the
	   children have been visited.  The aliases of a single
	   node are visited in arbitrary order.
	*)
	fun DFSApply CG f =
	    let val roots = findRoots CG
		fun DFSVisit (n as Node {mark,aliases,successors,...}) =
		    (case (!mark) of
			 VISITED => ()
		       | NOTVISITED =>
			     (mark := VISITED;
			      japp (DFSVisit o op!!) (!successors);
			      f aliases))
	    in
		dotPrint CG "cg.dot";
		if (length roots) <> 1 then
		    print "Warning:  Disconnected graph\n"
		else
		    ();
		unMark CG;
		app (fn root => DFSVisit (!!(getNode CG root)))
		    roots
	    end

	and escape_chars s = 
	    String.translate (fn #"\"" => "\\\""
	                       | #"\\" => "\\\\"
			       | #"\t" => "\\t"
			       | #"\n" => "\\n"
			       | c => String.str c) s


	(* Output the call graph in a format parseable by dot to
	 a file *)
	and dotPrint CG fileinfo =
	     let val os = TextIO.openOut fileinfo
		 fun outputSpace () = TextIO.output(os, " ")
		 fun outputNode n =
		     let val n' as Node {aliases,...} = !!(getNode CG n)
			 fun quote n =
			     (
			      TextIO.output1 (os,#"\"");
			      TextIO.output (os,escape_chars (NodeKey.toString n));
			      TextIO.output1 (os,#"\"")
			     )
		     in
			 japp quote aliases
		     end
		 fun outputEdge (sinfo, tinfo) =
		     (outputNode sinfo;
		      TextIO.output(os, " -> ");
		      outputNode tinfo;
		      TextIO.output(os,";\n"))
	     in
		 TextIO.output(os, "digraph G {\n");
		 appSuccEdges outputEdge CG;
		 TextIO.output(os, "size=\"7.5,11\"");
		 TextIO.output(os, "}\n");
		 TextIO.closeOut os
	     end

	type node = key

	val addEdge = fn G => fn {from,to} => addEdge G (from,to)

    end
