signature MEM_USAGE =
  sig
    type report_func = int * int * int -> string
      
    type time = Time.time
    val second: time
    val tenthSecond: time
    val quarterSecond: time
      
    (* memUsage (file,report,timeOut) func -
     * Spawn a child process that monitors the memory usage
     * of func. All data are printed to the file; each line
     * is printed using report. The monitor waits, roughly,
     * timeOut between each report.
     *)
    val memUsage: (string * report_func option * time) -> (unit -> unit) -> unit
  end

structure MemUsage : MEM_USAGE =
  struct
    
    local
      val c = ref 0
    in fun getMemUsageCounter() = !c
      fun incMemUsageCounter() = c := !c + 1
      fun resetMemUsageCounter() = c := 0
    end

    fun pidName () =
      let val pid = Posix.ProcEnv.getpid ()
      in SysWord.fmt StringCvt.DEC (Posix.Process.pidToWord pid)
      end

    type report_func = int*int*int -> string

    type time = Time.time
    val second = Time.fromMilliseconds 1000
    val longTime = Time.fromMilliseconds 10000
    val tenthSecond = Time.fromMilliseconds 100 
    val quarterSecond = Time.fromMilliseconds 250

    (* default function for printing memory usage *)
    fun showOutput NONE (count, size, rss) =
      concat [Int.toString count, " ", 
	      "Size= ", Int.toString size, ", ",
	      "RSS= ", Int.toString rss]
      | showOutput (SOME report) (count, size, rss) = report(count,size,rss)
      
    fun monitorMemoryUsage timeOut stream report pid =
      let val {size, rss} = Info.getInfo pid
      in
	TextIO.output(stream, (showOutput report (getMemUsageCounter(), size, rss)) ^ "\n")
	before incMemUsageCounter();
	TextIO.flushOut stream;
	Posix.Process.sleep timeOut;
	monitorMemoryUsage timeOut stream report pid   (* loop call *)
      end

    fun addHandler (signal, handler) =
      let open Signals
	val oldHandler = inqHandler signal
	fun signalHandler (signal, times, next) =
	  (handler ();
	   case oldHandler 
	     of HANDLER oldHandler =>  
	       (* Process already provided handler; remove our handler
		* and call the old one. *)
	       (setHandler (sigINT, HANDLER oldHandler);
		oldHandler(signal, times, next))
	      |   action =>
	       (* Process did *not* provide handler; just remove our
		* handler and proceed. *)
	       (setHandler (Signals.sigINT, action);
		next))
      in
	setHandler (signal, HANDLER signalHandler)
      end

    fun killChild pid =
      (print "parent: killing child\n";
       Posix.Process.kill (Posix.Process.K_PROC pid, Posix.Signal.term);
       Posix.Process.wait ();
       print "parent: done killing\n")
	   
    fun memUsage (file,report,timeOut) func =
      let val myPid = pidName ()   (* used by child to lookup 
				    * memory usage for the parent *)
      in
	resetMemUsageCounter();
	print ("Forking child (parent = " ^ myPid ^ ")\n");
	case Posix.Process.fork () 
	  of SOME pid => (* We're in the parent process *)
	    let (* val _ = Posix.Process.sleep second *)
	      fun cleanUp () = killChild pid
	    in
	      addHandler (Signals.sigINT, cleanUp);
	      func();
	      cleanUp() (* terminate child and close childs outstream *)
	    end
	   | NONE => (* We're in the child process *)
	    let val os = TextIO.openOut file
	      fun cleanUp () = (print "child: closing output stream\n"; 
				TextIO.closeOut os)
	    in 
	      addHandler (Signals.sigTERM, cleanUp);
	      monitorMemoryUsage timeOut os report myPid
	    end
      end
  end

