/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         decl.cpp
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of declarations.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#include <assert.h>

#include "decl.h"
#include "express.h"
#include "stemnt.h"

#include "token.h"
#include "tree.h"
#include "gram.h"

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
static
void
printStorage( ostream& out, StorageType storage )
{
    switch (storage)
    {
        case ST_None:
        case ST_Auto:
            break;

        case ST_Register:
            out << "register ";
            break;

        case ST_Static:
            out << "static ";
            break;

        case ST_Extern:
            out << "extern ";
            break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
static
void
printQual( ostream& out, TypeQual qualifier )
{
    if (qualifier & TQ_Const)
        out << "const ";

    if (qualifier & TQ_Volatile)
        out << "volatile ";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
static
Type*
recurBuildType(treeNode *node, Symbol*& typeName)
{
    Type *result = NULL;

    if (node)
    {
      switch (node->which)
      {
        default:
            assert(0);
            break;

        case LEAF_T:
            switch (node->type)
            {
                case TN_TYPE:
                    assert(0);
                    break;

                case TN_IDENT:
                    typeName = buildSymbol((leafNode*) node);
                    break;

                default:
                    assert(0);
                    break;
            }
            break;

        case NODE_T:
            switch (node->type)
            {
                case TN_TYPE_LIST:
                case TN_DECLS:
                case TN_PARAM_LIST:
                case TN_IDENT_LIST:
                case TN_COMP_DECL:
                case TN_DECL:
                    // We shouldn't see these.
                    cerr << "Saw: " << name_of_node(node->type) << endl;
                    assert(0);
                    break;

                case TN_BIT_FIELD:
                    {
                      BitFieldType *bft = new BitFieldType();

                      // Get the name.                        
                      recurBuildType(node->lnode,typeName);

                      if (node->rnode)
                          bft->size = buildExpr(node->rnode);

                      result = bft;
                    }
                    break;

                case TN_PNTR:
                    {
                      PtrType *pt = new PtrType();

                      pt->subType
                        = recurBuildType(node->lnode,typeName);

                      Decl decl;
                      buildBaseType(&decl,node->rnode);
                      pt->qualifier = decl.qualifier;

                      result = pt;
                    }
                    break;

                case TN_ARRAY_DECL:
                    {
                      ArrayType *at = new ArrayType();
                        
                      at->subType
                        = recurBuildType(node->lnode,typeName);

                      if (node->rnode)
                          at->size = buildExpr(node->rnode);

                      result = at;
                    }
                    break;

                case TN_FUNC_DECL:
                    {
                      FunctionType *ft = new FunctionType();

                      ft->subType
                        = recurBuildType(node->lnode,typeName);

                      ft->getParams(node->rnode);
                      result = ft;
                    }
                    break;

                case TN_ENUM_LIST:
                    {
                      EnumType *et = new EnumType();

                      et->getEnumElements(node);
                      et->tag = typeName->dup();

                      result = et;
                    }
                    break;

                case TN_FIELD_LIST:
                    {
                      StructType *st = new StructType();

                      st->tag = typeName;
                      typeName = NULL;

                      st->getStructComponents(node);

                      result = st;
                    }
                    break;

                default:
                    cerr << "Unhandled case: "
                         << name_of_node(node->type) << endl;
                    assert(0);
                    break;
            }
            break;
      }
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type::Type(TypeType _type /* =TT_Base */)
{
    type      = _type;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type::~Type()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Type::print(ostream& out, Symbol *name, int level) const
{
    out << __PRETTY_FUNCTION__ << endl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Decl::Decl()
{
    clear();
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Decl::~Decl()
{
    delete name;
    delete tag;
    delete form;
    delete initializer;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Decl::clear()
{
    storage   = ST_None;
    qualifier = TQ_None;
    baseType  = BT_NoType;

    name        = NULL;
    tag         = NULL;
    typeName    = NULL;
    form        = NULL;
    initializer = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Decl*
Decl::halfdup() const
{
    Decl *ret = NULL;

    if (this)
    {
        ret = new Decl();

        ret->storage   = storage;
        ret->qualifier = qualifier; 
        ret->baseType  = baseType;

        ret->tag       = tag->dup();
        ret->typeName  = typeName->dup();

        // Don't duplicate the other stuff
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Decl*
Decl::dup() const
{
    Decl *ret = halfdup();

    if (ret)
    {
        ret->name = name->dup();
        ret->tag      = tag->dup();
        ret->typeName = typeName->dup();

        if (form)
            ret->form = form->dup();

        if (initializer)
            ret->initializer = initializer->dup();
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Decl::copy(const Decl& decl)
{
    storage     = decl.storage;
    qualifier   = decl.qualifier; 
    baseType    = decl.baseType;
    name        = decl.name;
    tag         = decl.tag;
    typeName    = decl.typeName;
    form        = decl.form;
    initializer = decl.initializer;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Decl::print(ostream& out, bool showBase, int level) const
{
    if (showBase)
    {
        printStorage(out,storage);
        printQual(out,qualifier);
    
        if (baseType & BT_UnSigned)
            out << "unsigned ";
    
        if (baseType & BT_Void)
            out << "void ";
        else if (baseType & BT_Char)
            out << "char ";
        else if (baseType & BT_Short)
            out << "short ";
        else if (baseType & BT_LongLong)
            out << "long long ";
        else if (baseType & BT_Float)
            out << "float ";
        else if ((baseType & BT_Double) && (baseType & BT_Long))
            out << "long double ";
        else if (baseType & BT_Double)
            out << "double ";
        else if (baseType & BT_Ellipsis)
            out << "...";
        else if (baseType & BT_Long)
            out << "long ";
        else if (baseType & BT_Struct)
        {
            out << "struct ";
            if (tag)
                out << *tag << " ";
        }
        else if (baseType & BT_Union)
        {
            out << "union ";
            if (tag)
                out << *tag << " ";
        }
        else if (baseType & BT_Enum)
        {
            out << "enum ";
            if (tag)
                out << *tag << " ";
        }
        else if (baseType & BT_UserType)
        {
            if (typeName)
                out << *typeName << " ";
        }
        else
        {
            out << "int ";        // Default
        }
    }

    if (form)
    {
        form->print(out,name,level);
    }

    else if (name)
    {
        out << *name;
    }

    if (initializer)
    {
        out << " = ";

        initializer->print(out);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Decl::findExpr( fnExprCallback cb )
{
    if (form)
        form->findExpr(cb);

    if (initializer != NULL)
        initializer->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
PtrType::PtrType()
        : Type(TT_Pointer)
{
    subType = NULL;
}

PtrType::~PtrType()
{
    delete subType;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type*
PtrType::dup() const
{
    PtrType *ret = NULL;

    if (this)
    {
        ret = new PtrType();

        ret->qualifier = qualifier; 

        if (subType)
            ret->subType = subType->dup();
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
PtrType::extend(Type *extension)
{
    if (subType)
        subType->extend(extension);
    else
        subType = extension;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
PtrType::print(ostream& out, Symbol *name, int level) const
{
    out << "*";

    if (subType)
    {
        subType->print(out,name,level);
    }
    else if (name)
    {
        out << *name;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
PtrType::findExpr( fnExprCallback cb )
{
    if (subType)
        subType->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ArrayType::ArrayType()
          : Type(TT_Array)
{
    subType = NULL;
    size    = NULL;
}

ArrayType::~ArrayType()
{
    delete subType;
    delete size;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type*
ArrayType::dup() const
{
    ArrayType *ret = NULL;

    if (this)
    {
        ret = new ArrayType();

        ret->size = size->dup();

        if (subType)
            ret->subType = subType->dup();
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ArrayType::extend(Type *extension)
{
    if (subType)
        subType->extend(extension);
    else
        subType = extension;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ArrayType::print(ostream& out, Symbol *name, int level) const
{
    if (subType)
    {
        if (subType->type != TT_Array)
            out << "(";

        subType->print(out,name,level);

        if (subType->type != TT_Array)
            out << ")";
    }
    else if (name)
        out << *name;

    out << "[";

    if (size)
        size->print(out);

    out << "]";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ArrayType::findExpr( fnExprCallback cb )
{
    if (subType)
        subType->findExpr(cb);

    if (size)
        size->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
BitFieldType::BitFieldType()
          : Type(TT_BitField)
{
    size    = NULL;
}

BitFieldType::~BitFieldType()
{
    delete size;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type*
BitFieldType::dup() const
{
    BitFieldType *ret = NULL;

    if (this)
    {
        ret = new BitFieldType();

        ret->size = size->dup();
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
BitFieldType::print(ostream& out, Symbol *name, int level) const
{
    if (name)
        out << *name;

    out << ":";

    if (size)
        size->print(out);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
BitFieldType::findExpr( fnExprCallback cb )
{
    if (size)
        size->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
FunctionType::FunctionType()
             : Type(TT_Function)
{
    size = nArgs = 0;

    args     = NULL;

    subType = NULL;
}

FunctionType::~FunctionType()
{
    for (int j=0; j < nArgs; j++)
    {
        delete args[j];
    }

    delete [] args;
    delete subType;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type*
FunctionType::dup() const
{
    FunctionType *ret = NULL;

    if (this)
    {
        ret = new FunctionType();

        ret->size    = size;
        ret->args = new Decl* [size];

        for (int j=0; j < nArgs; j++)
            ret->addArg(args[j]->dup());

        if (subType)
            ret->subType = subType->dup();
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionType::extend(Type *extension)
{
    if (subType)
        subType->extend(extension);
    else
        subType = extension;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionType::print(ostream& out, Symbol *name, int level) const
{
    if (subType)
    {
        out << "(";
        subType->print(out,name,level);
        out << ")";
    }
    else if (name)
    {
        out << *name;
    }

    out << "(";

    if (nArgs > 0)
    {
        args[0]->print(out,true);
        for (int j=1; j < nArgs; j++)
        {
            out << ", ";
            args[j]->print(out,true);
        }
    }

    out << ")";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionType::addArg(Decl *arg)
{
    if (size == nArgs)
    {
        if (size == 0)
            size = 4;
        else
            size += size;

        Decl   **oldArgs = args;

        args = new Decl* [size];

        for (int j=0; j < nArgs; j++)
        {
            args[j] = oldArgs[j];
        }

        delete [] oldArgs;
    }

    args[nArgs] = arg;
    nArgs++;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionType::getParams(treeNode *node)
{
    if (node)
    {
        switch (node->which)
        {
            default:
                assert(0);
                break;
    
            case LEAF_T:
                switch (node->type)
                {
                    default:
                        assert(0);
                        break;

                    case TN_IDENT:
                        {
                        // This can happen for K&R-style declarations.

                          // Declare as "int" for now, we'll fix it later.
                          Decl *arg = new Decl();
                          buildDecl(arg,node);
                          addArg(arg);
                        }
                        break;

                    case TN_ELLIPSIS:
                        {
                          Decl *arg = new Decl();
                          arg->baseType = BT_Ellipsis;
                          addArg(arg);
                        }
                        break;
                }
                break;

            case NODE_T:
                switch (node->type)
                {
                    default:
                        break;
   
                    case TN_DECL:
                        {
                          Decl *arg = new Decl();
                          buildDecl(arg,node);

                          // Hook the arg's symtable entry back to here.
                          if ((arg->name != NULL) && (arg->name->entry != NULL))
                          {
                            arg->name->entry->uVarDecl = arg;
                          }

                          addArg(arg);
                        }
                        break;
 
                    case TN_IDENT_LIST:
                            // This can happen for K&R-style declarations.

                    case TN_PARAM_LIST:
                        getParams(node->lnode);
                        getParams(node->rnode);
                        break;
                }
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionType::fixParam(Decl *arg)
{
    for (int j=0; j < nArgs; j++)
    {
        if (args[j]->name->name->equal(arg->name->name))
        {
            delete args[j];
            args[j] = arg->dup();
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionType::findExpr( fnExprCallback cb )
{
    if (subType)
        subType->findExpr(cb);

    for (int j=0; j < nArgs; j++)
    {
        args[j]->findExpr(cb);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
StructType::StructType()
           : Type(TT_Struct)
{
    tag = NULL;
    size = 0;
    nComponents = 0;

    components = NULL;
}

StructType::~StructType()
{
    delete tag;

    for (int j=0; j < nComponents; j++)
    {
        delete components[j];
    }

    delete [] components;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Decl*
StructType::lastDecl()
{
    return components[nComponents-1];
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type*
StructType::dup() const
{
    StructType *ret = NULL;

    if (this)
    {
        ret = new StructType();

        ret->size    = size;
        ret->components = new Decl* [size];

        for (int j=0; j < nComponents; j++)
            ret->addComponent(components[j]->dup());

        ret->type = type;
        ret->tag = tag->dup();
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
StructType::print(ostream& out, Symbol *name, int level) const
{
    if (tag)
        out << *tag;

    out << "{\n"; 

    for (int j=0; j < nComponents; j++)
    {
        indent(out,level+1);
        components[j]->print(out,true,level+1);
        out << ";\n";
    }

    indent(out,level);
    out << "}"; 

    if (name)
        out << " " << *name;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
StructType::findExpr( fnExprCallback cb )
{
    for (int j=0; j < nComponents; j++)
    {
        components[j]->findExpr(cb);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
StructType::addComponent(Decl *comp)
{
    if (size == nComponents)
    {
        if (size == 0)
            size = 4;
        else
            size += size;

        Decl **oldComps = components;

        components = new Decl* [size];

        for (int j=0; j < nComponents; j++)
        {
            components[j] = oldComps[j];
        }

        delete [] oldComps;
    }

    components[nComponents] = comp;
    nComponents++;

    // Hook this component's symtable entry back to here.
    if ((comp->name != NULL) && (comp->name->entry != NULL))
    {
        comp->name->entry->uComponent = comp;
        comp->name->entry->Container = this;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Decl*
StructType::findComponent( Symbol* fieldName )
{
    Decl* result = NULL;

    for (int j=0; j < nComponents; j++)
    {
        if (components[j]->name->name->equal(fieldName->name))
        {
            result = components[j];
            break;
        }
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
StructType::getStructComponents(treeNode *node)
{
    Decl  *decl;

    if (node)
    {
        switch (node->type)
        {
            default:
                cerr << "Unknown type: " << name_of_node(node->type)
                     << endl;
                assert(0);
                break;

            case TN_IDENT:
            case TN_PNTR:
            case TN_ARRAY_DECL:
            case TN_FUNC_DECL:
            case TN_BIT_FIELD:
                decl = lastDecl();

                if (decl->name)
                {
                    // That one is already in use.
                    decl = decl->halfdup();
                    buildDecl(decl,node);
                    addComponent(decl);
                }
                else 
                {
                    buildDecl(decl,node);

                    // Hook this component's symtable entry back to here.
                    if (decl->name->entry != NULL)
                    {
                        decl->name->entry->uComponent = decl;
                    }
                }

                break;

            case TN_TYPE_LIST:
                decl = new Decl();
                buildBaseType(decl,node);
                addComponent(decl);
                break;

            case TN_FIELD_LIST:
            case TN_DECL:
            case TN_COMP_DECL:
            case TN_DECLS:
                getStructComponents(node->lnode);
                getStructComponents(node->rnode);
                break; 
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
EnumType::EnumType()
         : Type(TT_Enum)
{
    tag = NULL;
    size = 0;
    nElements = 0;

    values = NULL;
}

EnumType::~EnumType()
{
    delete tag;

    for (int j=0; j < nElements; j++)
    {
        delete values[j];
    }

    delete [] values;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
EnumType::addElement(Symbol *nme, Expression *val /* =NULL */ )
{
    if (size == nElements)
    {
        if (size == 0)
            size = 4;
        else
            size += size;

        Symbol     **oldNames = names;
        Expression **oldVals = values;

        names  = new Symbol* [size];
        values = new Expression* [size];

        for (int j=0; j < nElements; j++)
        {
            names[j]  = oldNames[j];
            values[j] = oldVals[j];
        }

        delete [] oldNames;
        delete [] oldVals;
    }

    names[nElements]  = nme;
    values[nElements] = val;
    nElements++;

    if (nme->entry != NULL)
    {
        nme->entry->uEnumValue = val;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type*
EnumType::dup() const
{
    EnumType *ret = NULL;

    if (this)
    {
        ret = new EnumType();

        ret->size  = size;
        ret->names = new Symbol* [size];
        ret->values = new Expression* [size];

        for (int j=0; j < nElements; j++)
        {
            Expression *val = values[j]->dup();
            ret->addElement(names[j]->dup(),val);
        }

        ret->tag = tag->dup();
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
EnumType::print(ostream& out, Symbol*, int level) const
{
    if (tag)
        out << *tag;

    out << " { "; 

    if (nElements > 0)
    {
        out << *names[0];

        if (values[0])
        {
            out << " = ";
            values[0]->print(out);
        }

        for (int j=1; j < nElements; j++)
        {
            out << ", ";
            out << *names[j];

            if (values[j])
            {
                out << " = ";
                values[j]->print(out);
            }
        }
    }

    out << " }"; 
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
EnumType::findExpr( fnExprCallback cb )
{
    for (int j=0; j < nElements; j++)
    {
        if (values[j] != NULL)
            values[j]->findExpr(cb);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
EnumType::getEnumElements(treeNode *node)
{
    Symbol     *sym;
    Expression *val;

    if (node)
    {
        switch (node->type)
        {
            case TN_ENUM_LIST:
                getEnumElements(node->lnode);
                getEnumElements(node->rnode);
                break;

            default:
                cerr << "Unknown type: " << name_of_node(node->type)
                     << endl;
                assert(0);
                break;

            case TN_ASSIGN:
                sym = buildSymbol((leafNode*) node->lnode);
                val = buildExpr(node->rnode);
                addElement(sym,val);
                break;

            case TN_IDENT:
                sym = buildSymbol((leafNode*) node);
                addElement(sym);
                break;
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
buildBaseType(Decl *decl, treeNode *node)
{
    if (node)
    {
        switch (node->which)
        {
            default:
                assert(0);
                break;
    
            case LEAF_T:
                assert(node->type == TN_TYPE);
                switch (node->tok)
                {
                    default:
                        cerr << "Unhandled token: "
                             << node->tok << endl;
                        assert(0);
                        break;

                    case AUTO:
                        decl->storage = ST_Auto;
                        break;
                    case STATIC:
                        decl->storage = ST_Static;
                        break;
                    case REGISTR:
                        decl->storage = ST_Register;
                        break;
                    case EXTRN:
                        decl->storage = ST_Extern;
                        break;

                    case CONST:
                        decl->qualifier |= TQ_Const;
                        break;
                    case VOLATILE:
                        decl->qualifier |= TQ_Volatile;
                        break;

                    case VOID:
                        decl->baseType |= BT_Void;
                        break;
                    case CHAR:
                        decl->baseType |= BT_Char;
                        break;
                    case SHORT:
                        decl->baseType |= BT_Short;
                        break;
                    case INT:
                        decl->baseType |= BT_Int;
                        break;
                    case LONG:
                        if (decl->baseType & BT_Long)
                        {
                            // C9X adds "long long" - (gross!)
                            decl->baseType &= ~BT_Long;
                            decl->baseType |= BT_LongLong;
                        }
                        else
                            decl->baseType |= BT_Long;
                        break;
                    case FLOAT:
                        decl->baseType |= BT_Float;
                        break;
                    case DOUBLE:
                        decl->baseType |= BT_Double;
                        break;

                    case SGNED:
                        decl->baseType |= BT_Signed;
                        break;
                    case UNSGNED:
                        decl->baseType |= BT_UnSigned;
                        break;

                    case TYPEDEF:
                        decl->baseType |= BT_TypeDef;
                        break;

                    case TYPEDEF_NAME:
                        decl->baseType |= BT_UserType;
                        decl->typeName = buildSymbol((leafNode*) node);
                        break;
                }
                break;
    
            case NODE_T:
                switch (node->type)
                {
                  case TN_TYPE_LIST:
                    buildBaseType(decl,node->lnode);
                    buildBaseType(decl,node->rnode);
                    break;

                  case TN_BIT_FIELD:
                    assert(0);
                    break;

                  case TN_OBJ_DEF:
                  case TN_OBJ_REF:
                    switch (node->tok)
                    {
                        case STRUCT:
                          decl->baseType |= BT_Struct;
                          break;

                        case UNION:
                          decl->baseType |= BT_Union;
                          break;

                        case ENUM:
                          decl->baseType |= BT_Enum;
                          break;
                    }

                    if (node->lnode)        // the tag
                    {
                        assert(node->lnode->type == TN_IDENT);
                        decl->tag = buildSymbol((leafNode*) node->lnode);
                    }

                    // The following is for definitions only.
                    if (node->rnode)
                    {
                        decl->form = recurBuildType(node->rnode,decl->name);

                        switch (node->tok)
                        {
                            case STRUCT:
                              break;
    
                            case UNION:
                              decl->form->type = TT_Union;
                              break;
    
                            case ENUM:
                              break;
                        }

                        if ((decl->tag != NULL)
                            && (decl->tag->entry != NULL))
                        {
                            decl->tag->entry->uStructDef = decl;
                        }
                    }
                    break;

                  default:
                    assert(0);
                    break;
                }
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Type*
buildType(treeNode *node, Symbol*& typeName)
{
    typeName = NULL;

    return recurBuildType(node,typeName);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Decl*
buildDecl(Decl *decl, treeNode *node)
{
    switch (node->type)
    {
        case TN_DECL:
        case TN_COMP_DECL:
        case TN_TYPE_NME:
            buildBaseType(decl,node->lnode);
            decl->form = buildType(node->rnode,decl->name);
            break;

        case TN_FUNC_DEF:
            {
            forNode *fnode = (forNode*) node;
            buildBaseType(decl,fnode->init);
            decl->form = buildType(fnode->test,decl->name);
            }
            break;

        case TN_ASSIGN:
            decl->initializer = buildExpr(node->rnode);
            decl->form = buildType(node->lnode,decl->name);
            break;

        default:
            if (decl->form)
                buildType(node,decl->name);
            else
                decl->form = buildType(node,decl->name);
            break;
    }

    return decl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
