/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         stemnt.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of statements.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#ifndef    STEMNT_H
#define    STEMNT_H

#include <stdlib.h>
#include <iostream.h>

#include "symbol.h"
#include "express.h"
#include "decl.h"
#include "callback.h"
#include "location.h"


// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum StatementType
{
    ST_NullStemnt=0,              // The null statement.

    ST_DeclStemnt,
    ST_TypedefStemnt,
    ST_ExpressionStemnt,

    ST_IfStemnt,
    ST_SwitchStemnt,

    ST_ForStemnt,
    ST_WhileStemnt,
    ST_DoWhileStemnt,

    ST_ContinueStemnt,
    ST_BreakStemnt,
    ST_GotoStemnt,
    ST_ReturnStemnt,

    ST_Block,

    ST_InclStemnt,           // #include
    ST_EndInclStemnt
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum LabelType
{
    LT_None=0,              // No label - invalid.

    LT_Default,             // default:
    LT_Case,                // case <expr>:

    LT_Goto                 // A named label (goto destination).
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Label
{
  public:
    Label( LabelType labelType );
    Label( Expression *_expr );
    Label( Symbol *sym );
   ~Label();

    Label *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );

    LabelType    type;

    union
    {
      Expression    *expr;    // for LT_Case
      Symbol        *name;    // for LT_Goto
    };

    Label    *next;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

class Statement
{
  public:
    Statement(StatementType stemntType, const Location& location);
    virtual ~Statement();

    void addLabel( Label *lbl );    // Add a label to this statement.

    virtual bool isBlock() const { return false; }
    virtual bool isFuncDef() const { return false; }
    virtual bool isDeclaration() const { return false; }
    virtual bool isTypedef() const { return false; }

    virtual bool needSemicolon() const { return (type != ST_NullStemnt); }

    virtual bool isInclude() const { return false; }
    virtual bool isEndInclude() const { return false; }

    virtual Statement *dup() const; 
    virtual void print(ostream& out, int level) const;

    virtual void findExpr( fnExprCallback cb );
    virtual void findStemnt( fnStemntCallback cb ) { (cb)(this); }

    StatementType    type;

    Label           *lblHead;  // List of labels.
    Label           *lblTail;

    Location        location;

    Statement       *next;     // For making a list of statements.

  friend ostream& operator<< (ostream& out, const Statement& stemnt);
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class InclStemnt : public Statement
{
  public:
    InclStemnt(char *incl, const Location& location);
    virtual ~InclStemnt();

    bool isInclude() const { return true; }
    Statement *dup() const; 
    void print(ostream& out, int level) const;

    bool    isStandard;      // use brackets
    char   *filename;
};

class EndInclStemnt : public Statement
{
  public:
    EndInclStemnt(const Location& location);
    virtual ~EndInclStemnt();

    bool isEndInclude() const { return true; }
    Statement *dup() const; 
    void print(ostream& out, int level) const;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ExpressionStemnt : public Statement
{
  public:
    ExpressionStemnt( Expression *expr, const Location& location);
    virtual ~ExpressionStemnt();

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );

    Expression    *expression;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class IfStemnt : public Statement
{
  public:
    IfStemnt( Expression *_cond,
              Statement *_thenBlk, const Location& location,
	      Statement *_elseBlk = NULL);
    ~IfStemnt();

    bool needSemicolon() const { return false; }

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );
    void findStemnt( fnStemntCallback cb );

    Expression    *cond;

    Statement     *thenBlk;
    Statement     *elseBlk;        // Can be NULL
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class SwitchStemnt : public Statement
{
  public:
    SwitchStemnt(Expression *_cond, Statement *_block,
		 const Location& location );
    ~SwitchStemnt();

    bool needSemicolon() const { return false; }

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );
    void findStemnt( fnStemntCallback cb );

    Expression    *cond;

    Statement     *block;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ForStemnt : public Statement
{
  public:
    ForStemnt( Expression *_init, Expression *_cond, Expression *_incr,
	       const Location& location, Statement *_block =NULL);
    ~ForStemnt();

    bool needSemicolon() const { return false; }

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );
    void findStemnt( fnStemntCallback cb );

    Expression    *init;
    Expression    *cond;
    Expression    *incr;

    Statement     *block;        // Can be NULL
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class WhileStemnt : public Statement
{
  public:
    WhileStemnt( Expression *_cond, Statement *_block,
		 const Location& location);
    ~WhileStemnt();

    bool needSemicolon() const { return false; }

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );
    void findStemnt( fnStemntCallback cb );

    Expression    *cond;
    Statement     *block;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class DoWhileStemnt : public Statement
{
  public:
    DoWhileStemnt( Expression *_cond, Statement *_block,
		   const Location& location);
    ~DoWhileStemnt();

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );
    void findStemnt( fnStemntCallback cb );

    Expression    *cond;
    Statement     *block;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class GotoStemnt : public Statement
{
  public:
    GotoStemnt( Symbol *_dest, const Location& location);
    ~GotoStemnt();

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    Symbol        *dest;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ReturnStemnt : public Statement
{
  public:
    ReturnStemnt( Expression *_result, const Location& location);
    ~ReturnStemnt();

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );

    Expression    *result;        // Can be NULL.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class DeclStemnt : public Statement
{
  public:
    DeclStemnt(const Location& location, StatementType stype = ST_DeclStemnt);
    virtual ~DeclStemnt();

    // Declarations print their own semicolon.
    bool needSemicolon() const { return false; }
    bool isDeclaration() const { return true; }

    Statement *dup() const; 
    virtual void print(ostream& out, int level) const;

    // Only converts if necessary.
    DeclStemnt *convertToTypedef();

    void addDecl(Decl *decl);

    void findExpr( fnExprCallback cb );

    Decl     base;

    // The type(s).
    int      size;
    int      nDecls;
    Decl   **decls;
};

class TypedefStemnt : public DeclStemnt
{
  public:
    TypedefStemnt(const Location& location);
   ~TypedefStemnt();

    bool isTypedef() const { return true; }

    void print(ostream& out, int level) const;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Block : public Statement
{
  public:
    Block(const Location& location);
    virtual ~Block();

    void add(Statement *stemnt);

    void addStatements(treeNode *node);

    virtual bool isBlock() const { return true; }
    bool needSemicolon() const { return false; }

    virtual Statement *dup() const; 
    virtual void print(ostream& out, int level) const;

    virtual void findExpr( fnExprCallback cb );
    void findStemnt( fnStemntCallback cb );

    void insert(Statement *stemnt, Statement *after=NULL);

    Statement     *head;    // List of statements in this block.
    Statement     *tail;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// Not really true, but this is very convienent.
class FunctionDef : public Block
{
  public:
    FunctionDef(const Location& location);
   ~FunctionDef();

    bool isFuncDef() const { return true; }

    Statement *dup() const; 
    void print(ostream& out, int level) const;

    void findExpr( fnExprCallback cb );

    Symbol  *FunctionName() const;

    Decl          *decl;    // The declaration.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

Label      *buildLabel(treeNode *node);
Statement  *buildStatement(treeNode *node);
DeclStemnt *buildDeclStemnt(DeclStemnt *stemnt, treeNode *node);

void        indent(ostream& out, int level);
void        printNull(ostream& out, int level);
void        printBlock(ostream& out, int level, Statement *block);


// For debugging.
char *nameOfStatementType( StatementType type );
char *nameOfLabelType( LabelType type );

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif  /* STEMNT_H */

